/*

 * xlate.c
 *
 * Format translation code (YUV->RGB 8, 16, 24)
 *
 *
 *
 * Mark Vrabel, Jan 96
 */

#include "xlate.h"

/*
/*
 * This module provides translation from YUV into RGB. 
 *
 * We convert on copying to 8-bit palettised, RGB555 or RGB-24, using
 * translation tables in all cases. 
 *
 * The (128 kbyte) rgb555 lookup table is built by BuildYUVToRGB555 whenever
 * the destination format is set to RGB555. This translates from YUV655 (6-bits of
 * Y) to RGB555 for better colour resolution at the cost of doubling the
 * table size.
 *
 * The (32 kbyte) palette lookup table is built whenever the format is set to
 * FmtPal8 or the palette is changed. We are not given a palette here - we are
 * given a rgb555-to-palette lookup table, and we build from it
 * a yuv-to-palette lookup. This translates from YUV555.
 *
 * The (128kbyte) rgb24 lookup table is built by BuildYUVToRGB24 whenever the
 * destination format is set to RGB24. This translates from YUV555 into 4-byte
 * RGBQUAD values.
 *
 */


/*
 * force x to be in the range lo to hi
 */
#define RANGE(x, lo, hi)	max(lo, min(hi, x))



/*
 * Convert a YUV colour into a 15-bit RGB colour.
 *
 * The input Y is in the range 16..235; the input U and V components
 * are in the range -128..+127. The conversion equations for this are
 * (according to CCIR 601):
 *
 *	R = Y + 1.371 V
 *	G = Y - 0.698 V - 0.336 U
 *	B = Y + 1.732 U
 *
 * To avoid floating point, we scale all values by 1024.
 *
 * The resulting RGB values are in the range 16..235: we truncate these to
 * 5 bits each. and return a WORD containing 5-bits each for R, G and B
 * with bit 15 set to 0.
 */
WORD
YUVToRGB15(int y, int u, int v)
{
    int ScaledY = RANGE(y, 16, 235) * 1024;
    int red, green, blue;
	// added rounding (+4) because of 5 bit output
#if 1 // old equations
    red = RANGE((ScaledY + (1404 * v)) / 1024+4, 0, 255);
    green = RANGE( (ScaledY - (715 * v) - (344 * u)) / 1024+4, 0, 255);
    blue = RANGE( (ScaledY + (1774 * u)) / 1024+4, 0, 255);
#else
    red = RANGE((ScaledY + (1398 * v)) / 1024, 0, 255);
    green = RANGE( (ScaledY - (717 * v) - (342 * u)) / 1024, 0, 255);
    blue = RANGE( (ScaledY + (1773 * u)) / 1024, 0, 255);
#endif
    return (WORD) (((red & 0xf8) << 7) | ((green & 0xf8) << 2) | ((blue & 0xf8) >>3) );
}



/*
 * Convert a YUV colour into 24-bit RGB colour.
 *
 * The input Y is in the range 16..235; the input U and V components
 * are in the range -128..+127. The conversion equations for this are
 * (according to CCIR 601):
 *
 *	R = Y + 1.371 V
 *	G = Y - 0.698 V - 0.336 U
 *	B = Y + 1.732 U
 *
 * To avoid floating point, we scale all values by 1024.
 *
 * The resulting RGB values are in the range 16..235: we return these in a
 * DWORD with 8 bits each for red, green and blue, and 8 bits free at
 * the top.
 */
DWORD
YUVToRGBQUAD(int y, int u, int v)
{
    int ScaledY = RANGE(y, 16, 235) * 1024;
    int red, green, blue;
    DWORD dwResult;

    red = RANGE((ScaledY + (1404 * v)) / 1024, 0, 255);
    green = RANGE( (ScaledY - (715 * v) - (344 * u)) / 1024, 0, 255);
    blue = RANGE( (ScaledY + (1774 * u)) / 1024, 0, 255);


    dwResult = ((red & 0xff) << 16) | ((green & 0xff) << 8) | (blue & 0xff);

    return dwResult;

}


/*
 * build a translation table to translate between YUV and RGB555.
 *
 * This builds a lookup table with 64k 1-word entries: truncate the YUV
 * to 16bits (5-5-5) and look-up in this xlate table to produce the
 * 15-bit rgb value.
 */
PVOID
BuildYUVToRGB555(void)
{
    PVOID pXlate;
    PWORD pRgb;
	UINT w;
    ULONG ulSizeXlate = sizeof(WORD) * 64 * 1024;  /* enough for 64k 16-bit values */

    pXlate = new char [ulSizeXlate];
    if (pXlate == NULL) {
		return(NULL);
    }

	pRgb = (PWORD)pXlate;

    /*
     * build a 16-bit yuv lookup table by stepping through each entry,
     * converting the yuv index to rgb and storing at that index. The index
     * to this table is a 16-bit value with the (6-bit) y in bits 15..10,
     * u in bits 9..5 and v in bits 4..0. All three components are unsigned.
     */
    for (w = 0; w < 64*1024; w++) {

		/*
		 * the YUVtoRGB55 conversion function takes values 0..255 for y,
		 * and -128..+127 for u and v. Pick out the relevant bits of the
		 * index for this cell, and shift to get values in this range.
		 * Subtract 128 from u and v to shift from 0..255 to -128..+127
		 */
 		
		*pRgb++ = YUVToRGB15(
			   	    (w &  0xfc00) >> 8,
				    ((w & 0x3e0) >> 2) - 128,
					((w & 0x1f) << 3) - 128);
    }

    return(pXlate);
}


/*
 * build a palette translation table.
 *
 * Other code has already converted the palette into a translation
 * table with one entry for each 15-bit RGB value, giving the relevant
 * 8-bit palette entry. We are passed a pointer to this table (protected
 * by VC_AccessData, so we need not worry about access violations).
 *
 * Indices into this table are 15-bit yuv (5:5:5) colours. We step through
 * each possible index, separate out the y, u and v components and convert
 * that into an RGB (5:5:5) colour. We then use the passed translation table
 * to convert the rgb colour to a palette index, and store that pal index
 * as the translation for the yuv colour used as index.
 *
 * The palette table we are passed in is user-mode memory: we create
 * our yuv->pal lookup in non-paged memory so we can use it at interrupt time.
 *
 * We are called via VC_AccessData: this passes us a PVOID context pointer that
 * is not used. VC_AccessData handles invalid data exceptions for us, so we
 * don't need to worry about bad user data.
 */

PVOID
BuildYuvToPal(
    PUCHAR pBuffer,
    ULONG Length
)
{
    PVOID pXlate;
    PUCHAR pRgb;
    int wRGB;
    UINT w;

    ULONG ulSizeXlate = 32 * 1024;	/* enough for 32k 1-byte entries */

    /* check that the length of the palette xlate table is good */
    if (Length < (32 * 1024)) {
		return (FALSE);
    }

    /* allocate our non-paged xlate table */
    pXlate = new char [ulSizeXlate];
    if (pXlate == NULL) {
    	return(FALSE);
    } 

	pRgb = (PUCHAR)pXlate;

    /*
     * build a 15-bit yuv lookup table by stepping through each entry,
     * converting the yuv index to rgb and then to a palette index. The index
     * to this table is a 15-bit value with the y component in bits 14..10,
     * u in bits 9..5 and v in bits 4..0. All three components are unsigned.
     */
    for (w = 0; w < 32*1024; w++) {

		/*
		 * the YUVtoRGB55 conversion function takes values 0..255 for y,
		 * and -128..+127 for u and v. Pick out the relevant bits of the
		 * index for this cell, and shift to get values in this range.
		 * Subtract 128 from u and v to shift from 0..255 to -128..+127
		 */
 		wRGB = YUVToRGB15(
		   		(w &  0x7c00) >> 7,
			    ((w & 0x3e0) >> 2) - 128,
			    ((w & 0x1f) << 3) - 128);
		/*
		 * having got the RGB555 value for this YUV index value, we can
		 * use the passed-in table to convert rgb555 to pal-index and
		 * store that as the conversion for this yuv value
		 */

		 *pRgb++ = pBuffer[wRGB];
    }

    return(pXlate);
}


/*
/*
 * build a translation table from yuv-555 to rgb-24. The table contains
 * 32k entries, each of which is a DWORD containing an RGBQUAD.
 */
PVOID
BuildYUVToRGB24(void)
{
    PVOID pXlate;
    PDWORD pRgb;
    UINT w;
    ULONG ulSizeXlate = sizeof(DWORD) * 32 * 1024;	/* enough for 32k 32-bit values */

    pXlate = new char [ulSizeXlate];
    if (pXlate == NULL) {
    	return(NULL);
    }

	pRgb = (PDWORD)pXlate;

    /*
     * build a 15-bit yuv lookup table by stepping through each entry,
     * converting the yuv index to rgb and store the rgb value. The index
     * to this table is a 15-bit value with the y component in bits 14..10,
     * u in bits 9..5 and v in bits 4..0. All three components are unsigned.
     */
    for (w = 0; w < 32*1024; w++) {

	/*
		 * the YUVtoRGB55 conversion function takes values 0..255 for y,
		 * and -128..+127 for u and v. Pick out the relevant bits of the
		 * index for this cell, and shift to get values in this range.
		 * Subtract 128 from u and v to shift from 0..255 to -128..+127
		 */
 		*pRgb++ = YUVToRGBQUAD(
		   			(w &  0x7c00) >> 7,
					((w & 0x3e0) >> 2) - 128,
					((w & 0x1f) << 3) - 128);
    }

    return(pXlate);
}


VOID
CopyYUVToRGB555(
    PWORD pXlate,		/* translation table yuv-655 to rgb-555 */
    DWORD Width,		/* width of copy rect in pixels */
    DWORD Height,		/* height of copy rect in lines */
    PUCHAR pDst,		/* destination pixels */
    PUCHAR pSrc,    	/* source pixels */	
    BOOL bYUVSwapBytes   	
)
{
    int RowInc;
    int i, j;
    DWORD uv55, dwPixel;
    int WidthBytes = Width * 2;		// dest pixel size is two bytes
    BOOL bDuplicate = FALSE;
    
    /*
     * adjust the destination to point to the start of the last line,
     * and work upwards (to flip vertically into DIB format)
     */
    pDst += (Height - 1) * WidthBytes;

	/*
	 * calculate the amount to adjust pDst by at the end of one line
	 * of copying. At this point we are at the end of line N. We need
	 * to move to the start of line N-1.
	 */
	RowInc = WidthBytes + (Width * 2);

	/* loop copying each scanline */
	for (i = Height; i > 0; i--) {

	    /* loop copying two pixels at a time */
	    for (j = Width ; j > 0; j -= 2) {

			/*
			 * get two pixels and convert to 16-bpp YUV
			 */
            if (bYUVSwapBytes) {
/*                ((BYTE*)&dwPixel)[0] = pSrc[1];  
                ((BYTE*)&dwPixel)[1] = pSrc[0];  
                ((BYTE*)&dwPixel)[2] = pSrc[3];  
                ((BYTE*)&dwPixel)[3] = pSrc[2];
*/
#if 1
				uv55=(pSrc[0] & 0xf8) <<2 | (pSrc[2] & 0xf8) >> 3;
				dwPixel = pXlate[ ((pSrc[1] & 0xfc) << 8) | uv55 ] |
					(pXlate[((pSrc[3] & 0xfc) << 8) | uv55 ] << 16);
				pSrc += sizeof(DWORD);
#else
				uv55=(pSrc[1] & 0xf8) <<2 | (pSrc[5] & 0xf8) >> 3;
				dwPixel = pXlate[ ((pSrc[3] & 0xfc) << 8) | uv55 ] |
					(pXlate[((pSrc[7] & 0xfc) << 8) | uv55 ] << 16);
				pSrc += sizeof(DWORD)*2;
#endif				
            } else {
			    dwPixel = *(DWORD *)pSrc;
				pSrc += sizeof(DWORD);
				uv55 = ((dwPixel & 0xf8000000) >> 27) |
					((dwPixel & 0x0000f800) >> 6);
				dwPixel = pXlate[ ((dwPixel & 0xfc) << 8) | uv55 ] |
					(pXlate[((dwPixel & 0xfc0000) >> 8) | uv55 ] << 16);
            }
            
            /* swap */

			/*
			 * dwPixel now has two pixels, in this layout (MSB..LSB):
			 *
			 *  V Y1 U Y0
			 *
			 * convert to 2 yuv655 words and lookup in xlate table
			 */

			/* get common u and v components to lower 10 bits */

			/* build each yuv-655 value by truncating
			 * y to 6 bits and adding the common u and v bits,
			 * look up to convert to rgb555, and combine two pixels
			 * into one dword
			 */

			/* write two pixels to destination */
			*(DWORD *)pDst = dwPixel;
			pDst += sizeof(DWORD);

	    } // loop per 2 pixels

	    /* move dest pointer back to next line */
	    pDst -= RowInc;

	} // loop per row
}



/*
 * translate YUV from pSrc into 8-bit palettised data in pDst.
 *
 * pXlate is an array of bytes, one for each 15-bit YUV555 value, giving
 * the corresponding palette entry.
 *
 * dwWidth and dwHeight give the size of the destination rectangle in pixels.
 *
 * Also flip the image vertically into a DIB format.
 *
 */
VOID
CopyYUVToPal8(
    PUCHAR pXlate,		/* translation table yuv-655 to rgb-555 */
    DWORD Width,		/* width of copy rect in pixels */
    DWORD Height,		/* height of copy rect in lines */
    PUCHAR pDst,		/* destination pixels */
    PUCHAR pSrc 		/* source pixels */		
)
{
    int RowInc;
    int i, j;
    DWORD uv55;
    int WidthBytes = Width;

    /*
     * adjust the destination to point to the start of the last line,
     * and work upwards (to flip vertically into DIB format)
     * if duplicating, remember we fill 2 lines worth
     */
    pDst += (Height - 1) * Width;

	/*
	 * calculate the amount to adjust pDst by at the end of one line
	 * of copying. At this point we are at the end of line N. We need
	 * to move to the start of line N-1.
	 *
	 */
	RowInc = WidthBytes + Width;

	/* loop copying each scanline */
	for (i = Height; i > 0; i--) {

	    /* loop copying four pixels at a time */
	    for (j = Width; j > 0; j -= 4) {

			DWORD dw1, dw2, destpix;

			/*
			* get two pixels and convert to 15-bpp YUV-555
			*/
			dw1 = *(DWORD *)pSrc;
			pSrc += sizeof(DWORD);
			dw2 = *(DWORD *)pSrc;
			pSrc += sizeof(DWORD);
	
			/*
			 * dwPixel now has two pixels, in this layout (MSB..LSB):
			 *
			 *  V Y1 U Y0
			 *
		     * convert to 2 yuv555 words and lookup in xlate table
		     */

		    /* get common u and v components to lower 10 bits */
			uv55 = ((dw1) >> 27) |
				((dw1 & 0x0000f800) >> 6);


			/* build each yuv-555 value by truncating
			 * y to 5 bits and adding the common u and v bits,
			 * look up to convert to palindex, and combine two pixels
			 * into one word.
			 */
			destpix = pXlate[ ((dw1 & 0xf8) << 7) | uv55 ] |
				(pXlate[((dw1 & 0xf80000) >> 9) | uv55 ] << 8);

            /* common u and v for second pixel pair */
			uv55 = ((dw2) >> 27) |
				((dw2 & 0x0000f800) >> 6);


            /* build second pixel pair and write all 4 pixels */
			destpix = pXlate[ ((dw1 & 0xf8) << 7) | uv55 ] |
				(pXlate[((dw1 & 0xf80000) >> 9) | uv55 ] << 8);

                *(DWORD *) pDst =
						(pXlate[ ((dw2 & 0xf8) << 7) | uv55 ] |
                        (pXlate[((dw2 & 0xf80000) >> 9) | uv55 ] << 8)
                        ) << 16 | destpix;

			pDst += sizeof(DWORD);
	    }

	    /* move dest pointer back to next line */
	    pDst -= RowInc;

	} // loop per row
}


/*
 * translate YUV from pSrc into 24-bit RGB in pDst.
 *
 * pXlate points to a yuv555 to rgb conversion table. This contains 32k entries
 * of one dword, each entry having (msb..lsb) 00RRGGBB for that yuv555 value.
 *
 * dwWidth and dwHeight give the size of the copy rectangle in pixels.
 * WidthBytes is the width of one source line in bytes.
 *
 * Also flip the image vertically into a DIB format.
 *
 */
VOID
CopyYUVToRGB24(
    PDWORD pXlate,		/* translation table yuv-655 to rgb-555 */
    DWORD Width,		/* width of copy rect in pixels */
    DWORD Height,		/* height of copy rect in lines */
    PUCHAR pDst,		/* destination pixels */
    PUCHAR pSrc 		/* source pixels */		
)
{
    int RowInc;
    int i, j;
    DWORD uv55, dwPixel, dwRGB;
    int WidthBytes = Width * 3;		// dest pixel size is three bytes

    /*
     * adjust the destination to point to the start of the last line,
     * and work upwards (to flip vertically into DIB format)
     */
    pDst += (Height - 1) * WidthBytes;

	/*
	 * calculate the amount to adjust pDst by at the end of one line
	 * of copying. At this point we are at the end of line N. We need
	 * to move to the start of line N-1.
	 */
	RowInc = WidthBytes + (Width * 3);

	/* loop copying each scanline */
	for (i = Height; i>0; i--) {

	    /* loop copying two pixels at a time */
	    for (j = Width ; j > 0; j -= 2) {

			/*
			 * get two pixels and convert to  YUV-555
			 */

			dwPixel = *(DWORD *)pSrc;
			pSrc += sizeof(DWORD);

			/*
			 * dwPixel now has two pixels, in this layout (MSB..LSB):
			 *
			 *  V Y1 U Y0
			 *
			 * convert to 2 yuv555 words and lookup in xlate table
			 */

			/* get common u and v components to lower 10 bits */
			uv55 = ((dwPixel & 0xf8000000) >> 27) |
				((dwPixel & 0x0000f800) >> 6);

			/*
			 * truncate y to 5 bits and add in common u, v bits. then
			 * lookup to convert to rgb dword.
			 */
			dwRGB = pXlate[ ((dwPixel & 0xf8) << 7) | uv55 ];

			/* write the red, green and blue bytes to destination */
			*(WORD UNALIGNED *) pDst = LOWORD(dwRGB);
			pDst += sizeof(WORD);
			*pDst = LOBYTE(HIWORD(dwRGB));
			pDst += sizeof(BYTE);

			/*
			 * same again for second pixel
			 */
			dwRGB = pXlate[((dwPixel & 0xf80000) >> 9) | uv55 ];

			/* write the red, green and blue bytes to destination */
			*(WORD UNALIGNED *) pDst = LOWORD(dwRGB);
			pDst += sizeof(WORD);
			*pDst = LOBYTE(HIWORD(dwRGB));
			pDst += sizeof(BYTE);

		} // loop per 2 pixels

	    /* move dest pointer back to next line */
	    pDst -= RowInc;

	} // loop per row
}
