#include <windows.h>
#include <windowsx.h>

#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "fft.h"

//internal function prototypes

/*
BITREV macro:
    BITREV(num, width)
This macro returns the bit reverse of the specified number
based on the specified width.  For speed the macro uses the
BitRevTab lookup table.  This limits num to < MAX_FFT_SIZE and
width to < LOG2_MAX_FFT_SIZE.  If either of these limits are 
exceded, the macro returs 0xffff,
*/


#define BITREV(num,width) \
    ( ((num) >= MAX_FFT_SIZE || (width) > LOG2_MAX_FFT_SIZE) ? 0xffff : \
            BitRevTab[(num) << (LOG2_MAX_FFT_SIZE-(width))] )

//global data
#define LOG2_MAX_FFT_SIZE 13
#define MAX_FFT_SIZE (1 << LOG2_MAX_FFT_SIZE )
#define SINE_TABLE_SIZE (MAX_FFT_SIZE*3/2)
#define MAX_WINDOW_SIZE (2*MAX_FFT_SIZE)

float *SineTab = NULL;        //3/4 sine table (form 0 to 3*pi/2, size = MAX_FFT_SIZE*3/2) 
WORD *BitRevTab = NULL;       //bit rev lookup table has length of MAX_FFT_SIZE
COMPLEX *fftdata = NULL;       //temp storage for fft calc, has length of MAX_FFT_SIZE
static inited = FALSE;              //flag TRUE if fft tabels are initialized
HANDLE hmodDLL;


void FreeFFT(void)
/*
de-allocate global memory
*/
{
    if (SineTab != NULL)
        GlobalFreePtr(SineTab);
    if (BitRevTab != NULL)
        GlobalFreePtr(BitRevTab);
    if (fftdata != NULL)
        GlobalFreePtr(fftdata);
        
    inited = FALSE;
    SineTab = NULL;
    BitRevTab = NULL;
    fftdata = NULL;
}
 
ECODE BuildWindow(WINDOWTYPE WindowType,UINT WindowSize,float *window)
{               
    UINT i,sz,j;
    double a0,a1,a2,a3,p,pinc;
    float total,scale;
                      
    //process rectangle window                      
    if (WindowType == RECTANGLE) {
        for (i=0;i<WindowSize;i++)
            window[i] = 1.0F;
    }
    
    //process other windows
    else {
        switch (WindowType) {   //fetch coefs
        
        case HANN:
            a0 = .8164966;
            a1 = .8164966;
            a2 = 0.0;
            a3 = 0.0;
            break;
        case HANNING:
            a0 = .5;
            a1 = .5;
            a2 = 0.0;
            a3 = 0.0;
            break;
        case HAMMING:
            a0 = .54;
            a1 = .46;
            a2 = 0.0;
            a3 = 0.0;
            break;
        case BLACKMAN:
            a0 = .42;
            a1 = .50;
            a2 = .08;
            a3 = 0.0;
            break;
        case BLACKMAN_HARRIS_67:
            a0 = .42323;
            a1 = .49755;
            a2 = .07922;
            a3 = 0.0;
            break;
        case BLACKMAN_HARRIS_92:
            a0 = .35875;
            a1 = .48829;
            a2 = .14128;
            a3 = .01168;
            break;
        default:
            return OLDSP_BAD_WINDOW_TYPE;
        }  //end switch WindowType
        
        //build the window
        pinc = 4.0*acos(0.0)/WindowSize;    // = 2pi/size
        //set start phase to .5*pinc if size is even, 0 if size is odd
        p = (WindowSize & 1) ? 0 : .5*pinc;
        total = 0.0F;        
        sz = WindowSize/2;          //round down to half size
        for (i=0;i<sz;i++) {
            window[i] = (float) (a0 - a1*cos(p) + a2*cos(2*p) - a3*cos(3*p));
            total += window[i];
            p += pinc;
        }
        total += total;     //add the other half of window
        //if odd, include the half way point
        if (WindowSize & 1) {       
            window[i] = (float) (a0 - a1*cos(p) + a2*cos(2*p) - a3*cos(3*p));
            total += window[i];
        }
        
        //normalize window if hanning, hamming, blackman, or blackman harris
        if (WindowType >= HANNING) {
            scale = WindowSize/total;
            sz = (WindowSize+1)/2;          //round up to half size
            for (i=0;i<sz;i++)
                window[i] *= scale;
        }

        //copy 2nd half of window
        sz = WindowSize/2;
        for(i=0, j=WindowSize-1; i<sz; i++, j--)
            window[j] = window[i];

    }  //end else (non-rectangle windows)
    
    return OLNOERROR;
}




ECODE DoRealFFT(float *in, float *power, float *logmag, float *phase, int log2size)
/*
This function performs a real fft on the specified input array using
the specified fft size.  (Size is passed as log to the base 2 of the fft size.)
The function writes the magnituce, the log magnitude (in dB) and the phase of the
frequency data into the specified data
arrays.  (if the data pointer is NULL, no data is written.)  The function returns
the appropriate status code.
*/
{
    ECODE status;
    int i;
    int log2_complex_size = log2size - 1;
    int complex_fft_size = 1 << log2_complex_size;
    COMPLEX *ComplexData;
    float powval;

    if (complex_fft_size > MAX_FFT_SIZE)
        return OLDSP_FFT_SIZE_TOO_LARGE;
        
    //init fft tables
    if ((status = InitFFT()) != OLNOERROR) 
        return status;
    
    //bit rev input and pack real data into imag part
    ComplexData = (COMPLEX *)in;
    for (i=0;i<complex_fft_size;i++) {
        int rev = BITREV(i,log2_complex_size);
        fftdata[i] = ComplexData[rev];
    }
    
    //perform complex fft
    DoFFT(fftdata,log2_complex_size);
    
    //perform even/odd un-winding
    Unwind(fftdata,complex_fft_size);
    
/*  comment out the direct computation (without unwinding)
    //bit rev input and pack real data into imag part
    for (i=0;i<complex_fft_size*2;i++) {
        int rev = BITREV(i,log2_complex_size+1);
        fftdata[i].re = in[rev];
        fftdata[i].im = 0.0F;
    }
    
    //perform complex fft
    DoFFT(fftdata,log2_complex_size+1);
*/    

    //calc phase
    if (phase != NULL) {
        for (i=0;i<complex_fft_size;i++)
            phase[i] = (float)atan2(fftdata[i].im,fftdata[i].re);
    }

    //calc magnitude & dB
    if (power != NULL && logmag != NULL) {
        for (i=0;i<complex_fft_size;i++) {
            powval = fftdata[i].im * fftdata[i].im + fftdata[i].re * fftdata[i].re;
            power[i] = powval;
            logmag[i] = PowerTodB(powval);
        }
    }
    else {
        if (power != NULL) {
            for (i=0;i<complex_fft_size;i++)
                power[i] = fftdata[i].im * fftdata[i].im + fftdata[i].re * fftdata[i].re;
        }
        if (logmag != NULL) {
            for (i=0;i<complex_fft_size;i++)
                logmag[i] = PowerTodB(fftdata[i].im * fftdata[i].im + fftdata[i].re * fftdata[i].re);
        }
    }
        
    return OLNOERROR;
}

ECODE InitFFT(void)
/*
This function is used to initialize fft sine and bit reverse tables.
The function returns the appropriate status code.  (The
function will only fail if it can not allocate table memory)  
*/
{   
    double pi;
    int i;
    
    if (inited)
        return OLNOERROR;
    
    //allocate memory
    SineTab = (float *)GlobalAllocPtr(GMEM_FIXED | GMEM_SHARE, (long)sizeof(float)*SINE_TABLE_SIZE);
    if (SineTab == NULL)
        return OLDSP_CANT_ALLOC_MEMORY;
    BitRevTab = (WORD *)GlobalAllocPtr(GMEM_FIXED | GMEM_SHARE, sizeof(WORD)*MAX_FFT_SIZE);
    if (BitRevTab == NULL) {
        GlobalFreePtr(SineTab);
        return OLDSP_CANT_ALLOC_MEMORY;
    }
    fftdata = (COMPLEX *)GlobalAllocPtr(GMEM_FIXED | GMEM_SHARE, (long)sizeof(COMPLEX)*MAX_FFT_SIZE);
    if (BitRevTab == NULL) {
        GlobalFreePtr(SineTab);
        GlobalFreePtr(BitRevTab);
        return OLDSP_CANT_ALLOC_MEMORY;
    }     
               
    //fill sine table with half sine wave
    pi = 2.0 * acos(0.0);
    for (i=0;i<SINE_TABLE_SIZE;i++)
        SineTab[i] = (float)sin(pi * i / MAX_FFT_SIZE);
    
    //fill bit rev table
    for (i=0;i<MAX_FFT_SIZE;i++) {
        WORD k,rev = 0;
        WORD num = i;
        //calc bit rev by shifting number down and bet rev up
        for (k=0;k<LOG2_MAX_FFT_SIZE;k++) {
            rev = (rev << 1) | (1 & num);
            num = num >> 1;
        }
        BitRevTab[i] = rev;
    }
    
    inited = TRUE;
    return OLNOERROR;   
}
                    
void DoFFT(COMPLEX *x,int log2size)
/*
This function computes a complex radix 2 DIT FFT.  The function computes the
algorithm in place on the specified data using the specified fft size 
(Size is passed as log to the base 2 of the fft size.)  This function makes
use of the SineTab global array to provide coefs
*/
{                                      
    int j,pass;
    int GroupPoints;
    int GroupBflys;
    int ci,cstep;
    register int i,ip;
    COMPLEX w,t;
    int fft_size = 1 << log2size;

    //loop for each pass
    GroupPoints = 2;
    GroupBflys = 1;
    cstep = MAX_FFT_SIZE;    //init coef inc to pi
    for (pass=1;pass<=log2size;pass++) {
    
        ci = 0;  //init coef index
        
        //loop for each butterfly in 'group'
        for (j=0;j<GroupBflys;j++) {
            
            //get coef for this (the jth) butterfly 
            w.im = -SineTab[ci];
            w.re =  SineTab[ci+(MAX_FFT_SIZE/2)]; //cos(a) = sin(a+pi/2)
            
            //loop for each group - perform the jth butterfly in each group
            for (i=j;i<fft_size;i+=GroupPoints) {
                /* perform butterfly */
                ip = i + GroupBflys;
                t.re = x[ip].re * w.re - x[ip].im * w.im;
                t.im = x[ip].re * w.im + x[ip].im * w.re;
                x[ip].re = x[i].re - t.re;
                x[ip].im = x[i].im - t.im;
                x[i].re += t.re;
                x[i].im += t.im;
            }
    
            ci += cstep;   //setup for the next coef
    
        }  /* each pass */
    
        //setup for next pass   
        GroupPoints <<= 1;
        GroupBflys <<= 1;
        cstep >>= 1;
    
    } /* each pass */ 
}

void Unwind(COMPLEX *x, int ComplexFftSize)
/*
This function performs the even-odd unwinding pass that turns an
N point complex FFT into a 2N real FFT (You prepare for this by
packing real data into the imag part of the complex input array.)
The function works in place on the input array using the spcified
FFT size (expressed as the size of the complex FFT -> = real fft size/2
*/

{                                        
    int i,ip;
    int ci,cstep;
    int half_size = ComplexFftSize/2;
    float a,b,c,d,e,f;    //1st point = a + jb, 2nd point = c + jd, coef = e +jf
    COMPLEX t;            //temp value to hold results of complex mult
                  
    //compute 0 and N/2 points
    x[0].re = x[0].re + x[0].im;
    x[0].im = 0.0F;
    x[ComplexFftSize/2].im = -x[ComplexFftSize/2].im;

    //compute all other points
    cstep = MAX_FFT_SIZE/ComplexFftSize;
    
    for (i=1,ci=cstep; i<half_size; i++, ci+=cstep) {
        ip = ComplexFftSize - i;
        e =  SineTab[ci+(MAX_FFT_SIZE/2)]; //cos(a) = sin(a+pi/2)
        f = -SineTab[ci];
        
        //compute unwind butterfly
            //unwind the data (you know use that old fft even-odd symetry)
        a = (x[i].re + x[ip].re) / 2.0F;
        b = (x[i].im - x[ip].im) / 2.0F;
        c = (x[i].im + x[ip].im) / 2.0F;
        d = (x[i].re - x[ip].re) / 2.0F;
             //complex multiply
        t.re = c*e + d*f;
        t.im = c*f - d*e;
             //complex add and complex subtract
        x[i].re = a + t.re;
        x[i].im = b + t.im;
        x[ip].re = a - t.re;
        x[ip].im = t.im - b;                   
    }
    
}


