/*
 * bwuser.h
 *
 * 32-bit Video Capture / Recompression  driver
 * Broadway App User-mode support library
 *
 * define functions providing access to video capture and 
 * recompression hardware, by accessing the kernel-mode 
 * device driver. 
 *
 * Mark Vrabel, March 96:
 */

#ifndef _BWUSER_H_
#define _BWUSER_H_


/* 
 * define audio settings
 */
#define AUDIO_NONE			0
#define AUDIO_MONO			1
#define AUDIO_JOINTSTEREO	2
#define AUDIO_DUALCHANNEL	3
#define AUDIO_STEREO		4


#define AUDIO_RATE_11KHZ    11025
#define AUDIO_RATE_22KHZ    22050
#define AUDIO_RATE_44KHZ    44100

/*
 * define error strings
 */
#define IDS_RT_FIRST_STRING				    2000	/* first non AVICAP string */
#define IDS_RT_WAVE_FORMAT_NOT_PCM_ERROR	2000	/* "Error: Audio format must be set to PCM for realtime capture." */
#define IDS_RT_WAVE_FORMAT_NOT_16BIT_ERROR	2001	/* "Error: Audio format must be set to 16 bit sample size for realtime capture." */
#define IDS_RT_VIDEO_DROP_ERROR				2002	/* "Error: video data was lost during capture, reduce capture rate." */
#define IDS_RT_STAT_AUDIOCURRENT			2003	/* "Captured %ld audio bytes %d.%03d sec." */
#define IDS_RT_STAT_AUDIOONLY				2004    /* "Captured %d.%03d sec. %ld audio bytes (%d,%03d sps)" */
#define IDS_RT_VIDEO_LOST_ERROR				2005	/* "Capture stopped due to problem with source video - for example, fast forward" */ 
                                                    /* "or other video control pressed or video stopped. See Troubleshoot problems," */
													/* "Capturing video in on-line help for more information." */
#define IDS_RT_AUDIO_DROP_ERROR				2006    /* "Error: Audio data was lost during MPEG capture. To avoid this, reduce the audio */
													/* format (in KHz) on the Audio Capture page in  Capture Options."*/
#define IDS_RT_CAPTURE_LIMIT_2GB			2007	/* "You have captured the maximum file size allowed (2GB).  Capture has stopped." */
#define IDS_RT_CAPTURE_LIMIT_4GB			2008    /* "You have captured the maximum file size allowed (4GB).  Capture has stopped." */

/*
 * File size constants
 */
#define FILE_SIZE_SINGLEFILE	0
#define FILE_SIZE_NOFILE		0xffffffff

/*
 * define capture modes
 */
typedef enum {
	
    CAPMODE_IFRAME = 0,		// current mode is avi file capture
    CAPMODE_RTMPEG,			// current mode is mpg file capture
    CAPMODE_IFRAME_ONLY		// avi capture - only mode supported by hardware
}  
CAPTURE_MODE;


/*
 * define file types
 */
typedef enum {
	
    FILETYPE_AVI = 0,		// avi file 
    FILETYPE_MPG_ONLY,		// mpg file only
    FILETYPE_MPA_ONLY,		// mpa file only
	FILETYPE_MPG_MPA		// mpg and mpa
}  
FILE_TYPE;



/* 
 * define callbacks structure passed to
 * initialize private capture interface
 */
 typedef struct _CAPCALLBACKS {

    FARPROC lpfnErrorCallback;
    FARPROC lpfnStatusCallback;
    FARPROC lpfnYieldCallback;
    FARPROC lpfnFrameCallback;
    FARPROC lpfnVideoStreamCallback;
    FARPROC lpfnWaveStreamCallback;
    FARPROC lpfnCapControlCallback;

 } CAPCALLBACKS, *PCAPCALLBACKS, FAR *LPCAPCALLBACKS;


/* 
 * this structure defines mpeg capture/compression preferences. 
 */
typedef struct _BW_RECODE_PREFERENCES 
{
	/* size of this structure */
    unsigned long ulStructSize; // must be set before calling any related functions

	/* target */
	TCHAR EncodePlaybackTarget[30];	// playback target string

	/* video settings */
	long nEncodeVideoEnable;	// TRUE=video; FALSE=no video
	long nEncodeVideoBitRate; 	// output video rate in bits/second
	long nEncodeSquarePixels;	// TRUE=square pixels, FALSE=standard SIF/QSIF
	long nEncodeSofteningLevel;	// softening level - not currently used
	long nLargeImage;        	// TRUE= SIF; FALSE=QSIF
	
	/* audio settings */
	long nEncodeAudioEnable;	// TRUE=video; FALSE=no video
	long nEncodeAudioMode;		// MONO,STEREO,JOINTSTEREO or DUALCHANNEL
	long nEncodeAudioBitRate;	// output video rate in bits/second

    /* estimated disk space */
	long nCompressionEstimate;	// estimated file size in bits/second

} BW_RECODE_PREFERENCES;


/* 
 * this structure defines avi and shared mpeg capture preferences. 
 */
typedef struct _BW_CAPTURE_PREFERENCES
{
	/* size of this structure */
    unsigned long ulStructSize; // must be set before calling any related functions

	/* sorce settings */
    long nAutoDetect;			// TRUE=use source autodetect, FALSE=use specified source
    long nNTSC;					// TRUE=NTSC; FALSE=PAL
    long nComposite;			// TRUE=composite input; FALSE=svideo input

	/* audio settings */
	long nAudioEnable;			// TRUE=audio enabled; FALSE=AUDIO disabled
    long nAudioMode;			// MONO or STEREO 
    long nAudioRate;			// 11250,22500 or 44100 kHz

	/* appearance settings */
	long nHue;					// hue color setting (0-100)
    long nSaturation;			// saturation color setting (0-100)
    long nIntensity;			// intensity color setting (0-100)
    long nContrast;				// contrast color setting (0-100)
	long nFilterValue;			// TRUE=use softening; FALSE=no softening

	/* video quality setting */
    long nBitRate;				// video quality setting (35-70 Mb/minute
    
	/* capture type */
	long nCaptureMode;			// I-Frame (AVI) of realtime IBP (MPG) or I-Frame Only

	long nBlankLines;           // blank lines at bottom of image

} BW_CAPTURE_PREFERENCES;



#ifdef __cplusplus
extern "C" {
#endif


/********************************************************************
 *
 *
 *  The following inline functions provide a usefull means for
 *	making sure that the application and the configuration dialogs
 *  are consistant in interpreting the values in the preference
 *	structures.
 *
 *
 *******************************************************************/

/*
 * Translate the audio preferences to one of the predefined 
 * audio modes. 
 *
 * In the preference structure AudioEnable 
 * refers to whether audio was selected or not, and audio mode
 * is the last selected mode (mono, stereo, ect). Even if audio 
 * is disabled, this is required by the dialog so at least one
 * of the radio buttons will be checked, though disabled. Use
 * this function to get a valid audio mode without even having 
 * think about it
 *
 */
_inline UINT BW_AudioToMode (int nAudioEnable, int nAudioMode) 
{
	return (nAudioEnable ?nAudioMode:AUDIO_NONE);
}


/*
 * Translate a color preference to a valid hardware setting
 *
 * The color dialog display color setting from 0 to 100
 * The hardware requires values of -1000 to 1000. This 
 * function scales the dialog setting to the hardware setting. 
 * For use only by the driver - an application stores the dialog
 * setting from the preference structure.
 */
_inline int BW_ColorToHardware	(int nSetting) 
{
	return ((nSetting * 20) - 1000);
}


/*
 * Translate a color hardware setting to a valid color 
 * preference
 *
 * The color dialog display color setting from 0 to 100
 * The hardware requires values of -1000 to 1000. This 
 * function scales the hardware setting to the dialog setting. 
 * For use only by the driver - an application stores the dialog
 * setting from the preference structure.
 */
_inline int BW_HardwareToColor	(int nSetting) 
{
	return ((nSetting + 1000) / 20);
}


/* 
 * Translate encode bitrate to disk megabytes per minute
 *
 * given the nEncodeVideoBitRate, nEncodeAudioBitRate or 
 * nCompressionEstimate (which is a combination of both),
 * this function will translate the value to the same  
 * disk megabytes (1024*1024) per minute value as it is displayed
 * in the dialog. this is the same formula used by the dialog, 
 * so it's probably what an application wants to display as
 * a preference.
 */ 
_inline double BW_EncodeRateToMBm (int nRate) 
{
	return (((double)nRate/8.0*60.0) / (double)(1024*1024));
}


/* 
 * translate encode bitrate to disk kilobytes per minute
 *
 * given the nEncodeVideoBitRate, nEncodeAudioBitRate or 
 * nCompressionEstimate (which is a combination of both),
 * this function will translate the value to the a disk 
 * kilobyte (1024)/ minute value usefule for determining 
 * available disk for compression.
 *
 */
_inline double BW_EncodeRateToKBm (int nRate) 
{
	return (((double)nRate/8.0*60.0) / (double)(1024));
}


 
/*******************************************************************
 *
 *
 *  The following functions provide an application with a means of 
 *  initiating a "private interface" between an application and the 
 *  capture driver. This interface is reqired for realtime IBP mpeg 
 *  capture.
 *
 *
 *******************************************************************/

/*
 * Init Private Capture Interface
 *
 * Initializes the driver's private capture interface
 * by subclassing the VFW avicap capture window.  It 
 * will fail if the current proccess 
 * does not have the driver open for capture.   
 */
BOOL WINAPI BW_SetPrivateCaptureInterface (
    HWND hWndCap, 
    LPCAPCALLBACKS lpCapCalbacks
);


/*
 * Release Private Capture Interface
 *
 * Releases the driver's private capture interface.
 * Will fail if the current proccess does not have 
 * the driver open for capture.   
 */
BOOL WINAPI BW_ReleasePrivateCaptureInterface (HWND hWndCap);


/*******************************************************************
 *
 *
 *  The following functions provide an application with a means of 
 *  retrieving the current capture/compreesion settings so they can
 *	be saved to a preference file. Also provided are functions for 
 *  loading the settings retrieved from a preference file to the
 *  device registry/hardware.  Useing the functions in a read/modify/
 *  write method allows for settings of any individual setting.
 *
 *
 *******************************************************************/

/*
 * Load Capture Preferences
 *
 * Loads the capture settings in the given 
 * configuration structres to the device 
 * registry.
 */
BOOL WINAPI BW_LoadCapturePreferences (
    BW_CAPTURE_PREFERENCES FAR *lpCapturePreferences,
    BW_RECODE_PREFERENCES *lpRecodePreferences
);


/*
 * Get Capture Preferences
 *
 * Retrieves the current capture settings of the device
 * from the registry and stores them in the given 
 * configuration structures.
 */
BOOL WINAPI BW_GetCapturePreferences (
    BW_CAPTURE_PREFERENCES FAR *lpCapturePreferences,
    BW_RECODE_PREFERENCES FAR *lpRecodePreferences
);


/*
 * Load Recode Preferences
 *
 * Loads the recode settings in the given 
 * configuration structure to the device 
 * registry.
 */
BOOL WINAPI BW_LoadRecodePreferences (
    BW_RECODE_PREFERENCES FAR *lpRecodePreferences
);


/*
 * Get Recode Preferences
 *
 * Retrieves the current recode settings
 * from the device registry and stores
 * them in the given configuration structure. 
 */
BOOL WINAPI BW_GetRecodePreferences (
    BW_RECODE_PREFERENCES FAR *lpRecodePreferences
);


/*******************************************************************
 *
 *
 *  The following functions provide an application with a means of 
 *  retrieving current current specific capture settings, or to 
 *  display the capture options property sheet.  Any of these 
 *  functions may be called regadless of whether the driver is open
 *  for capture of not.
 *
 *
 *******************************************************************/

/*
 * Capture Configuration.
 *
 * This function displays the capure options property
 * sheet.
 */
BOOL WINAPI BW_ConfigCapture (HWND hWndParent);

/*
 * Set the # Blank Lines in overlay mode
 */

BOOL WINAPI BW_SetBlankLines(long BlankLines);

/*
 * Capture Mode
 *
 * This function reports back the last set Capture Mode 
 * capture setting. 
 */
BOOL WINAPI BW_CaptureMode (UINT *pMode);


/*
 * File Type
 *
 * This function reports back the current output 
 * file type(s) generated. 
 */
BOOL WINAPI BW_FileType (UINT *pType);


/* 
 * CalibrateAudio
 *
 * This function will perform audio/video synchronization calibration
 */
double BW_CalibrateAudio (HWND hWnd);


/*
 * Audio Mode                                                                  
 *
 * This function reports back the last set Audio 
 * capture mode.  
 */
BOOL WINAPI BW_AudioMode (UINT *pAudioMode);


/*
 * Audio Rate                                                                  
 *
 * This function reports back the last set Audio 
 * capture rate.  
 */
BOOL WINAPI BW_AudioRate (UINT *pAudioRate);


/*
 * Quality
 *
 * This function reports back the last set Quality 
 * capture Setting in Megabytes per minute.   
 */
BOOL WINAPI BW_Quality (UINT *pQuality);


/*
 * Real Time Quality
 *
 * This function reports back the last set real time
 * MPEG quality capture Setting in in bytes per second.   
 */
BOOL WINAPI BW_RealTimeQuality (DWORD *pQuality);


/*
 * Hardware Revision
 *
 * This function reports back the board hardware 
 * revision.  
 */
BOOL WINAPI BW_HardwareRevision (
    UINT *pHardwareRevision
);


/*
 * Hardware Caps 
 *
 * This function reports back the board hardware 
 * capiblities takeing into respect the type of broadway board.  
 */

BOOL WINAPI BW_HardwareCaps (DWORD *pHardwareCaps); 


/*
 * File size
 *
 * This function sets the maximum mpeg file size.
 * file size of 0 indicate no limit to file size.
 * This is a request from PictureTel. File size is
 * given in megabytes. 
 */
BOOL WINAPI BW_SetFileSize (DWORD dwSize);


/*
 * GetCapturePlayBackTarget
 *
 * This function reports back the current
 * capture play back target
 */
BOOL WINAPI BW_GetCapturePlayBackTarget (LPTSTR lpszTarget, UINT cbSize);


/*
 * SetCapturePlayBackTarget
 *
 * This function sets the current
 * capture play back target and updates
 * the registry seting to reflect the last set
 * options for that target
 */
BOOL WINAPI BW_SetCapturePlayBackTarget (LPTSTR lpszTarget);


/*
 * GetCapturePlayBackType
 *
 * This function reports back the current
 * capture play back type
 */
BOOL WINAPI BW_GetCapturePlayBackType (LPTSTR lpszType, UINT cbSize);


/*
 * GetCapturePlayBackType
 *
 * This function sets the current
 * capture play back type
 */
BOOL WINAPI BW_SetCapturePlayBackType (LPTSTR lpszType);


/*******************************************************************
 *
 *
 *  The following functions provide an application with a means of 
 *  retrieving current current specific compression (recode) 
 *  settings, to compression options dialog. Any of these 
 *  functions may be called regadless of whether the driver is open
 *  for compression of not.
 *
 *
 *******************************************************************/

/*
 * Recompress Configuration.
 *
 * This function performs recompression 
 * setup affecting the target format.
 */
BOOL WINAPI BW_ConfigRecompress (HWND hWndParent);


/*
 * CompressEstimate
 *
 * This function reports back the total stream data 
 * rate in bytes/second
 */
BOOL WINAPI BW_CompressEstimate (DWORD *pEstimate);


/*
 * GetCompressPlayBackTarget
 *
 * This function reports back the current
 * compression play back target
 */
BOOL WINAPI BW_GetCompressPlayBackTarget (LPTSTR lpszTarget, UINT cbSize);


/*
 * SetCompressPlayBackTarget
 *
 * This function sets the current
 * compression play back target and updates
 * the registry seting to reflect the last set
 * options for that target
 */
BOOL WINAPI BW_SetCompressPlayBackTarget (LPTSTR lpszTarget);



/*******************************************************************
 *
 *
 *  The following functions provide an application with a means of 
 *  performing file based compresion of avi files to output mpeg 
 *  files.
 *
 *
 *******************************************************************/

/*
 * RecompressBegin
 *
 * This function is a wrapper around BW_Recompress to handl parameter
 * validation and error reporting.
 *
 * This function will fail if there are no idle boards 
 * in the system.
 *
 */
BOOL WINAPI BW_RecompressBegin (
    LPTSTR pAviFileName, 
    LPTSTR pMpegFileName, 
    HWND hWndParent, 
    int nAudioDelay
);


/*
 * Recompression
 *
 * This function will recompress the input file 
 * The resulting mpeg system stream is written to 
 * the output MpegFile. The function will not return 
 * until the recompression is complete or an error is 
 * encountered.  
 *
 * This function will fail if ther are no idle boards 
 * in the system.
 */
DWORD WINAPI BW_Recompress (
    LPTSTR pAviFileName, 
    LPTSTR  pMpegFileName, 
    HWND hWndParent, 
    int nAudioDelay
);


/*
 * RecompressStreamsBegin
 *
 * This function is a wrapper around BW_RecompressStream to handle parameter
 * validation and error reporting.
 *
 * This function will fail if there are no idle boards 
 * in the system.
 *
 */
BOOL WINAPI BW_RecompressStreamsBegin (
	PAVISTREAM pAudioStream,
	PAVISTREAM pVideoStream,
    LPTSTR  pMpegFileName, 
    HWND hWndParent, 
    int nAudioDelay
);

/*
 * Stream Recompression
 *
 * This function will recompress the input Avi streams to full mpeg. 
 * The resulting mpeg system stream is written to the output MpegFile.
 * The function will not return until the recompression is complete 
 * or an error is encountered.  
 *
 * This function will fail if ther are no idle boards in the system.
 *
 */
DWORD WINAPI BW_RecompressStreams (
	PAVISTREAM pAudioStream,
	PAVISTREAM pVideoStream,
    LPTSTR  pMpegFileName, 
    HWND hWndParent,
    int nAudioDelay
);

#ifdef __cplusplus
}
#endif


#endif //_VCUSER_H_
