/* recode.cpp
This file contains the implementation for the CIBPRecode class. This class is used to 
compress an I-frame mpeg avi file into a full mpeg system stream file.
*/

#define BUILD_RECODE_DLL
#include "recode.h"
#include "video.h"
#include "errors.h"
#include "audio\encode.h"
#include "stream\stream.h"
#include "stream\videoCD.h"
#include "resource.h"
#include "bwuser.h"
#include <commctrl.h>
#include <math.h>
#include "convaud.h"
#include <rcuser.h>

HANDLE g_hInstance;    //saves our module's instance handle

BOOL WINAPI DllMain(HANDLE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
/*
Stash our instance handle
*/
{
	if (ul_reason_for_call == DLL_PROCESS_ATTACH)
		g_hInstance = hInst;
	return TRUE;
}

	
//constructor, takes board and dma buffer info 
CIBPRecode::CIBPRecode ( 
                 DEVID DevID)		//handle to VT_access board object
                 : mRecodeVideo(DevID)
{
	InitCommonControls();
//	AVIFileInit();
	mDevID=DevID;
	mCompressThread = NULL;
	mAudioThread = NULL;
	mStreamThread = NULL;
	mStatus = VT_NO_ERROR;
	mOldCursor = NULL;
	mhMpgFile=NULL;
	mSecondsRemaining=0;
	mbStreamBased = FALSE;
	mTimeReset=FALSE;
	mSecondsAvg=FALSE;
} 
 

//destructor
CIBPRecode::~CIBPRecode ( void)
{
	KillThreads();
//	if (!r->mbStreamBased) 
//   	AVIFileExit();
}

void CIBPRecode::KillThreads(void)
/*
Terminate any active threads
*/
{
	//make sure all threads are stopped
	if (mCompressThread != NULL) {
		TerminateThread (mCompressThread, 0);
		CloseHandle(mCompressThread);
	}
	if (mAudioThread != NULL) {
		TerminateThread (mAudioThread, 0);
		CloseHandle(mAudioThread);
	}
	if (mStreamThread != NULL) {
		TerminateThread (mStreamThread, 0);
		CloseHandle(mStreamThread);
	}

	mCompressThread = NULL;
	mAudioThread = NULL;
	mStreamThread = NULL;
}


ECODE CIBPRecode::Compress(
	        //interface parameters
		    const char *AviFileName, //full path name of input avi file
		    const char *MpgFileName, //full path name of output mpg file
			HWND hWnd,				 //window handle of parrent window for progress dialog
	
		    //audio re-compression params
		    int AudioMode,          //none, mono, stereo, daul channel, joint stereo (see bwuser.h)
		    int AudioDataRate,		//output audio data rate in kbits/second (1000 bits/second)

		    //video re-compression params
		    int VideoDataRate,		//output video rate in bits/second
		    BOOL UseQSIF,			//TRUE to select QSIF resolution, FALSE to select SIF

		    //stream mix parameters
            EStreamMixType StreamMixType,  //type of mixing, normal, none, or VideoCD
		    double StreamDataRate,  //total stream data rate in bytes/second, 0 = use optimum rate
		    int PackSize,			//in bytes
		    int PacketsPerPack,		//number of data packets per pack
									// 0 => use 1 audio packet and 1 video packet in each pack
		    double AudioDelay)		//delay audio form video (in seconds, can be negative)
	
/*
This is the main interface function of this class, it copies the input
parameters to the object, and spawns a thread to perform the compression.

This funciton returns the appropriate error code, VT_NO_ERROR (0) if success
*/

{
    /* before we do anything - try to get the recode settings */
	mRecodeOptions.ulStructSize=sizeof(mRecodeOptions);
    mRecodeOptions.nMode = MODE_FETCH_SETTINGS;
    mRecodeOptions.hWndParent = NULL;

	mBroadwayPref.ulStructSize=sizeof(BW_RECODE_PREFERENCES);
	mRecodeOptions.lpPreferences=&mBroadwayPref;

    if (!BW_GetRecodeOptions(&mRecodeOptions)) {
        return VT_ERR_BOARD_NOT_FOUND;
    }

	mEnableVideo=mBroadwayPref.nEncodeVideoEnable;

	//save parameters
	if (!mbStreamBased)
		strncpy(mAviFileName, AviFileName, sizeof(mAviFileName)-1);

	strncpy(mMpgFileName, MpgFileName, sizeof(mMpgFileName)-1);
	mAudioDataRate = AudioDataRate;

	mEnableAudio = TRUE;
	switch (AudioMode) {
	case AUDIO_NONE:
		mEnableAudio = FALSE;
		break;
	case AUDIO_MONO:
		mAudioType = MONO;
		break;
	case AUDIO_JOINTSTEREO:
		mAudioType = JOINT_STEREO;
		break;
	case AUDIO_DUALCHANNEL:
		mAudioType = DUAL_CHANNEL;
		break;
	case AUDIO_STEREO:
		mAudioType = STEREO;
		break;
	}

	/* if stream based, make sure we have the streams we'll need!!! */
	if (mbStreamBased) {

		if (mEnableAudio && mAudioStream == NULL)
			if (!mEnableVideo)
				return VT_ERR_NO_AUDIO_CHANNELS;
			else
				mEnableAudio = FALSE;

		if (mEnableVideo && mVideoStream == NULL)
			if (!mEnableVideo)
				return VT_ERR_CANT_OPEN_VIDEO_STREAM;
			else
				mEnableVideo = FALSE;
	}

	mVideoDataRate = VideoDataRate;
	mUseQSIF = UseQSIF;
    mStreamMixType = StreamMixType;
	mStreamDataRate = StreamDataRate;
	mPackSize = PackSize;
	mPacketsPerPack = PacketsPerPack;
	mAudioDelay = AudioDelay;

	//init video compression progress
	mRecodeVideo.SetCompressProgress(0);

	//create thread to do the work
	mStatus = VT_NO_ERROR;
    DWORD ThreadID;
	if (StreamMixType == NoMix || !mEnableVideo) {
		//we are in no-mix mode, produce an audio and a video file
		mCompressThread = CreateThread(NULL, 0, CompressThread, this, 0, &ThreadID);
	}
	else {
		//otherwise, stream mix
		mCompressThread = CreateThread(NULL, 0, CompressThreadMixed, this, 0, &ThreadID);
	}

	if (mCompressThread == NULL)
        mStatus = VT_ERR_CANT_CREATE_COMPRESS_THREAD;
	else {
		//thread ok, show the progress dialog, pass object pointer to init dialog
		time(&mStartTime);
		DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_PROGRESS),
							hWnd, (DLGPROC)ProgressDialogProc,(LPARAM) this);
	}

	return mStatus;
}

ECODE CIBPRecode::Compress(
	        //interface parameters
			PAVISTREAM pAudioStream, //input audio stream
			PAVISTREAM pVideoStream,//input video stream
		    const char *MpgFileName, //full path name of output mpg file
			HWND hWnd,				 //window handle of parrent window for progress dialog
	
		    //audio re-compression params
		    int AudioMode,          //none, mono, stereo, daul channel, joint stereo (see bwuser.h)
		    int AudioDataRate,		//output audio data rate in kbits/second (1000 bits/second)

		    //video re-compression params
		    int VideoDataRate,		//output video rate in bits/second
		    BOOL UseQSIF,			//TRUE to select QSIF resolution, FALSE to select SIF

		    //stream mix parameters
            EStreamMixType StreamMixType,  //type of mixing, normal, none, or VideoCD
		    double StreamDataRate,  //total stream data rate in bytes/second, 0 = use optimum rate
		    int PackSize,			//in bytes
		    int PacketsPerPack,		//number of data packets per pack
									// 0 => use 1 audio packet and 1 video packet in each pack
		    double AudioDelay)		//delay audio form video (in seconds, can be negative)
	
/*
This is the stream based interface function of this class, it stores the streams and
calls the main interface function .

This funciton returns the appropriate error code, VT_NO_ERROR (0) if success
*/
	
{
    /* before we do anything - try to get the recode settings */
	mRecodeOptions.ulStructSize=sizeof(mRecodeOptions);
    mRecodeOptions.nMode = MODE_FETCH_SETTINGS;
    mRecodeOptions.hWndParent = NULL;

	mBroadwayPref.ulStructSize=sizeof(BW_RECODE_PREFERENCES);
	mRecodeOptions.lpPreferences=&mBroadwayPref;

    if (!BW_GetRecodeOptions(&mRecodeOptions)) {
        return VT_ERR_BOARD_NOT_FOUND;
    }

	mbStreamBased = TRUE;

	mEnableVideo=mBroadwayPref.nEncodeVideoEnable;
	if (mEnableVideo)
		mVideoStream = pVideoStream;
	else
		mVideoStream = NULL;

	mEnableAudio = TRUE;
	switch (AudioMode) {
	case AUDIO_NONE:
		mEnableAudio = FALSE;
		break;
	case AUDIO_MONO:
		mAudioType = MONO;
		break;
	case AUDIO_JOINTSTEREO:
		mAudioType = JOINT_STEREO;
		break;
	case AUDIO_DUALCHANNEL:
		mAudioType = DUAL_CHANNEL;
		break;
	case AUDIO_STEREO:
		mAudioType = STEREO;
		break;
	}

	if (mEnableAudio)
		mAudioStream = pAudioStream;
	else
		mAudioStream = NULL;

	ECODE Status = Compress((const char *)NULL
							,MpgFileName
							,hWnd
							,AudioMode
							,AudioDataRate
							,VideoDataRate
							,UseQSIF
							,StreamMixType
							,StreamDataRate
							,PackSize
							,PacketsPerPack
							,AudioDelay
							);

	return Status;
}

BOOL CALLBACK CIBPRecode::ProgressDialogProc(HWND hwndDlg, UINT uMsg,
												WPARAM wParam, LPARAM lParam)
/*
Dialog funciton for progress dialog box, dispatch messages to appropriate handler
routines.
*/
{
	if (uMsg == WM_INITDIALOG) {
		//if init dialog, store our object pointer in the window user data
		SetWindowLong(hwndDlg, GWL_USERDATA, lParam);
	}

	//fecth our object pointer
	CIBPRecode *r = (CIBPRecode *)GetWindowLong(hwndDlg, GWL_USERDATA);

	switch (uMsg) {
	case WM_INITDIALOG:
		return r->OnInitDialog(hwndDlg);
	case WM_COMMAND:
		if (LOWORD(wParam) == IDCANCEL) {
			r->OnCancel(hwndDlg);
			return TRUE;
		}
		break;
	case WM_TIMER:
		r->UpdateProgress(hwndDlg);
		return TRUE;
	}

	return FALSE;
}

BOOL CIBPRecode::OnInitDialog(HWND hwndDlg)
/*
This function inits the specifed progress dialog window, returns TRUE if
focus is not set.
*/
{
	//center the dialog on the parent window
	HWND parent = GetParent(hwndDlg);
	if (parent == NULL)
		parent = GetDesktopWindow();
	RECT ParentRect, OurRect;
	GetWindowRect(parent, &ParentRect);
	GetWindowRect(hwndDlg, &OurRect);
	MoveWindow(hwndDlg,
		 ParentRect.left + (ParentRect.right - ParentRect.left - OurRect.right + OurRect.left)/2,
		 ParentRect.top + (ParentRect.bottom - ParentRect.top - OurRect.bottom + OurRect.top)/2,
		 OurRect.right - OurRect.left,
		 OurRect.bottom - OurRect.top,
		 FALSE);
	
	//show output file name

	if (!mEnableVideo) {
		char AudioFileName[MAX_PATH];
		strncpy(AudioFileName, mMpgFileName,sizeof(AudioFileName)-1);
		AudioFileName[sizeof(AudioFileName)-5] = '\0'; //make sure we have enough space for ext
		char *ptr = AudioFileName;
		while (*ptr != '.' && *ptr != '\0')				 //find extention
			ptr++;
		strcpy(ptr,".mpa");

		SetWindowText(GetDlgItem(hwndDlg, IDC_FILENAME), AudioFileName);
	} else {
		SetWindowText(GetDlgItem(hwndDlg, IDC_FILENAME), mMpgFileName);
	}

	//start timer
	SetTimer(hwndDlg, 1, 1000, NULL);  //get timer every 500 milliseconds

	return TRUE;
}

void CIBPRecode::UpdateProgress(HWND hwndDlg)
/*
update the progress bar based on the current compleation status
*/
{
	//fetch progress / time remaining data

	int PercentDone ;
	time_t CurrentTime;
	time(&CurrentTime);
	double TimeUsed=difftime(CurrentTime,mStartTime);
	int FramesProc=mRecodeVideo.GetFramesProcessed();
	char str[100];
	char label[100];
	
	if (mEnableVideo)
	{
		if (!mTimeReset)
		{
			time(&mStartTime);  // reset start time after one frame is processed
			if  (FramesProc>1)
				mTimeReset=TRUE;
			TimeUsed=0;
			mSecondsRemaining = mRecodeVideo.GetSecondsRemaining();

		}
		PercentDone = (int)mRecodeVideo.GetCompressProgress();
		TimeUsed+=1;
		DOUBLE seconds=TimeUsed*(100.0/mRecodeVideo.GetCompressProgress()-1);
#ifdef _DEBUG
		wsprintf(str,"Calc Sec %d ",(int)seconds);
		OutputDebugString(str);
#endif
		if (!mSecondsAvg)
		{
			if (TimeUsed<=10)
			{
				seconds = seconds*TimeUsed/10+mRecodeVideo.GetSecondsRemaining()*(10-TimeUsed)/10;
			}
			else
				mSecondsAvg=TRUE;
		}
		if (abs((int)seconds-mSecondsRemaining)>10)
			mSecondsRemaining = (int)seconds;
		else
		{
			mSecondsRemaining = mSecondsRemaining*3+(int)seconds;
			mSecondsRemaining>>=2; // now useing 25% averageing
		}
#ifdef _DEBUG
		sprintf(str,"TimeUsed %d TimeLeft %d Frame %d Percent %d Video Time %d\n",
			(int)TimeUsed,mSecondsRemaining, FramesProc,PercentDone,mRecodeVideo.GetSecondsRemaining());
		OutputDebugString(str);
#endif
		
	}
	else
	{
		PercentDone = mAudioFractionDone/655; // 100/0x10000;
		if (PercentDone>0)
		{
			mSecondsRemaining += (int)(TimeUsed*0x10000/mAudioFractionDone-TimeUsed);
			mSecondsRemaining>>=1;
		}
		else
			mSecondsRemaining=0;
	}
    if (mSecondsRemaining >= 60)	{
		wsprintf(str, "%d:%02d", mSecondsRemaining/60, mSecondsRemaining - 60*(mSecondsRemaining/60));
		LoadString(g_hInstance, IDS_MIN_SEC_LABEL, label, sizeof(label));
	}
	else {
		wsprintf(str, "%d", mSecondsRemaining);
		LoadString(g_hInstance, IDS_SEC_LABEL, label, sizeof(label));
	}

	//update progress bar & time remaining data
	if (PercentDone > 0 || FramesProc>10) {
		SendDlgItemMessage(hwndDlg, IDC_PROGRESS, PBM_SETPOS, PercentDone, 0);
		ShowWindow(GetDlgItem(hwndDlg, IDC_TIMEREMAINING_LABEL), SW_SHOW);
		SetWindowText(GetDlgItem(hwndDlg, IDC_TIMEREMAINING), str);
		SetWindowText(GetDlgItem(hwndDlg, IDC_TIME_LABEL), label);
	}

	//check if we are all done
	if (mCompressThread && WaitForSingleObject(mCompressThread,0) == WAIT_OBJECT_0) {
		CloseHandle(mCompressThread);
		mCompressThread = NULL;

		//reload codec microcode to board
		if (mOldCursor == NULL)
			mOldCursor=SetCursor(LoadCursor(NULL,MAKEINTRESOURCE(IDC_WAIT)));
//		VT_SetMode(mDevID,VT_IDLE);
//		VT_SetMode(mDevID,VT_CODEC);
		SetCursor(mOldCursor);
		mOldCursor = NULL;

		EndDialog(hwndDlg, 0);  //if we are done, it's time to exit
	}
}

void CIBPRecode::OnCancel(HWND hwndDlg)
/*
cancel the current compression: Set flag to abort compression
This will cause timer processing (UpdateProgress) to exit.
*/
{
	mOldCursor=SetCursor(LoadCursor(NULL,MAKEINTRESOURCE(IDC_WAIT)));
	mAbort = TRUE;
	if (WaitForSingleObject(mCompressThread,5000) != WAIT_OBJECT_0) {
		//if the compression is hung for some reason, force it
		KillThreads();
		if (mhMpgFile!=NULL) CloseHandle(mhMpgFile);

//		VT_SetMode(mDevID,VT_IDLE);/
//		VT_SetMode(mDevID,VT_CODEC);
		SetCursor(mOldCursor);
		mOldCursor = NULL;
		EndDialog(hwndDlg, 0);
	}

}

DWORD WINAPI CIBPRecode::CompressThread(LPVOID ThrdInfo)
/*
This is the main funciton of the worker thread for the CIBPRecode class.  
This funciton returns the appropriate error code, VT_NO_ERROR (0) if success
This function outputs data to a separate mpeg audio and mpeg video file.
*/
{
	CIBPRecode *r = (CIBPRecode *)ThrdInfo;  //get pointer to our object

	r->mStatus = VT_NO_ERROR;
	r->mAbort = FALSE;

    //open input avi file 
	if (!r->mbStreamBased) {

		AVIFileInit();

		if (0 != AVIFileOpen(&r->mAviFile, r->mAviFileName, OF_SHARE_DENY_WRITE, NULL)) {
			r->mStatus = VT_ERR_CANT_OPEN_AVI_FILE; 
	   		AVIFileExit();
			return r->mStatus;
		}

		//open audio stream if enabled
		if (r->mEnableAudio) {
			
			if (0 != AVIFileGetStream(r->mAviFile, &r->mAudioStream, streamtypeAUDIO, 0)) {

				r->mAudioStream = NULL;
				if (!r->mEnableVideo)  // if video is not enabled punt
				{
					AVIFileRelease(r->mAviFile);
	   				AVIFileExit();
					r->mStatus=VT_ERR_NO_AUDIO_CHANNELS;
					return VT_ERR_NO_AUDIO_CHANNELS;
				}
			}
		}
	}

    //do we need to encode audio?
    if (!r->mEnableAudio || r->mAudioStream == NULL) {
        //if no-audio, our job is easy
		if (r->mbStreamBased) {
	        r->mStatus = r->mRecodeVideo.CompressVideoStream( r->mVideoStream, r->mMpgFileName, &r->mAbort,
		                                                         r->mVideoDataRate,r->mUseQSIF);
		} else {
	        r->mStatus = r->mRecodeVideo.CompressVideoStream( r->mAviFile, r->mMpgFileName, &r->mAbort,
		                                                         r->mVideoDataRate,r->mUseQSIF);
		}

		if (!r->mbStreamBased) {
		    AVIFileRelease(r->mAviFile);
	   		AVIFileExit();
		}
        
		return r->mStatus;
    }

    //setup for audio processing
    if (r->mStatus == VT_NO_ERROR)
	    r->mStatus = r->SetupAudioEncoding(r->mAudioDataRate, r->mUseJointStereo);

	//set audio file name = base_name.mpa
    if (r->mStatus == VT_NO_ERROR) {
		strncpy(r->mAudioFileName, r->mMpgFileName,sizeof(r->mAudioFileName)-1);
		r->mAudioFileName[sizeof(r->mAudioFileName)-5] = '\0'; //make sure we have enough space for ext
		char *ptr = r->mAudioFileName;
		while (*ptr != '.' && *ptr != '\0')				 //find extention
			ptr++;
		strcpy(ptr,".mpa");
	}

	// Do we need to Encode Video? if not just compress audio and exit

    if (!r->mEnableVideo) {
    
		//if no-video, our job is easy
		r->AudioEncodeThread(r);
       	if (!r->mbStreamBased) {
			AVIFileRelease(r->mAviFile);
		   	AVIFileExit();
		}
        
		return r->mStatus;
    }

    //spawn thread to encode the audio
    DWORD ThreadID;
    if (r->mStatus == VT_NO_ERROR) {
        r->mAudioThread = CreateThread(NULL, 0, r->AudioEncodeThread, r, 0, &ThreadID);
	    if (r->mAudioThread == NULL)
            r->mStatus = VT_ERR_CANT_CREATE_AUDIO_THREAD;
    }

	//do all the video recode work in the CRecodeVideo class
    if (r->mStatus == VT_NO_ERROR) {

		if (r->mbStreamBased) {
	        r->mStatus = r->mRecodeVideo.CompressVideoStream( r->mVideoStream, r->mMpgFileName, &r->mAbort,
		                                                         r->mVideoDataRate,r->mUseQSIF);
		} else {
	        r->mStatus = r->mRecodeVideo.CompressVideoStream( r->mAviFile, r->mMpgFileName, &r->mAbort,
		                                                         r->mVideoDataRate,r->mUseQSIF);
		}

        if (r->mStatus != VT_NO_ERROR)
            TerminateThread (r->mAudioThread, VT_NO_ERROR);  //if video error, stop audio thread
    }
  
    //wait for audio to finish
    if (r->mStatus == VT_NO_ERROR)
        WaitForSingleObject(r->mAudioThread, INFINITE);
    
    //get audio/video status
    if (r->mStatus == VT_NO_ERROR)
        GetExitCodeThread(r->mAudioThread, &r->mStatus);

    //clean up
    CloseHandle(r->mAudioThread);
	r->mAudioThread = NULL;

	if (!r->mbStreamBased) {
	    AVIFileRelease(r->mAviFile);
   		AVIFileExit();
	}

    return r->mStatus;
}


DWORD WINAPI CIBPRecode::CompressThreadMixed(LPVOID ThrdInfo)
/*
This is the main funciton of the worker thread for the CIBPRecode class.  
This funciton returns the appropriate error code, VT_NO_ERROR (0) if success.
This function outputs data to a single system stream mpeg file.
*/
{
	CIBPRecode *r = (CIBPRecode *)ThrdInfo;  //get pointer to our object

	r->mStatus = VT_NO_ERROR;
	r->mAbort = FALSE;

    //open input avi file 
	if (!r->mbStreamBased) {
		AVIFileInit();

		if (0 != AVIFileOpen(&r->mAviFile, r->mAviFileName, OF_SHARE_DENY_WRITE, NULL)) {
			r->mStatus = VT_ERR_CANT_OPEN_AVI_FILE; 
	   		AVIFileExit();
			return r->mStatus;
		}

		//open audio stream if enabled
		if (r->mEnableAudio) {
			if (0 != AVIFileGetStream(r->mAviFile, &r->mAudioStream, streamtypeAUDIO, 0))
				r->mAudioStream = NULL;
		}
	}

    //do we need to encode audio?
    if (!r->mEnableAudio || r->mAudioStream == NULL) {
        //if no-audio, our job is easy
		if (r->mbStreamBased) {
	        r->mStatus = r->mRecodeVideo.CompressVideoStream( r->mVideoStream, r->mMpgFileName, &r->mAbort,
		                                                         r->mVideoDataRate,r->mUseQSIF);
		} else {
	        r->mStatus = r->mRecodeVideo.CompressVideoStream( r->mAviFile, r->mMpgFileName, &r->mAbort,
		                                                         r->mVideoDataRate,r->mUseQSIF);
		}

		if (!r->mbStreamBased) {
		    AVIFileRelease(r->mAviFile);
	   		AVIFileExit();
		}
        return r->mStatus;
    }

    //open output file 
    r->mhMpgFile =  CreateFile(r->mMpgFileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
                                FILE_ATTRIBUTE_NORMAL, NULL);
    if (r->mhMpgFile == INVALID_HANDLE_VALUE) {
		r->mStatus = VT_ERR_CANT_OPEN_OUTPUT_FILE; 
    }

    //setup for audio processing
    if (r->mStatus == VT_NO_ERROR) {
		r->mAudioFileName[0] = '\0';	//don't send audio output to a file
	    r->mStatus = r->SetupAudioEncoding(r->mAudioDataRate, r->mUseJointStereo);
	}

	// open the stream mixer
    //create/init stream mix object
    CStreamMix *stream;
    if (r->mStreamMixType == VideoCDMix)
        stream = new CVideoCDMix;
    else
        stream = new CStreamMix;

    stream->Initialize ( r->mhMpgFile,
                         r->mPackSize,
                         r->mPacketsPerPack,
                         r->mStreamDataRate,
                         r->mAudioDelay,
                         3072 /*VBV padding*/,
                         0.0 /*video frame rate*/);

	//setup for data transfer    
    r->mVideoPin = stream->GetVideoPin();
    r->mAudioPin = stream->GetAudioPin();

    //spawn thread to encode the audio
    DWORD ThreadID;
    if (r->mStatus == VT_NO_ERROR) {
        r->mAudioThread = CreateThread(NULL, 0, r->AudioEncodeThread, r, 0, &ThreadID);
	    if (r->mAudioThread == NULL)
            r->mStatus = VT_ERR_CANT_CREATE_AUDIO_THREAD;
    }

	//do all the video recode work in the CRecodeVideo class
    if (r->mStatus == VT_NO_ERROR) {
		if (r->mbStreamBased) {
	        r->mStatus = r->mRecodeVideo.CompressVideoStream( r->mVideoStream, r->mMpgFileName, &r->mAbort,
		                                                         r->mVideoDataRate,r->mUseQSIF,r->mVideoPin);
		} else {
	        r->mStatus = r->mRecodeVideo.CompressVideoStream( r->mAviFile, r->mMpgFileName, &r->mAbort,
		                                                         r->mVideoDataRate,r->mUseQSIF,r->mVideoPin);
		}
        if (r->mStatus != VT_NO_ERROR)
            TerminateThread (r->mAudioThread, VT_NO_ERROR);  //if video error, stop audio thread
    }
  
    //wait for audio to finish
    if (r->mStatus == VT_NO_ERROR)
        WaitForSingleObject(r->mAudioThread, INFINITE);
    
    //get audio/video status
    if (r->mStatus == VT_NO_ERROR)
        GetExitCodeThread(r->mAudioThread, &r->mStatus);

    //flush all stream data to output file
    if (!r->mAbort)
		stream->FlushData();

    //clean up
    CloseHandle(r->mAudioThread);
	r->mAudioThread = NULL;
    CloseHandle(r->mhMpgFile);
	r->mhMpgFile=NULL;
    delete stream;

	if (!r->mbStreamBased) {
	    AVIFileRelease(r->mAviFile);
   		AVIFileExit();
	}

    return r->mStatus;
}


ECODE CIBPRecode::SetupAudioEncoding(int AudioDataRate, BOOL UseJointStereo)
/*
This funciton open's checks the audio stream of our avi file.  It also allocates
and inits the audio encoding class
*/
{
    //read and check audio stream info
    WAVEFORMATEX FmtData;
    long FmtSize = sizeof(FmtData);
    FmtData.nSamplesPerSec = 0;
    AVIStreamReadFormat(mAudioStream, 0, &FmtData, &FmtSize);

    ECODE status = VT_NO_ERROR;
	if (FmtData.wFormatTag != WAVE_FORMAT_PCM)
        status = VT_ERR_NOT_PCM_AUDIO;
	else if (FmtData.nChannels < 1)
        status = VT_ERR_NO_AUDIO_CHANNELS;
	else if (FmtData.nChannels > 2)
        status = VT_ERR_TOO_MANY_AUDIO_CHANNELS;

    if (status != VT_NO_ERROR) {
        if (!mbStreamBased) 
    		AVIStreamRelease(mAudioStream);
        return status;
    }

    //set stereo/mono
    mAudioEncode = new CAudioEncode(SF_44100_HZ, mAudioType, AudioDataRate);
    if (mAudioEncode == NULL) {
		if (!mbStreamBased) 
			AVIStreamRelease(mAudioStream);
        return VT_ERR_CANT_ALLOC_AUDIO_ENCODER;
    }

    return VT_NO_ERROR;
}


	
DWORD WINAPI CIBPRecode::AudioEncodeThread(LPVOID ThrdInfo)
/*
this is the main function of the thread that performs mpeg-1 layer II
audio compression.  The input parameter is a pointer to the CIBPRecode
object that we are compressing.
*/
{
    //get our object data
    CIBPRecode *recode = (CIBPRecode *)ThrdInfo;

	//init local data
	BYTE *OutData = NULL;
	FILE *OutFile = NULL;
	BOOL UseFiles = recode->mAudioFileName[0] != '\0';
	ECODE status = VT_NO_ERROR;
 	CConvertStream ConvertStream;
	SAMPLE *InData = NULL;
	long TotalLength=AVIStreamLength(recode->mAudioStream);
    //setup for file output if required
	if (UseFiles) {
		OutFile = fopen(recode->mAudioFileName,"wb");
		if (OutFile == NULL)
			status = VT_ERR_CANT_OPEN_OUTPUT_FILE;
		if (status == VT_NO_ERROR) {
			OutData = new BYTE[recode->mAudioEncode->GetMaxFrameSize()];
			if (OutData == NULL)
				status = VT_ERR_MEMORY_ERROR;
		}
	}

	//setup stream access
	if (status == VT_NO_ERROR) {
		if (!ConvertStream.init(recode->mAudioStream, 44100.0))
			status = VT_ERR_CANT_OPEN_AUDIO_STREAM;
	}

	//alloc data
	if (status == VT_NO_ERROR) {
	    InData = new SAMPLE[INPUTSIZE+1];
		if (InData == NULL)
			status = VT_ERR_MEMORY_ERROR;
	}

	//do the mpeg encode thing
	int SamplesProcessed=0;
	if (status == VT_NO_ERROR) {
		long SamplesRead = ConvertStream.read(InData, INPUTSIZE);

		while ( SamplesRead > 0 && !recode->mAbort) {
			//zero pad input if needed
			SamplesProcessed+=SamplesRead;
			{
				CAutoLock Lock(&recode->mCritSec);
				recode->mAudioFractionDone=(int)((double)ConvertStream.GetSamplesRead()*0x10000/TotalLength);
			}
			while (SamplesRead < INPUTSIZE) {
				InData[SamplesRead][0] = 0.0F;
				InData[SamplesRead][1] = 0.0F;
				SamplesRead++;
			}
			//encode buffer, and output data
			int OutSize = recode->mAudioEncode->GetNextFrameSize();
			if (UseFiles) {
				//send data to a file
				int nMaxSB;
				recode->mAudioEncode->ProcessFrame(InData, OutData, FALSE, &nMaxSB);
				fwrite(OutData, sizeof(BYTE), OutSize, OutFile);
			}
			else {
				//send data to the stream mixer
				SBuff *buff=recode->mAudioPin->GetBuffer(1000);

				while (buff==NULL)
				{
					if (recode->mVideoPin->IsInputFinished() || recode->mAbort)
						break; // if too much audio quit now
					buff=recode->mAudioPin->GetBuffer(1000);
				} 
				int nMaxSB;
				if (buff==NULL)
				{
					SamplesRead=0;
					break;
				}
				
				recode->mAudioEncode->ProcessFrame(InData, (BYTE*)buff->data, FALSE, &nMaxSB);
				buff->ValidBytes = (int)OutSize;
				recode->mAudioPin->PutBuffer(buff);
				
			}
			//fetch next buffer of input audio
			SamplesRead = ConvertStream.read(InData, INPUTSIZE);
		}
	}

    //clean up
    if (!recode->mbStreamBased) 
	   AVIStreamRelease(recode->mAudioStream);
	if (OutFile != NULL)
		fclose(OutFile);
	if (OutData != NULL)
		delete [] OutData;
	if (!UseFiles)
		recode->mAudioPin->InputFinished();  //tell pin that we have no more data
	if (InData != NULL)
	    delete [] InData;
	delete recode->mAudioEncode;

	return status;

}



DWORD WINAPI CIBPRecode::StreamMix(LPVOID ThrdInfo)
/*
This is the main function of the thread that builds an mpeg-1 system stream
from the output of the audio and video streams.  This version of the code uses
files to store the intemediate audio and video data (c:\audio.mpg & c:\video.mpg).
The input parameter is a pointer to the CIBPRecode
object that we are stream mixing.  This function returns the appropriate error code,
VT_NO_ERROR if successfull.
*/
{
    //get our object data
    CIBPRecode *recode = (CIBPRecode *)ThrdInfo;

    //open video file
	char *FileName = "c:\\Video.mpg";
    HANDLE VideoFile =  CreateFile(FileName, GENERIC_READ, FILE_SHARE_READ, NULL,
                                OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (VideoFile == INVALID_HANDLE_VALUE) {
        return VT_ERR_CANT_OPEN_VIDEO_MPG_FILE;
    }

    //open audio file
	FileName = "c:\\Audio.mpg";
    HANDLE AudioFile =  CreateFile(FileName, GENERIC_READ, FILE_SHARE_READ, NULL,
                                OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (AudioFile == INVALID_HANDLE_VALUE) {
        CloseHandle(VideoFile);
        return VT_ERR_CANT_OPEN_AUDIO_MPG_FILE;
    }

    //create/init stream mix object
    CStreamMix stream;
    stream.Initialize ( recode->mhMpgFile,
                         recode->mPackSize,
                         recode->mPacketsPerPack,
                         recode->mStreamDataRate,
                         recode->mAudioDelay,
                         3072 /*VBV padding*/,
                         0.0 /*video frame rate*/);

	//setup for data transfer    
    CBasePin * VideoPin = stream.GetVideoPin();
    CBasePin * AudioPin = stream.GetAudioPin();
	BOOL VideoReady = TRUE;
	BOOL AudioReady = TRUE;

    //feed data to the pins till there is no more
	while (VideoReady || AudioReady) {
		DWORD size;

		//feed video if ready
		if (VideoReady) {
			SBuff *buff = VideoPin->GetBuffer();
		    ReadFile(VideoFile, buff->data, buff->DataSize, &size, NULL);
			buff->ValidBytes = (int)size;
			VideoPin->PutBuffer(buff);
			if (size == 0) {
				//no more data, close file
				CloseHandle(VideoFile);
				VideoFile = NULL;
			}
		}

		//feed audio if ready
		if (AudioReady) {
			SBuff *buff = AudioPin->GetBuffer();
		    ReadFile(AudioFile, buff->data, buff->DataSize, &size, NULL);
			buff->ValidBytes = (int)size;
			AudioPin->PutBuffer(buff);
			if (size == 0) {
				//no more data, close file
				CloseHandle(AudioFile);
				AudioFile = NULL;
			}
		}

		VideoReady = VideoPin->BufferReady() && VideoFile != NULL;
		AudioReady = AudioPin->BufferReady() && AudioFile != NULL;
	} //end while


    //flush all stream data to output file
    stream.FlushData();

    return VT_NO_ERROR;
}

void ProcessError(int ErrorNum)
/*
called when an error is detected in stream code. Exits stream thread with error message
*/
{
    ExitThread(ErrorNum);
}

