/*video.cpp
This module contains the implementation of the CRecodeVideo class.  This class uses the
broadway hardware to re-compress the input data into IBP mpeg-1 video.  The compressed
data is written to the specified output file
*/

//
//  mpeg-1 video encode filter
//


//include our stuff
#include "video.h"
#include "errors.h"
#include "stream\basepin.h"

//standard lib includes
#include <memory.h>
#include <stdio.h>

#define MAX_INPUT_DWORDS (0x20000 / sizeof(DWORD)) // 128K bytes

#ifdef _DEBUG
static char* DebugMessage=__FILE__ ": ERROR %d returned from line %d\n";
static char DebugBuffer[120];
#define CHECK_ERROR(__func) {ECODE __ecode=__func; \
if (__ecode) {sprintf(DebugBuffer,DebugMessage,__ecode,__LINE__); \
OutputDebugString(DebugBuffer);return __ecode;} } 
#else
#define CHECK_ERROR(__func) {ECODE __ecode=__func; if (__ecode) return __ecode;} 
#endif


#define COMPRESS_TIME 3.5  //compres to IBP in 3 times real time

/* List of class IDs and creator functions for class factory */

//#define GREEN_FIX

ECODE CRecodeVideo::CompressVideoStream(PAVIFILE pAviFile, const char *OutFileName,	BOOL *AbortFlag,
                                                int DataRate/* = 1150000*/ , BOOL UseQSIF /*=FALSE*/,
												CBasePin *pVideoPin /* =NULL*/)
/*
This function uses the broadway hardware to compress the 1st video stream in the
specified AVI file.  The output mpeg-1 video stream is written to the specified file
*/
{
	mVideoPin=pVideoPin;
	//alloc data
    DWORD *InData = new DWORD[MAX_INPUT_DWORDS];
	VT_AquireMutex();
	CHECK_ERROR(InitCompress(pAviFile, OutFileName, DataRate,UseQSIF,InData));
	SetCompressProgress(0);


	//setup for progress calculations
    long TotalFrames = AVIStreamLength(mVideoStream);
	double FrameCompressTime = COMPRESS_TIME / mFrameRate;

    //do the mpeg encode thing
    long SamplesRead = 0;
    long BytesRead = 0;
    int InSize = MAX_INPUT_DWORDS * sizeof(DWORD); 
#ifdef GREEN_FIX
	BytesRead=mLastGoodFrameBytes;
	SamplesRead=1;
#else
	AVIStreamRead(mVideoStream, mFrameCount, 1, InData, InSize, &BytesRead, &SamplesRead); 
#endif
    //encode each frame in the avi stream
    while ( SamplesRead > 0 && !*AbortFlag) {
        //printf("frame %3d bytes %d\n", start, BytesRead);  //test code
		mFrameCount++;

        //check for dropped frame
        if (BytesRead == 0) {
            //if this is a dropped frame, send last good frame (it's still in the buffer)
            BytesRead = mLastGoodFrameBytes;
		}
		else {
			if ((InData[0] & 0xff000000)==0xff000000 
				 && (InData[0] & 0x00ffffff) <= (DWORD)BytesRead)
				BytesRead=InData[0] & 0x00ffffff;
			mLastGoodFrameBytes = BytesRead;
		}

        //encode it & write to file!
        CHECK_ERROR(ProcessInput(InData, BytesRead/sizeof(DWORD)));

        //update compression progress
		SetCompressProgress(100 * mFrameCount / TotalFrames);
		SetSecondsRemaining((int)((TotalFrames - mFrameCount) * FrameCompressTime));

		//read more data
        SamplesRead = 0;
	    AVIStreamRead(mVideoStream, mFrameCount, 1, InData, InSize, &BytesRead, &SamplesRead); 
    }	

    if (!*AbortFlag)
		FinishProcessing(InData);
	VT_ReleaseMutex();

    //clean up
    AVIStreamRelease(mVideoStream);
    CloseHandle(mOutFile);
    delete [] InData;

	return VT_NO_ERROR;


}



ECODE CRecodeVideo::InitCompress(PAVIFILE pAviFile, const char *OutFileName,
								 int DataRate, BOOL UseQSIF, DWORD *pBuffer)
/*
This function set's up the memper data for the next compression.  It should be called 
before each compression.
*/
{
    //set hardware for video re-encoding
   //CHECK_ERROR( VT_Reset(mDevID));
   //CHECK_ERROR( VT_SetupVideoMode(mDevID,3,1));


    //create output file
    if (!mVideoPin)
	{
		mOutFile = CreateFile(OutFileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
								FILE_ATTRIBUTE_NORMAL, NULL);
		if (mOutFile == NULL)
			return VT_ERR_CANT_OPEN_OUTPUT_FILE;
	}
	else 
		mOutFile = NULL;

	//open video stream
	if (0 != AVIFileGetStream(pAviFile, &mVideoStream, streamtypeVIDEO, 0)) {
        CloseHandle(mOutFile);
        return VT_ERR_CANT_OPEN_VIDEO_STREAM;
	}
    
    //check for I-frame input video
    AVISTREAMINFO StreamInfo;
    StreamInfo.fccHandler = 0;
    AVIStreamInfo(mVideoStream, &StreamInfo, sizeof(StreamInfo)); 
    if ( StreamInfo.fccHandler != mmioFOURCC('B','W','1','0')
		&&  StreamInfo.fccHandler != mmioFOURCC('b','w','1','0') // codec seems to give lower case
        && StreamInfo.fccHandler != mmioFOURCC('B','M','P','G') )
	{
		// init for bypass here 
		//mBypass=TRUE;
        CloseHandle(mOutFile);
        AVIStreamRelease(mVideoStream);
        return VT_ERR_BAD_INPUT_VIDEO_TYPE;
    }

	//read frame rate form AVI file
	mFrameRate = StreamInfo.dwRate / StreamInfo.dwScale;
	//for now we will set to known NTSC or PAL rates
	if (mFrameRate!=30.0)
		mFrameRate = (mFrameRate >= 27.5) ? 29.97 : 25.00;

	//read frame size form AVI file
    long ReqSize;
    AVIStreamReadFormat(mVideoStream, 0, NULL, &ReqSize);
    BITMAPINFO *VideoFormat = (BITMAPINFO *)malloc(ReqSize);
    if (VideoFormat == NULL) {
        CloseHandle(mOutFile);
        AVIStreamRelease(mVideoStream);
        return VT_ERR_CANT_OPEN_VIDEO_STREAM;
	}
    AVIStreamReadFormat(mVideoStream, 0, VideoFormat, &ReqSize);
	long FrameWidth = VideoFormat->bmiHeader.biWidth;
	long FrameHeight = VideoFormat->bmiHeader.biHeight;
	free(VideoFormat);

	//download microcode to broadway hardware
	RECODEINFO RecodeInfo;
	RecodeInfo.StructSize= sizeof(RECODEINFO);
	VT_GetRecodeInfo(mDevID,&RecodeInfo);
    RecodeInfo.FrameRate = (int)(mFrameRate*100);
    RecodeInfo.DataRate  = DataRate;
	RecodeInfo.FrameWidth = (short)FrameWidth;
	RecodeInfo.FrameHeight = (short)FrameHeight;
	if (UseQSIF)
	{
		if (FrameHeight==288)
			RecodeInfo.FrameType=FT_PAL_QSIF;
		else if (FrameWidth==320)
			RecodeInfo.FrameType=FT_SQPIXEL_QSIF;
		else
			RecodeInfo.FrameType=FT_NTSC_QSIF;			
	}
	else
	{
		if (FrameHeight==288)
			RecodeInfo.FrameType=FT_PAL;
		else if (FrameWidth==320)
			RecodeInfo.FrameType=FT_SQPIXEL;
		else
			RecodeInfo.FrameType=FT_NTSC;			
	}
	VT_SetRecodeInfo(mDevID,&RecodeInfo);
    ECODE status =  VT_SetMode(mDevID,VT_RECODE);
	if (status == VT_NO_ERROR)
		status = VT_Start(mDevID);
	if (status != VT_NO_ERROR) {
		//if error starting, clean-up and return
        CloseHandle(mOutFile);
        AVIStreamRelease(mVideoStream);
        return status;
	}

	DWORD DMAWords;
	// These next two lines are a cludge to get the Buffer info from VT_access
	VT_ReadData(mDevID,&mDMALinearAddr, &DMAWords);
	VT_GetCurrentTransferAddress(mDevID,2,&mDMAPhysAddr);

    mOutputBuffer.Initialize((CBT*)mDMALinearAddr, DMAWords);
    //init state flags

	mLastGoodFrameBytes = 0;  //flag no good frames yet
    mFirstWrite = TRUE;
	mFrameCount = 0;

#ifdef GREEN_FIX
	char FileName[MAX_PATH];
	GetSystemDirectory(FileName,sizeof(FileName));
    lstrcat (FileName, "\\BWTABLES\\");
	switch (RecodeInfo.FrameType)
	{
		case FT_PAL:
		case FT_PAL_QSIF:
			lstrcat(FileName,"blpal.mpg");
			break;
		case FT_NTSC:
		case FT_NTSC_QSIF:
			lstrcat(FileName,"blntsc.mpg");
			break;
		default:
			return VT_NO_ERROR;
	}
	// now open the file and read it in
	HANDLE tFile = CreateFile(FileName, GENERIC_READ, 0, NULL, OPEN_EXISTING,
								FILE_ATTRIBUTE_NORMAL, NULL);
	if (tFile == NULL)
		return VT_NO_ERROR;
    if (!ReadFile(tFile,pBuffer,MAX_INPUT_DWORDS * sizeof(DWORD) 
		,(DWORD*)&mLastGoodFrameBytes,NULL))	
		return VT_NO_ERROR;
	CloseHandle(tFile);
#endif
    return VT_NO_ERROR;
}


ECODE CRecodeVideo::ProcessInput(DWORD *InputData, int InputSize)
/*
This function compresses the specifed input video data, and writes the results to
the object's output file. (
*/
{
    //nothing to do if buffer empty
    if (InputSize==0)
        return 0;

    //insure the frame starts with the start code
    //current codec and capture use start code to store frame size
    if (InputData[0]!=0 && InputData[0]!=0xb7010000) 
		InputData[0] = 0xb3010000;

    //read the current DMA address 
    DWORD addr;
	ECODE status=VT_GetCurrentTransferAddress(mDevID,2,&addr);
	if (status) return status;
    int loc = (addr-mDMAPhysAddr)/sizeof(CBT);

    //a bit of a kludge to chop off the 1st 16 DWORDS of zeros
    //so the Quartz mpeg player doesn't get messed-up, also
	//the 1st buffer must contain 1 complete video header
    if (mFirstWrite) {
        if (loc >= 1024) {
            mOutputBuffer.SetReadLocation(16);
            mFirstWrite = FALSE;
        }
    }
	else
    // remove the valid data from the DMA buffer
	{
   		mOutputBuffer.SetWriteLocation(loc);
		if (mVideoPin)
			mOutputBuffer.FlushToPin(mVideoPin); // do not flush 0 sized buffers
		else
			mOutputBuffer.FlushToFile(mOutFile);
	}

	// pad out the buffer
	DWORD size;
	if (InputSize<0x400)
		size=0x400-InputSize;
	else
		// INPUT size must be padded to blocks of 0x200 DWORDS
		size=(InputSize & 0xf) ? 0x10-(InputSize & 0xF):0;
	if (size)
	{
		ZeroMemory(&InputData[InputSize],size*sizeof(DWORD));
		InputSize+=size;
	}


    //send data to the board
    DWORD XferedCount;
	CHECK_ERROR(VT_WriteData(mDevID, InputData, InputSize, &XferedCount));
//	Sleep(10);
    //any remaining data to send, keep trucken 'tel it's gone
    size = InputSize-XferedCount;
	DWORD t1=timeGetTime();
    while (size )  {
        Sleep(5);  //don't use CPU while waiting
		CHECK_ERROR(VT_WriteData(mDevID, &InputData[XferedCount], size, &size));
        XferedCount += size;
        size = InputSize-XferedCount;
		if (timeGetTime()-t1 > 1000) 
			return VT_ERR_VIDEO_ENCODE_TIMEOUT;
        
    }
#ifdef _DEBUG
	char buffer[128];
	sprintf(buffer,"Processed %x size frame\n",InputSize);
	OutputDebugString(buffer);
#endif
	return 0;
}

#define STOP_CODE 0xb7010000

void CRecodeVideo::FinishProcessing(DWORD *pLastFrame)
/*
This function compresses any remaining data in the input buffer
*/
{
//    VT_Stop(mDevID);   //stop encoding, flush data from board  
	int count;
	int loops=0;
	DWORD temp;
	DWORD data[0x400];
    int size1,size2;
    CBT *ptr1, *ptr2;
    DWORD addr;
    ZeroMemory(data,sizeof(data));
	int size=mLastGoodFrameBytes/sizeof(DWORD);
	for (count=0;count<1;count++) ProcessInput(pLastFrame,size); // send one extra last frame
	while (VT_WriteData(mDevID, NULL, 0, &temp) ==0 ) ; // flush
	Sleep(500);         //wait for data to be output
	data[0]=0xb7010000; // stop code
	ProcessInput(data,0x400);
	while (VT_WriteData(mDevID, NULL, 0, &temp) ==0 ) ; // flush
	//	while (!ProcessInput(pLastFrame,size) && count++<3) // allow extra frames for padding
//		while (VT_WriteData(mDevID, NULL, 0, &temp ==0 )) ;
	//Sleep(2000);         //wait for data to be output
	// wait for stop code ..
	BOOL found=FALSE;
	while (!found)
	{
		Sleep(100);
		VT_GetCurrentTransferAddress(mDevID, 2, &addr);
		mOutputBuffer.SetWriteLocation((addr-mDMAPhysAddr)/sizeof(CBT));
		// now scan the buffer
		mOutputBuffer.GetDataPointers(&ptr1, &size1, &ptr2, &size2);
		for (count=0;count<size1;count++)
		{
			if (*ptr1++==STOP_CODE)
			{
				mOutputBuffer.SetWriteLocation(ptr1);
				found=TRUE;
				break;
			}
		}
		for (count=0;count<size2;count++)
		{
			if (*ptr2++==STOP_CODE)
			{
				mOutputBuffer.SetWriteLocation(ptr2);
				found=TRUE;
				break;
			}
		}
		if (loops++>40)
			break;
				
	}

	if (mVideoPin)
	{
		while (mOutputBuffer.ValidData())
		{
			mOutputBuffer.FlushToPin(mVideoPin);
		}
		mVideoPin->InputFinished();
	}
	else
		mOutputBuffer.FlushToFile(mOutFile);
	VT_Stop(mDevID);
	/*add end of sequence code to file
	DWORD eos = 0xb7010000;
    DWORD OutSize;
    WriteFile(mOutFile, &eos, sizeof(eos), &OutSize, NULL);
    */

	VT_SetMode(mDevID,VT_IDLE);
//	VT_Reset(mDevID);

}


#if 0

ECODE CRecodeVideo::LoadRecoder(BITMAPINFOHEADER *Bi,double FrameRate)
{
	mFrameRate=FrameRate;

	// first we fix up frame rate
	if (mFrameRate!=30.0 && mFrameRate >24.1)
		mFrameRate = (mFrameRate >= 27.5) ? 29.97 : 25.00;

	//read frame size form AVI file
    long ReqSize;
    AVIStreamReadFormat(mVideoStream, 0, NULL, &ReqSize);
    BITMAPINFO *VideoFormat = (BITMAPINFO *)malloc(ReqSize);
    if (VideoFormat == NULL) {
        CloseHandle(mOutFile);
        AVIStreamRelease(mVideoStream);
        return VT_ERR_CANT_OPEN_VIDEO_STREAM;
	}
    AVIStreamReadFormat(mVideoStream, 0, VideoFormat, &ReqSize);
	long FrameWidth = VideoFormat->bmiHeader.biWidth;
	long FrameHeight = VideoFormat->bmiHeader.biHeight;
	free(VideoFormat);

	//download microcode to broadway hardware
	RECODEINFO RecodeInfo;
	RecodeInfo.StructSize= sizeof(RECODEINFO);
	VT_GetRecodeInfo(mDevID,&RecodeInfo);
    RecodeInfo.FrameRate = (int)(mFrameRate*100);
    RecodeInfo.DataRate  = DataRate;
	RecodeInfo.FrameWidth = (short)FrameWidth;
	RecodeInfo.FrameHeight = (short)FrameHeight;
	if (UseQSIF)
	{
		if (FrameHeight==288)
			RecodeInfo.FrameType=FT_PAL_QSIF;
		else if (FrameWidth==320)
			RecodeInfo.FrameType=FT_SQPIXEL_QSIF;
		else
			RecodeInfo.FrameType=FT_NTSC_QSIF;			
	}
	else
	{
		if (FrameHeight==288)
			RecodeInfo.FrameType=FT_PAL;
		else if (FrameWidth==320)
			RecodeInfo.FrameType=FT_SQPIXEL;
		else
			RecodeInfo.FrameType=FT_NTSC;			
	}
	VT_SetRecodeInfo(mDevID,&RecodeInfo);
    ECODE status =  VT_SetMode(mDevID,VT_RECODE);
	if (status == VT_NO_ERROR)
		status = VT_Start(mDevID);
	if (status != VT_NO_ERROR) {
		//if error starting, clean-up and return
        CloseHandle(mOutFile);
        AVIStreamRelease(mVideoStream);
        return status;
	}

	DWORD DMAWords;
	// These next two lines are a cludge to get the Buffer info from VT_access
	VT_ReadData(mDevID,&mDMALinearAddr, &DMAWords);
	VT_GetCurrentTransferAddress(mDevID,2,&mDMAPhysAddr);

    mOutputBuffer.Initialize((CBT*)mDMALinearAddr, DMAWords);
    //init state flags

}
#endif