/*pins.h
interface for the CStreamMixAudioPin & CStreamMixVideoPin classes
*/

class CStreamMix;

class CStreamMixAudioPin : public CBasePin {
    protected:
    volatile int mFullQueueBytes;         //total valid bytes in full queue
    SBuff *mCurrentBuff;         //current (full) buffer being transfered
    int mBytesToCopy;            //bytes left to copy in current buffer
    char *mCurrentBuffPtr;       //pointer to remaining data on current buffer
    CStreamMix *mOwner;          //pointer to the owner class
    HANDLE mEmptyBuffSemaphore;

    BOOL mHeaderReady;           //TRUE if audio header data has been read
    double mFrameRate;           //rate of buffers in buffers/second (from stream header)
    double mDataRate;            //stream data rate in bytes/second
    double mAveFrameSize;        //in bytes
    double mBytesRemaining;      //number of bytes remaining in current frame
    int mFrameCount;             //number of complete frames xfered
    CCritSec mCritSec;           //critical section object
    BOOL mInputFinished;         //TRUE if all input is finsihed (no more ButBuffer()'s)


    virtual void ProcessData(void);  //called after each put buffer

    public:
	
	CStreamMixAudioPin(void) : CBasePin() {
        //constructor
        mEmptyBuffSemaphore = NULL;
    }

    ~CStreamMixAudioPin() {
        //destructor
        if (mBytesToCopy)
            FreeBuff(mCurrentBuff);
        if (mEmptyBuffSemaphore != NULL)
            CloseHandle(mEmptyBuffSemaphore);
    }

    void PutBuffer(SBuff * buff);
    
    SBuff *GetBuffer(int Timeoutms /*=0*/) {
        // fetch next empty buffer, block if no buffer available
        if (WaitForSingleObject(mEmptyBuffSemaphore, Timeoutms)==WAIT_TIMEOUT)
			return NULL;
		else
			return mEmpty.get();
    }

    int ValidData(void) {
        //return current number of valid bytes
        return mFullQueueBytes + mBytesToCopy;
    }

    void InputFinished(void);

    void SetOwner(CStreamMix *owner) {
        //set the class that owns us.
        //we will call our owners ProcessData() method, when we have data ready
        mOwner = owner;
    }

    BOOL initialize(int NumBuffs, int BuffSize) {
        //initialize member variables, semaphore and base class, return TRUE if ok
        mFullQueueBytes = mBytesToCopy = 0;
        mHeaderReady = FALSE;
        mInputFinished = FALSE;

		mFrameCount = 0;
        mAveFrameSize = 100.0;
        mBytesRemaining = 0.0;

        mEmptyBuffSemaphore = CreateSemaphore(NULL, NumBuffs, NumBuffs, NULL);
        if (mEmptyBuffSemaphore == NULL)
            return FALSE;
        return CBasePin::initialize(NumBuffs, BuffSize);
    }

	int RemainingFrameSize(void) {
		// return number of bytes left in the current audio frame (aduio access unit)
		return (int)mBytesRemaining; // may have rounding error
	}

	int BytesToCopy(void) {
		return mBytesToCopy;
	}

	int GetNextFrameNumber(void) {
		//return the number of the next full frame (audio access unit), first frame = 0
		return RemainingFrameSize() ? mFrameCount+1 : mFrameCount;
	}

    BOOL IsHeaderReady(void) { return mHeaderReady;};
    BOOL IsInputFinished(void) { return mInputFinished;};
    double GetFrameRate(void) { return mFrameRate;};
    double GetDataRate(void) { return mDataRate;};
    int MoveData(char *OutPtr, int OutSize);


    protected:
    //helper functions
    void ReadHeader(SBuff * buff);
};


//the following class is used to parse the input mpeg-1 video stream to retrive
//picture frame information
class CParseVideo {
	unsigned int mAcch;			  //bytes 7-4 of the saved stream accumulator
    unsigned int mAccl;           //bytes 3-0 of the saved stream accumulator
	int mBytesSinceLastFrame;
	int mFrameNumber;
	int mPresNumber;
	int mGroupPresFrame;		  //1st presentation frame number in current group of pictures
	int mMaxPresNumber;			  //max presentation number we have encountered
	int mLastFrameSize;
    double mVBVDelay;             //VBV delay in seconds (see 11172-2)
                                  //this is realy: VideoStreamTime (at end of frame)
                                  //                      - DecodeTime for the current frame
    static const int mPictureStartCode;  //start code bits
    static const int mGOPStartCode;  //group of pictures start code bits

	public:
	CParseVideo(void);
	BOOL FindNextPicture(SBuff * buff, int *position);
	int GetFrameNumber(void) {return mFrameNumber;}
	int GetPresNumber(void) {return mPresNumber;}
	int GetLastFrameSize(void) {return mLastFrameSize;}
    double GetVBVDelay(void) {return mVBVDelay;}
};


class CStreamMixVideoPin : public CStreamMixAudioPin {
    protected:
    struct SFrameRef {
        int FrameNumber; //frame number (1st frame in stream = 0)
        int PresNumber;  //frame number in presentation order
        int FrameSize;   //size of current frame in bytes
        double VBVDelay; //in seconds
    };

    CBuffQueue mFrameRefs;         //queue of current frame reference structures
    CBuffQueue mUnusedRefs;        //queue of unused frame reference structures

	SBuff *mInputFrameRefBuff;		 //buffer that contains frame ref for newest input data
    SFrameRef *mCurFrameRef;         //points to the frame ref of the output frame being xfered
	SBuff *mCurFrameRefBuff;		 //buffer that contains mCurFrameRef
	int mFrameBytesToXfer;			 //number of bytes remaining on the current frame
	volatile int mFrameBytesInQueue; //total bytes of frames that are 
									 //on the mFrameRefs queue
	CParseVideo ParseVideo;			 //helper class to parse input video 
	int mVBVSize;				     //size of VBV buffer in bytes


//	HANDLE mTestFile;				//file to recieve raw video data for test

    public:

    CStreamMixVideoPin(void) : CStreamMixAudioPin() {
        //constructor
		mCurFrameRefBuff = NULL;
		mInputFrameRefBuff = NULL;
    };

    ~CStreamMixVideoPin(void) {
        //distructor
        if (mCurFrameRefBuff != NULL)
            delete mCurFrameRefBuff;
        if (mInputFrameRefBuff != NULL)
            delete mInputFrameRefBuff;
    };

    BOOL initialize(int NumBuffs, int BuffSize);
    void PutBuffer(SBuff * buff);
    int MoveData(char *OutPtr, int OutSize);

    int ValidData(void) {
        //return current number of valid bytes
        return mFrameBytesInQueue;
    };

	int RemainingFrameSize(void) {
		// return number of bytes left in the current video frame
		return mFrameBytesToXfer;
	};

	int GetNextFrameNumber(void) {
		//return the number of the next full frame
        if (mCurFrameRefBuff == NULL)
            return -1;  //flag we are in start-up state
		if (mFrameBytesToXfer != mCurFrameRef->FrameSize
                                    || mCurFrameRef->FrameNumber == -1)
			return mCurFrameRef->FrameNumber+1;
		else
			return mCurFrameRef->FrameNumber;
	};

    //routines to fetch frame info for the next full frame
    //These routines should only be called if ValidData() >= RemainingFrameSize(void)
    int GetNextPresentationNumber(void) {return GetNextFrameRef()->PresNumber;}
    double GetNextVBVDelay(void) {return GetNextFrameRef()->VBVDelay;}
    int GetNextFrameSize(void) {return GetNextFrameRef()->FrameSize;}

    double GetAveFrameSize(int length);
    int GetVBVSize(void) {return mVBVSize;};
    void flush(void);

    //helper functions
    protected:
	BOOL ReadHeader(SBuff * buff);
    void UpdateFrameRef(SBuff *RefBuff);
	SBuff *GetEmptyFrameRef(void);
    SFrameRef *GetNextFrameRef(void);


};

