#ifndef __STREAM__
#define __STREAM__
/*stream.h
this file contains the interface of the CStreamMix and supporting classes.
This code takes an input mpeg-1 layer II audio stream and an mpeg-1 video
stream to produce an mpeg-1 system stream to the parameters specified in
in the CStreamMix::initialize() function.

Note that this code must be compiled with a 32 bit compiler.  It tracks time
by counting the total number of frames.  It can produce output files up to
1.8 years long without overflowing the time counts.  (how big is your disk?) 
*/
#include "basepin.h"
#include <windows.h>
#include <stdio.h>
#include "critsec.h"  //CCritSec & CAutoLock classes
#include "pins.h"

inline double MAX(double x, double y) {return (x<y) ? y : x;}
inline double MIN(double x, double y) {return (x>y) ? y : x;}
inline int MAX(int x, int y) {return (x<y) ? y : x;}
inline int MIN(int x, int y) {return (x>y) ? y : x;}


//header data structures and size definitions
struct SStamp {
    unsigned char   TypeByte;       //type, bits 32-30, and marker bit (= 1) 
    unsigned char   MidWord[2];     //bits 29-15 and marker bit (= 1)
    unsigned char   LowWord[2];     //bits 14-0  and marker bit (= 1)
};
#define TIME_STAMP_SIZE sizeof(SStamp)
#define VIDEO_STAMP_SIZE (2*TIME_STAMP_SIZE)

struct SPackHeader {
    char PackStartCode[4];
    char StreamTime[5];     //time code and marker bits
    char MuxRate[3];        //22 bits of data & 2 marker bits
};
#define PACK_HEADER_SIZE sizeof(SPackHeader)

struct SPacketHeader {
    char PacketStartCode[3];
    char StreamID;
    char PacketLength[2];
};

#define PACKET_HEADER_SIZE sizeof(SPacketHeader)
#define STD_BUFF_DATA_SIZE 2
#define END_CODE_SIZE 1

#define AUDIO_PACKET_HEADER_SIZE (PACKET_HEADER_SIZE + STD_BUFF_DATA_SIZE + TIME_STAMP_SIZE)
#define VIDEO_PACKET_HEADER_SIZE (PACKET_HEADER_SIZE + VIDEO_STAMP_SIZE)


class CStreamMix {

	//data structures
	protected:
	enum ETimeStampType {
		SINGLE_STAMP = 2,			//one single time stamp
		FIRST_STAMP  = 3,			//first stamp in a double time stamp
		SECOND_STAMP = 1			//second stamp in a double time stamp
	};

    enum EPacketType {
        PADDING_STREAM, AUDIO_STREAM, VIDEO_STREAM
    };

	//object data
	protected:
	CStreamMixAudioPin mAudioPin;
	CStreamMixVideoPin mVideoPin;
    CCritSec mCritSec;          //critical section object
	HANDLE mOutputFile;			//output file handle
	char *mPacketData;			//pointer to packet data buffer
	double mAudioStartTime;		//time in seconds to start audio presentation
	double mVideoStartTime;		//time in seconds to start video presentation
	double mAudioFrameRate;		//in frames (access units) per second
	double mVideoFrameRate;		//in frames (access units) per second
    int mVBVPadding;            //extra STD buffer padding to use (bytes)
    double mAudioDelay;         //delay in seconds for starting audio vs. starting video
	BOOL mStreamStarted;		//TRUE if we have started (StartUpStream() returns TRUE)

	
	private:
    double mVBVPaddingTime;     //extra STD buffer padding to use (seconds)
 	int mMinAudioSize;			//min amount of audio data required to build packet
	int mMinVideoSize;			//min amount of video data required to build packet
	int mPackSize;				//total size of pack data
	double mStreamRate;			//total ouput rate in bytes/second
    BOOL mJustInTimeAudio;      //TRUE if we are to send full audio packets just before they
                                //are needed.  Otherwise, we send 2 packets per pack
                                //(typicaly a large video packet and a small audio packet)
	double mStreamTime;			//time of current position in stream (in seconds)
	int mPacketsPerPack;		//number of packets in a pack


    BOOL mFixedRate;            //TRUE if fixed data rate, false if variable data rate
    double mPacketRate;         //in packets/second
    double mOverHeadRate;       //ave rate of header data (bytes/second)
	int mCurPacketNum;			//number of current packet, ranges form 0 to PacketsPerPack - 1
	int mBasePacketSize;		//packet size to use when adding bytes is not needed (in bytes)
    int mCurPacketSize;         //size of current packet in bytes
    int mUsedPackBytes;         //number of bytes already used in pack
    int mMaxPacketSize;         //largest possible packet (size of mPacketData buffer)
    int mCurVideoFrameSize;     //size of current video frame
	double mAudioTimeAhead;		//(in seconds) if decoders buffers are less then this far
								// ahead, it is time to send more audio
    double mMaxStreamRate;      //max stream rate used in bytes/second
	int mAvePacketSize;			//average packet size
    int mSkipAudioPacketCount;  //number of initial packs to skip audio.  not used in
                                //just-in-time audio mode
    double mAveAudioPacketSize; //required ave audio packet size, not used in just-in-time
                                //audio mode.
    double mAudioBytesUnder;    //runing total of the number of audio bytes under the target,
                                //not used in just-in-time audio mode.


	//interface functions
	public:
    CStreamMix(void)  {
        mAudioPin.SetOwner(this);
        mVideoPin.SetOwner(this);
        mPacketData = NULL;
    };
    virtual ~CStreamMix() {
        if (mPacketData != NULL)
            delete [] mPacketData;
    };
	virtual void ProcessData(void);
    CBasePin *GetAudioPin(void) {return &mAudioPin;};
    CBasePin *GetVideoPin(void) {return &mVideoPin;};
    virtual void FlushData(void);
    int Initialize( HANDLE OutputFile,    //win32 file handle
                    int PackSize,         //in bytes
                    int PacketsPerPack,   //if 0, we will use 2 packets per pack
                                          //the first will be an video packet, the 
                                          //second an audio packet.  If this is non-zero
                                          //we use the specified number of same
                                          //size packets in just-in-time audio mode.
                                          //in this case audio is sent just before the
                                          //audio buffers underflow.

                    double StreamRate,    //rate of output stream in bytes/second
                                          //(if set to zero, min required rate
                                          //will be used.  Ignored for non-fixed rate video)
                    double AudioDelay,    //delay audio form video presentation in seconds
                                          //(can be negative)
                    int    VBVPadding,    //extra buffer padding size in bytes
                    double VideoFrameRate); //in frames/second, 0.0 => use rate as specified
                                            //video stream


	//utility functions (used by derived classes)
	protected:
	int OutputPadding(int NumBytes);
	int AddVideoTimeStamp(char * OutPtr, BOOL pad=TRUE);
	void AddTimeStamp(char * OutPtr, ETimeStampType type, double time);
	void AddStuffing(char * OutPtr, int size, BOOL EndHeader =TRUE);
	BOOL InitRateData(double MinPaddingFraction);

	double VideoFrameToTime(int FrameNum) {
		return mVideoStartTime + FrameNum / mVideoFrameRate;
	}
	double AudioFrameToTime(int FrameNum) {
		return mAudioStartTime + FrameNum / mAudioFrameRate;
	}
	void BuildPacketHeader(char *HeaderData, EPacketType PacketType, int NumBytes);
    void LoadStdBuffSize(char *StdSizeData, EPacketType PacketType);
    virtual int GetVideoBufferSize(void);   //get req size of decoder's buffer
    virtual int GetAudioBufferSize(void);   //get req size of decoder's buffer

	//helper functions
	private:

	int NeedAudio(void);
	void OutputPackHeader(void);

    void ProcessPacket(double CurOutputRate);
	void BuildVideoPacket(BOOL DisableTimeStamp=FALSE);
    void AdjustRates(void);
	void BuildAudioPacket(void);
	BOOL StartUpStream(void);
	double GetOutputRate(void);
	int CalcPacketSize(void);

    void OutputSystemHeader(void);
    BOOL IsCSPS(void);
	int PackBufferSize(int BuffSuze,BOOL VideoScale=FALSE);
	int OutputFirstPack(void);

};



#endif //#ifndef __STREAM__
