/*
 * vtinit.c
 *
 *
 * Video Tool WinNT kernel-mode driver 
 * kernel-mode support library - initialization routines
 *
 * Mark Vrabel, June 96
 */

#include "vtkernel.h"

/* forward declaration of local functions */
NTSTATUS VT_CreateNumberedName(int DeviceNumber, LPWSTR pHeader, LPWSTR pBase, PUNICODE_STRING pUniqueName);
NTSTATUS VT_CreateLink(PUNICODE_STRING pusName);
VOID VT_MapPort(PDEVICE_EXT pDevExt, PHYSICAL_ADDRESS PortBase, ULONG NrOfPorts);
NTSTATUS VT_AllocFrameMemory (PDEVICE_EXT pDevExt, ULONG Size);
VOID VT_DeAllocFrameMemory(PDEVICE_EXT pDevExt);
NTSTATUS VT_AllocDevInfo (PDEVICE_EXT pDevExt);
VOID VT_DeAllocDevInfo (PDEVICE_EXT pDevExt);
BOOL VT_MapKerneltoUser(PVOID pKernAddr, ULONG ulSize, PVOID *ppUserAddr,PMDL *ppMdl);
VOID VT_UnMapKerneltoUser(PVOID *ppUserAddr,PMDL *ppMdl);
PWCHAR VT_SaveRegistryPathName(PUNICODE_STRING RegistryPathName);
BOOLEAN VT_FindPCIDeviceEx(USHORT VendorId, USHORT DeviceId, int index, PPCI_COMMON_CONFIG pPCICfg, ULONG CfgLen, ULONG *pBusNum, PPCI_SLOT_NUMBER pSlotNum, UCHAR *pRevisionID);
VOID VT_Cleanup(PDRIVER_OBJECT pDriverObject);



/*
 * write a dword value to this device's section of the registry or profile.
 * ValueName is a unicode string representing the registry value name or
 * profile key, and ValueData is the dword data written. Returns TRUE if
 * successfully written.
 */
BOOL
VT_WriteProfile(PDEVICE_EXT pDevExt, PWCHAR ValueName, DWORD ValueData)
{
    NTSTATUS Status;

    if ((pDevExt == NULL) || (pDevExt->ParametersKey == NULL)) {
		return(FALSE);
    }

    Status = RtlWriteRegistryValue(RTL_REGISTRY_ABSOLUTE
								  ,pDevExt->ParametersKey
								  ,ValueName
								  ,REG_DWORD
								  ,&ValueData
								  ,sizeof(DWORD)
								  );

    return(NT_SUCCESS(Status));
}


/*
 * write a unicode string value to this device's section of the registry 
 * or profile.ValueName is a unicode string representing the registry value 
 * name or profile key, and ValueData is the unicode string written. Value
 * size is the length (in bytes) of the unicode string. Returns TRUE if 
 * successfully written.
 */
BOOL
VT_WriteProfileString(
	PDEVICE_EXT pDevExt, 
	PWCHAR ValueName, 
	DWORD ValueData,
	DWORD ValueSize
)
{
    NTSTATUS Status;

    if ((pDevExt == NULL) || (pDevExt->ParametersKey == NULL)) {
		return(FALSE);
    }

    Status = RtlWriteRegistryValue(RTL_REGISTRY_ABSOLUTE
								  ,pDevExt->ParametersKey
								  ,ValueName
								  ,REG_SZ
								  ,&ValueData
								  ,ValueSize
								  );

    return(NT_SUCCESS(Status));
}


/*
 * create a possible device name by appending a given number to
 * the standard basename, and appending that to the string \device.
 * Puts the result in the counted string pName, which must be initialised
 * to point to sufficient storage.
 */
NTSTATUS
VT_CreateNumberedName(
    int DeviceNumber,               // number to append
    LPWSTR pHeader,			        // header for device (eg \Device)
    LPWSTR pBase,                   // base name - eg VidCap, WaveOut
    PUNICODE_STRING pUniqueName     // result goes here
)
{
    UNICODE_STRING UnicodeNumber;
    WCHAR Number[12];   // big enough for any 32-bit number

    /*
     * start the string off with \device
     * -assumes the string is correctly inited to 0 length.
     */
    if (pHeader != NULL) {
		RtlAppendUnicodeToString(pUniqueName, pHeader);
    }

    /*
     * add the base name (eg VidCap)
     */
    RtlAppendUnicodeToString(pUniqueName, pBase);

    /*
     * convert the number to a unicode string and append that
     */
    UnicodeNumber.Buffer = Number;
    UnicodeNumber.MaximumLength = sizeof(Number);

    RtlIntegerToUnicodeString(DeviceNumber, 10, &UnicodeNumber);
    RtlAppendUnicodeStringToString(pUniqueName, &UnicodeNumber);

    /* done!: we now have eg \Device\VidCap0 - its up to the caller
     * to establish if this is unique or not
     */
    return(STATUS_SUCCESS);
}


/*
 * given a name in \device, create a symbolic link into \dosdevices
 * for this name so that the device can be opened by Win32 applications.
 */
NTSTATUS
VT_CreateLink(PUNICODE_STRING pusName)
{
    UNICODE_STRING LinkObject;
    WCHAR LinkName[80];

    LinkName[0] = UNICODE_NULL;

    RtlInitUnicodeString(&LinkObject, LinkName);

    LinkObject.MaximumLength = sizeof(LinkName);

    RtlAppendUnicodeToString(&LinkObject, L"\\DosDevices");

    /*
	 * DeviceSize is a known compile time constant - use a define
     * rather than a variable as this reduces code size.
	 */
    #define  DeviceSize (sizeof(L"\\Device") - sizeof(UNICODE_NULL))

    pusName->Buffer += DeviceSize / sizeof(WCHAR);
    pusName->Length -= DeviceSize;

    RtlAppendUnicodeStringToString(&LinkObject, pusName);

    pusName->Buffer -= DeviceSize / sizeof(WCHAR);
    pusName->Length += DeviceSize;

    return (IoCreateSymbolicLink(&LinkObject, pusName));
}


/*
 * translate the bus-relative port address using the HAL,
 * and map the port address into memory if necessary
 */
VOID VT_MapPort(PDEVICE_EXT pDevExt, PHYSICAL_ADDRESS PortAddress, ULONG NrOfPorts)
{
    PHYSICAL_ADDRESS MappedAddress;

    pDevExt->PortMemType = 0;     // memory space
    HalTranslateBusAddress(pDevExt->BusType
						   ,pDevExt->BusNumber
						   ,PortAddress
						   ,&pDevExt->PortMemType
						   ,&MappedAddress
						   );

    if (pDevExt->PortMemType == 0) {
		/* memory mapped i/o - map the physical address into system space*/
		pDevExt->pRegs = MmMapIoSpace(MappedAddress, NrOfPorts, FALSE);

    } else {
		pDevExt->pRegs = (PUCHAR) MappedAddress.LowPart;
    }

    pDevExt->NrOfPorts = NrOfPorts;
}


/*
 * Allocates device frame buffer memory.
 */
NTSTATUS
VT_AllocFrameMemory (PDEVICE_EXT pDevExt, ULONG Size)
{
    PHYSICAL_ADDRESS MaxAddr;
    	
	MaxAddr.LowPart = 0xFFFFFFFF;
	MaxAddr.HighPart = 0xFFFFFFFF;

	pDevExt->dwMemoryBytes = Size;
    pDevExt->pLinearAddr=MmAllocateContiguousMemory(pDevExt->dwMemoryBytes, MaxAddr);
	  
    /* if nogo, fail */
	if (pDevExt->pLinearAddr == NULL) {                                                
	    KdPrint(("Unable to allocate frame buffer memory\n"));
		return STATUS_INSUFFICIENT_RESOURCES;
    }

    pDevExt->pPhysAddr = MmGetPhysicalAddress(pDevExt->pLinearAddr);
    return(STATUS_SUCCESS);
}


/*
 * Free previously allocated device frame 
 * buffer memory.
 */
VOID 
VT_DeAllocFrameMemory(PDEVICE_EXT pDevExt)
{
    if (pDevExt->pLinearAddr != NULL) {
		MmFreeContiguousMemory(pDevExt->pLinearAddr);
	}
   
	pDevExt->pLinearAddr = NULL;
}


/*
 * Allocate shared device information structure.
 */
NTSTATUS 
VT_AllocDevInfo (PDEVICE_EXT pDevExt)
{
    /* Allocate area for the reserved data */
	pDevExt->pUserInfo = ExAllocatePool (NonPagedPool, PAGE_SIZE);
    if (pDevExt->pUserInfo == NULL) {
	    KdPrint(("Unable to allocate user info\n"));
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	/* zero out the memory */
	RtlZeroMemory(pDevExt->pUserInfo, PAGE_SIZE);

	pDevExt->pUserInfo->dwStructSize   = PAGE_SIZE;
	pDevExt->pUserInfo->dwSizeReserved = PAGE_SIZE-FIELD_OFFSET(USER_INFO,dwSizeReserved);
	pDevExt->pUserInfo->State		   = IDLE;

   	return(STATUS_SUCCESS);
}


/*
 * VT_AllocDevInfo:
 *
 *   Frees shared device information structure.
 *
 * Arguments
 *   pDevExt - pointer to device extention data
 *
 * Return Value
 *
 *   None.
 */
VOID 
VT_DeAllocDevInfo (PDEVICE_EXT pDevExt)
{
    if (pDevExt->pUserInfo != NULL) {
		ExFreePool (pDevExt->pUserInfo);
		pDevExt->pUserInfo = NULL;
    }
}


/* 		
 * Map a user mode address to kernel mode
 */
BOOL 
VT_MapUsertoKernel(PVOID pUserAddr, ULONG ulSize, PVOID *ppKernAddr,PMDL *ppMdl) 
{
	*ppKernAddr  = NULL;

	__try {
		*ppMdl=MmCreateMdl(NULL,pUserAddr,ulSize);
		if (*ppMdl == NULL) {
			return FALSE;
		}
   
		MmProbeAndLockPages(*ppMdl,KernelMode,IoModifyAccess);
		*ppKernAddr=MmGetSystemAddressForMdl(*ppMdl);
   
	}
	__finally {
		return TRUE;
	}
}  


/* 
 * Unmaps a previously mapped user mode address
 */
VOID 
VT_UnMapUsertoKernel(PMDL pMdl) 
{

	if (pMdl == NULL) {
		return;
    } 
	
	__try {
		MmUnlockPages(pMdl);
		IoFreeMdl(pMdl);
	}
	__finally {
		return;
	}
}

/* 
 * Map a kernel mode address to process specific user mode
 * virtual address space
 */
BOOL
VT_MapKerneltoUser(PVOID pKernAddr, ULONG ulSize, PVOID *ppUserAddr,PMDL *ppMdl) 
{
	*ppUserAddr  = NULL;
	
	__try {
		*ppMdl = IoAllocateMdl(pKernAddr,ulSize,FALSE,TRUE,NULL);
		if (*ppMdl == NULL) {
		    return FALSE;
		}
     
		MmBuildMdlForNonPagedPool(*ppMdl);
		*ppUserAddr=MmMapLockedPages(*ppMdl,UserMode);
		KdPrint(("KmPtr 0x%x mapped to 0x%x\n",pKernAddr,*ppUserAddr)); 
	}
   
	__finally {
		return TRUE;
	}
}  


/* 
 * release a kernel mode to user mode address mapping
 */
VOID
VT_UnMapKerneltoUser(PVOID *ppUserAddr,PMDL *ppMdl) 
{
	
	if (ppMdl == NULL) {
		return;
	}
 
	if (*ppMdl == NULL) {
		return;
	}
	
	__try {
		
		if (ppUserAddr != NULL) {
			if (*ppUserAddr != NULL) {
				MmUnmapLockedPages(*ppUserAddr,*ppMdl);
			}
		}

		IoFreeMdl(*ppMdl);
		*ppUserAddr = NULL;
		*ppMdl = NULL;
	}
  
	__finally {
		return;
	}
}


/*
 * VT_MapDevInfoToUser:
 *
 *   Maps device infomation structure and it's
 *   associated pointers into user mode.
 *
 * Arguments
 *   pDevExt - pointer to device extention data
 *
 * Return Value
 *
 *   Final status from mapping.
 */
NTSTATUS 
VT_MapDevInfoToUser(PDEVICE_EXT pDevExt)
{
	PEPROCESS pProcess = IoGetCurrentProcess();
	PINSTANCE_INFO pInstance = NULL;
	int i;
	
	PVOID pUserRegs = NULL;
	PVOID pUserLinearAddr = NULL;
	PVOID pUserInfo = NULL;

	PMDL pMdlRegs = NULL;
	PMDL pMdlLinearAddr = NULL;
	PMDL pMdlInfo = NULL;
	
	/* 
	 * try to find an instance structure for this process. We use an array of
	 * instance info structures as opposed to a linked list so we don't have 
	 * to allocate new instance structures on the fly. Since each allocation
	 * from the non paged pool is on page boundries, we would have to grab 
	 * 4096 bytes from the pool for each instance info structure, and the 
	 * non paged pool is a very limited resource. The device extention contains
	 * an array of 256 instance info structures, enough for 256 seperate proccess
	 * spaces to gain access to the device information.  The likely hood of that 
	 * number being exceeded is nil.
	 */
	for  (i=0;i<MAX_INSTANCES;i++) {
		
		/* does one already exist??? */
		if (pDevExt->InstanceList[i].pProcess == pProcess) {
			pInstance = &pDevExt->InstanceList[i];
			break;
		}
		
		/* 
		 * keep track of the first unused structure so we can use it if 
		 * one does not already exist for this structure.
		 */
		if (pInstance == NULL && pDevExt->InstanceList[i].pProcess == NULL) {
			pInstance = &pDevExt->InstanceList[i];
		}
	}

	/* did we get one */
	if (pInstance == NULL) {
		KdPrint(("Unable to map device info to user mode\n"));
		return(STATUS_INSUFFICIENT_RESOURCES);
	} 
	
	/* 
	 * if we didn't have one for this process, we need to map the required 
	 * pointers to user mode.
	 */
	if (pInstance->pProcess == NULL) {

		/* map device register memory to Usermode */
		if (!VT_MapKerneltoUser(pDevExt->pRegs
	  							,pDevExt->NrOfPorts
						        ,&pUserRegs
						        ,&pMdlRegs
						        )) {

			goto exit_error;
		}

		/* map device frame buffer memory to Usermode */
		if (!VT_MapKerneltoUser(pDevExt->pLinearAddr
							   ,pDevExt->dwMemoryBytes
							   ,&pUserLinearAddr
							   ,&pMdlLinearAddr
	    					   )) {

			goto exit_error;
		}

		/* map device shared user memory to Usermode */
		if (!VT_MapKerneltoUser(pDevExt->pUserInfo
	  							,pDevExt->pUserInfo->dwStructSize
						        ,&pUserInfo
						        ,&pMdlInfo
						        )) {

			goto exit_error;
		}

		/* fill in the instance data */
		pInstance->pProcess			= pProcess;
		
		pInstance->pLinearAddr		= pUserLinearAddr;
		pInstance->pRegs			= pUserRegs;
		pInstance->pUserInfo		= pUserInfo;

		pInstance->pMdlLinearAddr	= pMdlLinearAddr;
		pInstance->pMdlRegs			= pMdlRegs;
		pInstance->pMdlInfo			= pMdlInfo;
	}

	/* increment the instance usage count */
	pInstance->UsageCount++;
    return(STATUS_SUCCESS);

exit_error:
    
	/* clean house and return with error */
	VT_UnMapKerneltoUser(&pUserRegs, &pMdlRegs);
	VT_UnMapKerneltoUser(&pUserLinearAddr, &pMdlLinearAddr);
	VT_UnMapKerneltoUser(&pUserInfo, &pMdlInfo); 
	
	KdPrint(("Unable to map device info to user mode\n"));
	return(STATUS_INSUFFICIENT_RESOURCES);
}


/*
 * VT_MapDevInfoToUser:
 *
 *   Maps device infomation structure and it's
 *   associated pointers into user mode.
 *
 * Arguments
 *   pDevExt - pointer to device extention data
 *
 * Return Value
 *
 *   Final status from mapping.
 */
NTSTATUS 
VT_GetDevInfoForUser(PDEVICE_EXT pDevExt, PDEVICE_INFO pDevInfo)
{
	PEPROCESS pProcess = IoGetCurrentProcess();
	PINSTANCE_INFO pInstance = NULL;
	int i;
	
	/* try to find an instance structure for this process */
	for  (i=0;i<MAX_INSTANCES;i++) {
		
		/* does one already exist??? */
		if (pDevExt->InstanceList[i].pProcess == pProcess) {
			pInstance = &pDevExt->InstanceList[i];
			break;
		}
	}

	/* did we get one */
	if (pInstance == NULL) {
		KdPrint(("Unable to map device info to user mode\n"));
		return(STATUS_INSUFFICIENT_RESOURCES);
	} 
	
	pDevInfo->dwStructSize	 = sizeof(DEVICE_INFO);
	pDevInfo->RevisionID	 = pDevExt->RevisionID;             
	
	pDevInfo->pRegs          = pInstance->pRegs;
	pDevInfo->pLinearAddr	 = pInstance->pLinearAddr;
	pDevInfo->pUserInfo		 = pInstance->pUserInfo;

    pDevInfo->pPhysAddr      = pDevExt->pPhysAddr.LowPart;
    pDevInfo->dwMemoryBytes  = pDevExt->dwMemoryBytes;

    pDevInfo->lpLinearAddr16 = NULL;
    pDevInfo->lpRegs16		 = NULL;		

    return(STATUS_SUCCESS);
}


/*
 * VT_UnMapDevInfo:
 *
 *   Frees device info mapping into user mode.
 *
 * Arguments
 *   pDevExt - pointer to device extention data
 *
 * Return Value
 *
 *   None.
 */
VOID 
VT_UnMapDevInfo(PDEVICE_EXT pDevExt) 
{
	PEPROCESS pProcess = IoGetCurrentProcess();
	PINSTANCE_INFO pInstance = NULL;
	int i;

	/* try to find the instance for this process */
	for  (i=0;i<MAX_INSTANCES;i++) {
		if (pDevExt->InstanceList[i].pProcess == pProcess) {
			pInstance = &pDevExt->InstanceList[i];
			break;
		}
	}

	/* did we get one??? */
	if (pInstance == NULL) {
		return;
	} 
	
	/* is this the last usage of this instance */
	if (--pInstance->UsageCount != 0) {
		return;
	} 

	/* unmap the user info */
	VT_UnMapKerneltoUser(&pInstance->pRegs, &pInstance->pMdlRegs);
	VT_UnMapKerneltoUser(&pInstance->pLinearAddr, &pInstance->pMdlLinearAddr);
	VT_UnMapKerneltoUser(&pInstance->pUserInfo, &pInstance->pMdlInfo);

	/* zero out the instance */
	RtlZeroMemory(pInstance, sizeof(INSTANCE_INFO));

}


/*
 * save the path to our registry entry (including parameters subkey)
 * in paged-pool memory.
 */
PWCHAR
VT_SaveRegistryPathName(PUNICODE_STRING RegistryPathName)
{
    int Length;
    PWCHAR SavedString;

    Length =
    RegistryPathName->Length + sizeof(PARMS_SUBKEY) +
         sizeof(UNICODE_NULL);                // Include backslash

    SavedString = ExAllocatePool(PagedPool, Length); // Only access on caller thread

    if (SavedString == NULL) {
	    KdPrint(("Unable to allocate parameters key\n"));
		return NULL;
    }

    /* Copy the character data */
    RtlCopyMemory(SavedString, RegistryPathName->Buffer,
        RegistryPathName->Length);

    SavedString[RegistryPathName->Length / sizeof(WCHAR)] = L'\\';
    SavedString[RegistryPathName->Length / sizeof(WCHAR) + 1] = UNICODE_NULL;

    /* Append the parameters suffix prepended by a backslash */
    wcscat(SavedString, PARMS_SUBKEY);

    return SavedString;
}

BOOLEAN
VT_PCIChipSetProbe (ULONG BusNum)
{
	PCI_COMMON_CONFIG PCICfg;
	ULONG CfgLen = sizeof(PCI_COMMON_CONFIG);
	PPCI_SLOT_NUMBER pSlotNum;
    ULONG SlotNum, Status, type;
	DWORD EnableShortBursts = 0;

	UNICODE_STRING Description;
    UNICODE_STRING UnicodeNumber;
	WCHAR DescriptionBuffer[256];
    WCHAR Number[12]; 
	

	/* set up the description string */
    Description.Buffer = DescriptionBuffer;
    Description.MaximumLength = sizeof(DescriptionBuffer);
    Description.Length = 0;

    RtlZeroMemory (Description.Buffer, Description.MaximumLength);

	/* go through all the slots and try to find the bridge */
	for (SlotNum = 0; SlotNum < MAX_PCI_SLOTS;SlotNum++) {
		
		/* get the configuation information */
		Status = HalGetBusData(PCIConfiguration,BusNum,SlotNum,&PCICfg,CfgLen);
		if (Status == BUS_NOTFOUND) {
		    KdPrint(("PCI bus not found\n"));
			return FALSE;
		}
		

		if (PCICfg.VendorID == 0xffff) {
			continue;
		}

		/* store device/vendor ID */
		type = MAKELONG (PCICfg.DeviceID, PCICfg.VendorID);
		
		/* see if it's a known type */
		switch (type) {
		
			case 0x80860486:
				RtlAppendUnicodeToString(&Description, L"Intel 82425EX PCI system controller");
				break;
		
			case 0x80860484:
				if (PCICfg.RevisionID == 3) {
				    RtlAppendUnicodeToString(&Description, L"Intel 82378ZB PCI-ISA Bridge");
				} else {
				    RtlAppendUnicodeToString(&Description, L"Intel 82378IB PCI-ISA Bridge");
				}
				break;
		
			case 0x80860483:
			    RtlAppendUnicodeToString(&Description, L"Intel 82424ZX (Saturn) cache DRAM controller");
				break;
		
			case 0x80860482:
			    RtlAppendUnicodeToString(&Description, L"Intel 82375EB PCI-EISA Bridge");
				break;

			case 0x808604a3:
				if (PCICfg.RevisionID == 16 || PCICfg.RevisionID == 17) {
				    RtlAppendUnicodeToString(&Description, L"Intel 82434NX (Neptune) PCI cache memory controller");
					EnableShortBursts = 1;
				} else {
				    RtlAppendUnicodeToString(&Description, L"Intel 82434LX (Mercury) PCI cache memory controller");
				}
				break;

			case 0x8086122d:
			    RtlAppendUnicodeToString(&Description, L"Intel 82437 (Triton) PCI cache memory controller");
				break;

			case 0x8086122e:
			    RtlAppendUnicodeToString(&Description, L"Intel 82371 (Triton) PCI-ISA Bridge");
				break;

			case 0x10390496:
			    RtlAppendUnicodeToString(&Description, L"SiS 85c496");
				break;
		
			case 0x10390406:
			    RtlAppendUnicodeToString(&Description, L"SiS 85c501");
				break;
		
			case 0x10390008:
			    RtlAppendUnicodeToString(&Description, L"SiS 85c503");
				break;
		
			case 0x10390601:
			    RtlAppendUnicodeToString(&Description, L"SiS 85c601");
				break;
		
			case 0x10140022:
			    RtlAppendUnicodeToString(&Description, L"IBM 82351 PCI-PCI Bridge");
				break;
		
			case 0x10110001:
			    RtlAppendUnicodeToString(&Description, L"DEC 21050 PCI-PCI Bridge");
				break;

			default:
				if (PCICfg.BaseClass == 0x06) {
				    RtlAppendUnicodeToString(&Description, L"PCI Standard Host CPU Bridge");
					break;
				}
		}

		/* if length is not zero, we found one */
		if (Description.Length != 0) {
			break;
		}
	}

	/* make sure we have a registry path to write to */
	RtlCreateRegistryKey(RTL_REGISTRY_ABSOLUTE, L"\\Registry\\Machine\\Software\\Data Translation");
	RtlCreateRegistryKey(RTL_REGISTRY_ABSOLUTE, L"\\Registry\\Machine\\Software\\Data Translation\\Broadway");
	RtlCreateRegistryKey(RTL_REGISTRY_ABSOLUTE, L"\\Registry\\Machine\\Software\\Data Translation\\Broadway\\Options");

	/* did we find the bridge??? */
	if (Description.Length != 0) {

		/* write the description to the registry */
	    Description.Buffer[(Description.Length / sizeof(WCHAR))+1] = UNICODE_NULL;
		Status = RtlWriteRegistryValue(RTL_REGISTRY_ABSOLUTE
			                           ,L"\\Registry\\Machine\\Software\\Data Translation\\Broadway\\Options"
									   ,L"PCIBridgeType"
									   ,REG_SZ
									   ,Description.Buffer
									   ,Description.Length
									   );

		/* create an ID sring */
		Description.Length = 0;
		RtlZeroMemory (Description.Buffer, Description.MaximumLength);
    
		RtlAppendUnicodeToString(&Description, L"VEN_");

		UnicodeNumber.Buffer = Number;
		UnicodeNumber.MaximumLength = sizeof(Number);

		RtlIntegerToUnicodeString(PCICfg.VendorID, 16, &UnicodeNumber);
		RtlAppendUnicodeStringToString(&Description, &UnicodeNumber);

		RtlAppendUnicodeToString(&Description, L"&DEV_");

		RtlIntegerToUnicodeString(PCICfg.DeviceID, 16, &UnicodeNumber);
		RtlAppendUnicodeStringToString(&Description, &UnicodeNumber);
	
		/* write the ID string to the registry */
	    Description.Buffer[(Description.Length / sizeof(WCHAR))+1] = UNICODE_NULL;
		Status = RtlWriteRegistryValue(RTL_REGISTRY_ABSOLUTE
			                           ,L"\\Registry\\Machine\\Software\\Data Translation\\Broadway\\Options"
									   ,L"PCIBridgeID"
									   ,REG_SZ
									   ,Description.Buffer
									   ,Description.Length
									   );
	} else {
		/* this should NEVER happen, but be prepared */
	    RtlAppendUnicodeToString(&Description, L"PCI Host CPU Bridge not found");

		/* write the description to the registry */
	    Description.Buffer[(Description.Length / sizeof(WCHAR))+1] = UNICODE_NULL;
		Status = RtlWriteRegistryValue(RTL_REGISTRY_ABSOLUTE
			                           ,L"\\Registry\\Machine\\Software\\Data Translation\\Broadway\\Options"
									   ,L"PCIBridgeType"
									   ,REG_SZ
									   ,Description.Buffer
									   ,Description.Length
									   );

		/* write the ID string to the registry */
		Status = RtlWriteRegistryValue(RTL_REGISTRY_ABSOLUTE
			                           ,L"\\Registry\\Machine\\Software\\Data Translation\\Broadway\\Options"
									   ,L"PCIBridgeID"
									   ,REG_SZ
									   ,Description.Buffer
									   ,Description.Length
									   );
	}

	/* write the short burst enable setting to the registry */
	Status = RtlWriteRegistryValue(RTL_REGISTRY_ABSOLUTE
		                           ,L"\\Registry\\Machine\\Software\\Data Translation\\Broadway\\Options"
								   ,L"EnableShortBursts"
								   ,REG_DWORD
								   ,&EnableShortBursts
								   ,sizeof(DWORD)
								   );
	
	return TRUE;
}


/* 
 * find a PCI device in the system and report it's bus and
 * slot number 
 */
BOOLEAN 
VT_FindPCIDeviceEx(
	USHORT VendorId,
	USHORT DeviceId, 
	int index, 
	PPCI_COMMON_CONFIG pPCICfg,
	ULONG CfgLen,
	ULONG *pBusNum,
	PPCI_SLOT_NUMBER pSlotNum,
    UCHAR *pRevisionID
) 
{
    ULONG BusNum,SlotNum;
    ULONG stat;


	/* index is zero based make it 1 based */
	index++; 
	KdPrint(("Looking for board #%d ...",index));
    
	for (BusNum = 0; BusNum < MAX_PCI_BUSES;BusNum++) {
		for (SlotNum = 0; SlotNum < MAX_PCI_SLOTS;SlotNum++) {
			stat=HalGetBusData(PCIConfiguration,BusNum,SlotNum,pPCICfg,CfgLen);
			if (stat == BUS_NOTFOUND) {
			    KdPrint(("PCI bus not found\n"));
				return FALSE;
			} else if ((pPCICfg->VendorID == VendorId) && (pPCICfg->DeviceID == DeviceId)) {
				if (--index == 0) {
					pSlotNum->u.AsULONG = SlotNum;
					*pBusNum  = BusNum;
					*pRevisionID = pPCICfg->RevisionID;
					KdPrint(("Found board in slot #%d bus #%d\n",SlotNum,BusNum));
					return TRUE;
				}
			}
		}
	}
	
	return FALSE;
}


/*
 * Create the device object, and any necessary related setup, and
 * allocate device extension data. The device extension data is
 * a DEVICE_EXT struct.
 */
PDEVICE_EXT
VT_Init(
    PDRIVER_OBJECT pDriverObject,
    PUNICODE_STRING RegistryPathName,
	USHORT DevId,
	USHORT VenId,
	ULONG FrameBufferSize
)
{
    NTSTATUS Status = IO_ERR_CONFIGURATION_ERROR;
    UNICODE_STRING usName;
	UNICODE_STRING DeviceRegName;
    WCHAR usNameBuffer[MAX_VIDCAP_NAME_LENGTH];
    PCM_RESOURCE_LIST pResourceList;
    PCM_PARTIAL_RESOURCE_LIST pPartialList;
    ULONG ResrcIdx;
    PHYSICAL_ADDRESS DeviceAddress;
    BOOLEAN Ok;
    PDEVICE_EXT pDevExt; 
    PDEVICE_OBJECT pDevObj;
    PCI_COMMON_CONFIG PCICfg;
    PCI_SLOT_NUMBER SlotNum;
    ULONG BusNum;
    UCHAR RevisionID;
	int DeviceNumber;


    /*
     * create the device object
     * - we need to create a unique name for this device, by trying
     * to create names with 0, then 1 etc appended until the name is unique.
     */
    for(DeviceNumber = 0; DeviceNumber < MAX_VIDCAP_DEVICES; DeviceNumber++) {

        usName.Buffer = usNameBuffer;
        usName.MaximumLength = sizeof(usNameBuffer);
        usName.Length = 0;

        VT_CreateNumberedName(DeviceNumber
							  ,L"\\Device\\"
							  ,DD_VIDCAP_DEVICE_NAME_U
							  ,&usName
							  );

		/* make sure the device is physically present */
		if (VT_FindPCIDeviceEx(VenId
		                       ,DevId
			    			   ,0
			    			   ,&PCICfg
				  			   ,sizeof(PCI_COMMON_CONFIG)
							   ,&BusNum
							   ,&SlotNum
							   ,&RevisionID
							   )) {
			

			Status = IoCreateDevice(pDriverObject
									,sizeof(DEVICE_EXT)
									,&usName
									,0
									,FILE_DEVICE_SOUND  // device type:multimedia == sound?
									,FALSE			    // not exclusive - more than one thread access
									,&pDevObj
									);

			if (NT_SUCCESS(Status)) {
				break;
			}
		}
    }

	/* fail if no devices were found */
    if (!NT_SUCCESS(Status)) {
	    KdPrint(("Cannot find device\n"));
		VT_Cleanup(pDriverObject);
		return(NULL);
    }

    /* get pointer to device extension and init data*/
    pDevExt = (PDEVICE_EXT) pDevObj->DeviceExtension;

	/* zero out the memory */
	RtlZeroMemory(pDevExt, sizeof(DEVICE_EXT));

    pDevExt->DeviceNumber = DeviceNumber;
    pDevExt->DeviceInUse = 0;
	
	pDevExt->BusType = PCIBus;
	pDevExt->BusNumber = BusNum;
	pDevExt->SlotNumber = SlotNum;

    /* ptr back to device object */
    pDevExt->pDeviceObject = pDevObj;

    /*
     * create a link into dosdevices for this same name so that win32
     * apps can open the device (they need to open eg \\.\VidCap0 )
     */
    if (!NT_SUCCESS(VT_CreateLink(&usName))) {
	    KdPrint(("Unable to creat symbolic link\n"));
		VT_Cleanup(pDriverObject);
		return(NULL);
    }

    /*
     * save the registry pathname in pDevExt, appending the
     * Parameters subkey
     */
    pDevExt->ParametersKey = VT_SaveRegistryPathName(RegistryPathName);
    if (pDevExt->ParametersKey == NULL) {
	    KdPrint(("Unable to save registry path name\n"));
		VT_Cleanup(pDriverObject);
		return(NULL);
    }

	/* 
	 * save the raw registry path as a unicode string
	 */
	pDevExt->RegistryPath.MaximumLength = RegistryPathName->MaximumLength;
	pDevExt->RegistryPath.Length = 0;
	pDevExt->RegistryPath.Buffer = ExAllocatePool (NonPagedPool, pDevExt->RegistryPath.MaximumLength);

	Status = RtlAppendUnicodeStringToString (&pDevExt->RegistryPath, RegistryPathName);
    
	/*
     * now store the device name in the registry. We need to store the
     * base name + number, not the leading \Device or \DosDevices, since
     * these are not seen by the win32 app (they prepend \\.\ to this
     * base name).
     */
    usName.Buffer = usNameBuffer;
    usName.MaximumLength = sizeof(usNameBuffer);
    usName.Length = 0;


    VT_CreateNumberedName(DeviceNumber
						  ,NULL
						  ,DD_VIDCAP_DEVICE_NAME_U
						  ,&usName
						  );

    /*
     * convert the counted string to a null-terminated string
     * before passing to the registry. Note that the Length field of
     * counted strings is a byte length.
     */
    usName.Buffer[usName.Length / sizeof(WCHAR)] = UNICODE_NULL;

    Status = RtlWriteRegistryValue(RTL_REGISTRY_ABSOLUTE
		                           ,pDevExt->ParametersKey
								   ,REG_DEVNAME
								   ,REG_SZ
								   ,usName.Buffer
								   ,usName.Length
								   );


    /* set up dispatch table */
    pDriverObject->DriverUnload                         = VT_Cleanup;
    pDriverObject->MajorFunction[IRP_MJ_CREATE]         = VT_Dispatch;
    pDriverObject->MajorFunction[IRP_MJ_CLOSE]          = VT_Dispatch;
    pDriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = VT_Dispatch;

	/* create the registry path so we can assign slot resources */
    usName.Buffer = usNameBuffer;
    usName.MaximumLength = sizeof(usNameBuffer);
    usName.Length = 0;


    VT_CreateNumberedName(DeviceNumber
						  ,NULL
						  ,DD_VIDCAP_DEVICE_NAME_U
						  ,&usName
						  );

	DeviceRegName.MaximumLength = MAX_NAME_SIZE;
    DeviceRegName.Buffer = ExAllocatePool (PagedPool, DeviceRegName.MaximumLength);
    RtlZeroMemory (DeviceRegName.Buffer, DeviceRegName.MaximumLength);
    DeviceRegName.Length = 0;

    RtlAppendUnicodeStringToString(&DeviceRegName, RegistryPathName);
    RtlAppendUnicodeToString(&DeviceRegName,L"\\");
    RtlAppendUnicodeStringToString(&DeviceRegName,&usName);

    /* Register resources with NT */
    Status=HalAssignSlotResources(&DeviceRegName
								  ,NULL
								  ,pDriverObject
								  ,pDevObj
								  ,pDevExt->BusType
								  ,pDevExt->BusNumber
								  ,pDevExt->SlotNumber.u.AsULONG
								  ,&pResourceList
								  );
    ExFreePool (DeviceRegName.Buffer);

    if (!NT_SUCCESS(Status)) {
	    KdPrint(("Cannot find device on second pass\n"));
		VT_Cleanup(pDriverObject);
		return(NULL);
    }


    /* 
     * Since ResourceList is bogus - see below - we need the info from PCI 
	 * config space - but since the first call returns bogus info, call it 
	 * again - why, cause thats the way it works!
	 */
    if (!VT_FindPCIDeviceEx(VenId
						   ,DevId
						   ,0
						   ,&PCICfg
						   ,sizeof(PCI_COMMON_CONFIG)
						   ,&pDevExt->BusNumber
						   ,&pDevExt->SlotNumber
						   ,&RevisionID
						   )) {
   
	    KdPrint(("Cannot find device on second pass\n"));
		VT_Cleanup(pDriverObject);
		return(NULL);
    }

    Ok = FALSE;
    ResrcIdx = 0;
    pPartialList = &pResourceList->List->PartialResourceList;

    /*
	 * Now complete the configuration of each resource
     * While we're at it, make sure at least 1 of the resources 
	 * is Memory Space
	 */
    while (ResrcIdx < pPartialList->Count) {
		
		if (pPartialList->PartialDescriptors[ResrcIdx].Type ==  CmResourceTypeMemory) {
			
			ULONG NrOfPorts;
			Ok = TRUE;
			NrOfPorts = pPartialList->PartialDescriptors[ResrcIdx].u.Memory.Length;

			DeviceAddress = pPartialList->PartialDescriptors[ResrcIdx].u.Memory.Start;
			VT_MapPort(pDevExt, DeviceAddress, NrOfPorts);

			break;
        }  

		ResrcIdx++;
    }

	/* free the resource list */
    ExFreePool(pResourceList);
	
	/* make sure that we got the memory mapped registers */
	if (!Ok) {
	    KdPrint(("Unable to determine port resouces\n"));
		VT_Cleanup(pDriverObject);
		return(NULL);
	}

	/* profile the PCI bridge type */
	VT_PCIChipSetProbe (pDevExt->BusNumber);
	pDevExt->RevisionID = RevisionID;

	/*	allocate the frame buffer memory  */
    if (!NT_SUCCESS(VT_AllocFrameMemory(pDevExt, FrameBufferSize))) {
	    KdPrint(("Unable to allocate frame buffer\n"));
		VT_WriteProfile(pDevExt, REG_STATUS, VC_INSUFFICIENT_RESOURCES);
		return pDevExt;
    }

	/* allocate the shared device info*/
    if (!NT_SUCCESS(VT_AllocDevInfo(pDevExt))) {
	    KdPrint(("Unable to allocate device info\n"));
		VT_WriteProfile(pDevExt, REG_STATUS, VC_INSUFFICIENT_RESOURCES);
		return pDevExt;
    }

    pDevExt->pUserInfo->State = INITIALIZED;
    VT_WriteProfile(pDevExt, REG_STATUS, VC_OK);
	return pDevExt;
}


/*
 * Cleanup on driver unload or aborted load.
 */
VOID
VT_Cleanup(PDRIVER_OBJECT pDriverObject)
{
    PDEVICE_EXT pDevExt;
    UNICODE_STRING usLink;
	UNICODE_STRING DeviceRegName;
    WCHAR LinkName[MAX_VIDCAP_NAME_LENGTH];
    NTSTATUS Status;


   /*
    * if there is no device created yet, then there can
    * be nothing to do
    */
	if (pDriverObject->DeviceObject == NULL) {
		return;
	}

	pDevExt = (PDEVICE_EXT) pDriverObject->DeviceObject->DeviceExtension;

	/* free the device info */
	VT_DeAllocDevInfo (pDevExt);

	/* free the frame buffer memory */
	VT_DeAllocFrameMemory(pDevExt);

    /* free memory mapping if memory-mapped i/o */
    if (pDevExt->PortMemType == 0 && pDevExt->NrOfPorts != 0) {
		MmUnmapIoSpace(pDevExt->pRegs, pDevExt->NrOfPorts);
    }

	/* create the registry path so we can deassign slot resources */
    usLink.Buffer = LinkName;
    usLink.MaximumLength = sizeof(LinkName);
    usLink.Length = 0;

    VT_CreateNumberedName(pDevExt->DeviceNumber
						  ,L"\\Device\\"
						  ,DD_VIDCAP_DEVICE_NAME_U
						  ,&usLink
						  );

    DeviceRegName.MaximumLength = MAX_NAME_SIZE;
    DeviceRegName.Buffer = ExAllocatePool (PagedPool, DeviceRegName.MaximumLength);
    RtlZeroMemory (DeviceRegName.Buffer, DeviceRegName.MaximumLength);
    DeviceRegName.Length = 0;

    RtlAppendUnicodeStringToString(&DeviceRegName, &pDevExt->RegistryPath);
    RtlAppendUnicodeToString(&DeviceRegName,L"\\");
    RtlAppendUnicodeStringToString(&DeviceRegName,&usLink);

    IoAssignResources(&DeviceRegName
					  ,NULL
					  ,pDriverObject
					  ,pDevExt->pDeviceObject
					  ,NULL
					  ,NULL
					  );

    ExFreePool (DeviceRegName.Buffer);

    /* free paged-pool registry pathname */
    if (pDevExt->ParametersKey != NULL) {
		ExFreePool(pDevExt->ParametersKey);
		pDevExt->ParametersKey = NULL;
    }

    /* free symbolic link */
    usLink.Buffer = LinkName;
    usLink.MaximumLength = sizeof(LinkName);
    usLink.Length = 0;

    VT_CreateNumberedName(pDevExt->DeviceNumber
						  ,L"\\DosDevices\\"
						  ,DD_VIDCAP_DEVICE_NAME_U
						  ,&usLink
						  );

    Status = IoDeleteSymbolicLink(&usLink);

    /* free device object */
    IoDeleteDevice(pDriverObject->DeviceObject);
}






