/*
 * rtloop.c
 *
 *					 
 * 32-bit Video Tool driver
 * realtime mpeg capture loop functions 
 * routines for Data Translation Video Tool.
 *
 * Mark Vrabel, October 96
 */

#include <vcdriver.h>
#include "hardware.h"
#include <bwdlgs.h>
#include <bwuser.h>
#include <rcuser.h>
#include <mmtimers.h>
#include <conio.h>
#include <stdio.h>
#include <math.h>
#include <direct.h>

#include "realtime.h"

// initial startup skip buffer count to compensate for c-cube's
// poor quality on first gop
#define RT_VIDEOSTARTCOUNT			(10)
#define RT_AUDIOSTARTCOUNT			(1)

// buffer allocation - try for 10 seconds - settle for 3 seconds
#define REQUIRED_SECONDS			(10.0)
#define MIN_SECONDS					(5.0)

// audio buffers are about 200 ms long (50 per 10 seconds).
// video buffer time depends on frame rate
#define AUDIO_BUFFER_TIME			(0.100)
#define VIDEO_BUFFER_TIME			((lpHw->VideoMode == NTSC) ?(.033367):(.04))

#define REQUIRED_AUDIOBUFFERS		((DWORD)(REQUIRED_SECONDS/AUDIO_BUFFER_TIME))
#define MIN_AUDIOBUFFERS			((DWORD)(MIN_SECONDS/AUDIO_BUFFER_TIME))

#define REQUIRED_VIDEOBUFFERS		((DWORD)(REQUIRED_SECONDS/VIDEO_BUFFER_TIME))
#define MIN_VIDEOBUFFERS			((DWORD)(MIN_SECONDS/VIDEO_BUFFER_TIME))

// since we don't know individual video frame size, we allocate
// a circular buffer of the required size
#define REQUIRED_VIDEOBUFFER_SIZE	((DWORD)(REQUIRED_SECONDS * (double)(lpRT->VideoDataRate/8)))
#define MIN_VIDEOBUFFER_SIZE		((DWORD)(MIN_SECONDS * (double)(lpRT->VideoDataRate/8)))
#define VIDEOBUFFER_SIZE_PER_SEC	((DWORD)(1.0 * (double)(lpRT->VideoDataRate/8)))
#define VIDEO_BUFFERS_PER_SECOND    ((DWORD)(1.0/VIDEO_BUFFER_TIME)) 

// allow .6 second to back up before we refuse to draw - since we're 
// allocating so much buffering, we can afford to be a little generous.
#define DRAW_AUDIO_BACKUP			(.6)
#define DRAW_VIDEO_BACKUP			(.6)
#define MAX_AUDIO_DONE				((DWORD)(DRAW_AUDIO_BACKUP/AUDIO_BUFFER_TIME))
#define MAX_VIDEO_DONE				((DWORD)(DRAW_VIDEO_BACKUP/VIDEO_BUFFER_TIME))

// max buffer size for VBV padding
#define PAD_BUFFERSIZE				(0x10000)

// the position in the VIDEOHDR.dwReserved array
// for storing the vbv padding count
#define PAD_SIZE_COUNT				(1)

/* declare local functions */
__declspec(dllexport) LRESULT CALLBACK MouseProc(int nCode, WPARAM wParam, LPARAM lParam);
__declspec(dllexport) LRESULT CALLBACK KeyboardProc(int nCode, WPARAM wParam, LPARAM lParam);

BOOL RT_MakeLastFrame (LPHWINFO lpHw, LPVIDEOHDR lpvh);
BOOL RT_FindLastFrame (LPHWINFO lpHw, LPVIDEOHDR lpvh);
BOOL RT_SetTimeCode (LPHWINFO lpHw, BYTE *lpData, DWORD dwBytes);
LRESULT RT_CheckFileSize (LPHWINFO lpHw);
BOOL RT_StopLoop (LPHWINFO lpHw);
VOID RT_StreamRemoveReadyHead(VCUSER_HANDLE vh, LPVIDEOHDR lpvh);
VOID RT_StreamAddReadyTail(VCUSER_HANDLE vh, LPVIDEOHDR lpvh);
VOID RT_StreamReplaceReadyHead(VCUSER_HANDLE vh, LPVIDEOHDR lpvh);
LRESULT RT_InitializeAudio (LPHWINFO lpHw);
LRESULT RT_InitializeVideo (LPHWINFO lpHw);
LRESULT RT_InitializeOutput (LPHWINFO lpHw);
LRESULT RT_SplitFile (LPHWINFO lpHw);
LRESULT RT_StartCapture (LPHWINFO lpHw);
LRESULT RT_StopCapture (LPHWINFO lpHw);
LRESULT RT_ProcessAudio (LPHWINFO lpHw); 
LRESULT RT_ProcessVideo (LPHWINFO lpHw); 
ULONG RT_RepeatFrame(LPHWINFO lpHw, LPUCHAR lpInput, ULONG InputLength, LPUCHAR lpOutput, ULONG OutputLength, LPULONG lpTimeStamp);

/*
 * needed to bybass subclassed window proc
 */
extern FARPROC gfpOrigCapWndProc; 
extern DWORD ghThreadId;

/* 
 * use globally for message strings
 */
char buffer[MAX_PATH];
char message[MAX_PATH];

#define INPUT_SIZE  (1152)
#define MAX_TIMEOUT (2000)
  
extern LPHWINFO glpHwCurrent;


__declspec(dllexport) LRESULT CALLBACK 
MouseProc(int nCode, WPARAM wParam, LPARAM lParam)
{
    LPHWINFO lpHw = glpHwCurrent;
	LPRTINFO lpRT = &lpHw->RTInfo;

	if (nCode < 0 || !lpHw->fCapturingNow || lpRT->bStop) {
		return CallNextHookEx(lpRT->hHookM, nCode, wParam, lParam);
	}

	if (lpRT->CapParams.fAbortRightMouse) {
	
		if (wParam == WM_RBUTTONDOWN || wParam == WM_NCRBUTTONDOWN) {
			SetEvent (lpRT->hCaptureStop);
			return (1);
		}
	}
	if (lpRT->CapParams.fAbortLeftMouse) {
	
		if (wParam == WM_LBUTTONDOWN || wParam == WM_NCRBUTTONDOWN) {
			SetEvent (lpRT->hCaptureStop);
			return (1);
		}
	}

	return CallNextHookEx(lpRT->hHookM, nCode, wParam, lParam);
}


__declspec(dllexport) LRESULT CALLBACK 
KeyboardProc(int nCode, WPARAM wParam, LPARAM lParam)
{
    LPHWINFO lpHw = glpHwCurrent;
	LPRTINFO lpRT = &lpHw->RTInfo;

	if (nCode < 0 || !lpHw->fCapturingNow || lpRT->bStop) {
		return CallNextHookEx(lpRT->hHookM, nCode, wParam, lParam);
	}

 	if (lParam & 0x80000000 && wParam == lpRT->CapParams.vKeyAbort) {
		SetEvent (lpRT->hCaptureStop);
		return (1);
	}
							
	return CallNextHookEx(lpRT->hHookM, nCode, wParam, lParam);
}						 


LRESULT 
RT_CaptureLoop (LPHWINFO lpHw)
{

	LPRTINFO lpRT = &lpHw->RTInfo;
	HWND hWnd = lpHw->hWndCap;
	LRESULT lResult;
	MSG Msg;

    OSVERSIONINFO VerInfo;
    
	/* initialize the realtime info structure */
	FillMemory (lpRT, sizeof(RTINFO), 0);
	lpRT->bCaptureOK = TRUE;
	((DWORD*)lpRT->StopCodeFrame)[0] = 0xb7010000;
	
    /* 
     * We need to get the platform information to determine 
     * if we are running on WinNT or Win95
     */
    VerInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    if (!GetVersionEx(&VerInfo)) {
		lpRT->ErrorCode = IDS_CAP_AVI_INIT_ERROR;
        return RT_StopCapture (lpHw);
    }
	
	/* 
	 * set the max file size - 2GB for win95, 4GB for winNT - 
	 * subtract 5MB for good measure
	 */ 
    if (VerInfo.dwPlatformId == VER_PLATFORM_WIN32_NT) {
		lpRT->dwMaxFileSize	= 0xffffffff - ((1024*1024)*5);
	} else {
		lpRT->dwMaxFileSize	= 0x80000000 - ((1024*1024)*5);
	}

	/* go to a wait cursor while we initialize */
	lpRT->OldCursor = SetCursor(LoadCursor(NULL,MAKEINTRESOURCE(IDC_WAIT)));

	/* initialize status info */
	lpHw->dwCurrentVideoFrame = 0; 
	lpHw->dwCurrentVideoFramesDropped = 0; 
	lpHw->dwCurrentWaveSamples = 0; 
	lpHw->dwCurrentTimeElapsedMS = 0; 
	lpHw->fCapturingNow = FALSE; 
	lpHw->wNumVideoAllocated = 0; 
	lpHw->wNumAudioAllocated = 0; 

	/* extract the compression options */
    lpRT->VideoDataRate = (DWORD)(lpHw->RecodePreferences.nEncodeVideoBitRate);
    lpRT->UseQSIF = lpHw->RecodePreferences.nLargeImage ?FALSE:TRUE;
    lpRT->AudioDataRate = lpHw->RecodePreferences.nEncodeAudioBitRate/1000;
	lpRT->AudioMode = BW_AudioToMode (lpHw->RecodePreferences.nEncodeAudioEnable, lpHw->RecodePreferences.nEncodeAudioMode);
    lpRT->StreamDataRate = lpHw->RecodeOptions.nEncodeUseOptimumRate ?0.0:(double)lpHw->RecodeOptions.nEncodeStreamDataRate; 
    lpRT->PackSize = lpHw->RecodeOptions.nEncodePackSize;
    lpRT->PacketsPerPack = lpHw->RecodeOptions.nEncodePacketsPerPack;
	lpRT->StreamMixType = (EStreamMixType)lpHw->RecodeOptions.nStreamMixType;

	lpRT->dwSizeGOP = (DWORD)((double)lpHw->RecodePreferences.nCompressionEstimate / (lpHw->VideoMode == NTSC ?26.6:22.2));

	if (lpHw->bCreateNoFile || lpHw->dwMaxFileSize == FILE_SIZE_NOFILE) {
		lpRT->bCreateNoFile = TRUE;
	} else {
		lpRT->bCreateNoFile = FALSE;
	}

	/* get the capture info */
	if (!capGetStatus (hWnd, &lpRT->CapStatus, sizeof(lpRT->CapStatus))) {
		lpRT->ErrorCode = IDS_CAP_AVI_INIT_ERROR;
        return RT_StopCapture (lpHw);
	}
	if (!capCaptureGetSetup (hWnd, &lpRT->CapParams, sizeof(lpRT->CapParams))) {
		lpRT->ErrorCode = IDS_CAP_AVI_INIT_ERROR;
        return RT_StopCapture (lpHw);
	}

	/* can't capture if already capturing */
	if (lpRT->CapStatus.fCapturingNow) {
		lpRT->ErrorCode = IDS_CAP_AVI_INIT_ERROR;
        return RT_StopCapture (lpHw);
	}

	/* are we capturing audio only, video only, or stream mixing ??? */
	lpRT->fCaptureAudio = (lpRT->AudioMode != AUDIO_NONE);
	lpRT->fCaptureVideo = lpHw->RecodePreferences.nEncodeVideoEnable;
	lpRT->bStreamMix = (lpRT->StreamMixType != NoMix && lpRT->fCaptureAudio && lpRT->fCaptureVideo && !lpRT->bCreateNoFile) ?TRUE:FALSE;

	/* setup any shtuff we need to stop */
	if (lpRT->CapParams.fYield) 
	{
		lpRT->hCaptureStop = CreateEvent (0, 0, 0, 0);
	
		/* should we stop on a mouse button??? */
		if (lpRT->CapParams.fAbortRightMouse || lpRT->CapParams.fAbortRightMouse) {
			lpRT->hHookM = SetWindowsHookEx(WH_MOUSE, MouseProc, ghDLLInst, ghThreadId);
		}

		/* should we stop on the escape key??? */
		if (lpRT->CapParams.vKeyAbort != 0) {
			lpRT->hHookK = SetWindowsHookEx(WH_KEYBOARD, KeyboardProc, ghDLLInst, ghThreadId);
		}
	}

	/* if capturing audio, initialize audio capture */
	if (lpRT->fCaptureAudio) {

		/* make sure we have an audio device */
		if (!lpRT->CapStatus.fAudioHardware) {
			lpRT->ErrorCode = IDS_CAP_WAVE_OPEN_ERROR;
	        return RT_StopCapture (lpHw);
		}

		/* initialize */										   
		if ((lpRT->ErrorCode = RT_InitializeAudio (lpHw)) != 0) {
	        return RT_StopCapture (lpHw);
		}
	}

	/* if capturing video, initialize video capture */
	if (lpRT->fCaptureVideo) {
		
		/* initialize */										   
		if ((lpRT->ErrorCode = RT_InitializeVideo (lpHw)) != 0) {
	        return RT_StopCapture (lpHw);
		}
	}

	/* initialize the output stream/file(s) */
	if ((lpRT->ErrorCode = RT_InitializeOutput (lpHw)) != 0) {
        return RT_StopCapture (lpHw);
	}
	 		    
	/* start 'em up!!! */
	if ((lpRT->ErrorCode = RT_StartCapture (lpHw)) != 0) {

		if (lpRT->ErrorCode == IDS_CAP_END) {
			lpRT->ErrorCode = 0;
		}

		return RT_StopCapture (lpHw);
	}
	 
	/* loop until signaled to stop */
	while (TRUE) {

		/* time to quit??? */
		if (lpRT->CapParams.fYield) {

			/* 
			 * we use mouse and keyboard hook procs when operating
			 * as a worker thread so we won't miss a message
			 */
			if (!lpRT->bStop && WaitForSingleObject (lpRT->hCaptureStop, 0) == WAIT_OBJECT_0) {		
				if (RT_StopLoop (lpHw)) {
					break;
				}
			}

			if (!lpRT->bStop && WaitForSingleObject (lpHw->hCaptureComplete, 0) == WAIT_OBJECT_0) {
				if (RT_StopLoop (lpHw)) {
					break;
				}
			} 

		} else {

			/*
			 * when not opperating as a worker thread we peek 
			 * for the specific messages to stop
			 */
			if (!lpRT->bStop && lpRT->CapParams.fAbortRightMouse) {
				if (PeekMessage (&Msg, NULL, WM_RBUTTONDOWN, WM_RBUTTONDOWN, PM_NOREMOVE)) {
					if (RT_StopLoop (lpHw)) {
						break;
					}
				} else if (PeekMessage (&Msg, NULL, WM_NCRBUTTONDOWN, WM_NCRBUTTONDOWN, PM_NOREMOVE)) {
					if (RT_StopLoop (lpHw)) {
						break;
					}
				}
			}

			if (!lpRT->bStop && lpRT->CapParams.fAbortLeftMouse) {
				if (PeekMessage (&Msg, NULL, WM_LBUTTONDOWN, WM_LBUTTONDOWN, PM_NOREMOVE)) {
					if (RT_StopLoop (lpHw)) {
						break;
					}
				} else if (PeekMessage (&Msg, NULL, WM_NCLBUTTONDOWN, WM_NCLBUTTONDOWN, PM_NOREMOVE)) {
					if (RT_StopLoop (lpHw)) {
						break;
					}
				}
			}

			if (!lpRT->bStop && lpRT->CapParams.vKeyAbort != 0) {
				if (PeekMessage (&Msg, NULL, WM_KEYDOWN, WM_KEYDOWN, PM_NOREMOVE)) {
					if (Msg.wParam == lpRT->CapParams.vKeyAbort && RT_StopLoop (lpHw)) {
						break;
					}
				}
			}
		}

		if (lpRT->bStop && timeGetTime() - lpRT->dwStopTime >= MAX_TIMEOUT) {
			break;
		}

		if (!lpRT->bStop && lpRT->CapParams.fLimitEnabled) {
			if (lpHw->dwCurrentTimeElapsedMS >= lpRT->CapParams.wTimeLimit*1000) {
				if (RT_StopLoop (lpHw)) {
					break;
				}
			}
		}

		/* 
		 * yield callback is called at least once per frame - but try 
		 * to call it anytime we're not busy 
		 */
		if (lpHw->lpfnYieldCallback && !lpRT->lpProcessWaveHdr && !lpRT->lpProcessVideoHdr && !lpRT->AudioDoneCount && !lpRT->VideoDoneCount) {
			lpHw->lpfnYieldCallback (hWnd);
		}

		/* if capturing audio, deal with it */
		if (lpRT->fCaptureAudio && !lpRT->bAudioDone) { 

			if ((lResult = RT_ProcessAudio (lpHw)) != 0) { 
				
				if (lResult != IDS_CAP_END) {
					lpRT->ErrorCode = lResult;
					break;
				}

				lpRT->bAudioDone = TRUE;
			}
		}

		/* yield a little to let the timer proc run */
		Sleep (0);

		/* if capturing video, deal with it */
		if (lpRT->fCaptureVideo) {

			/* enter the critical section while we process video */
			EnterCriticalSection (&lpHw->RTLoopCritical);
	
			lResult = RT_ProcessVideo (lpHw);

			/* leave the critical after processing */
			LeaveCriticalSection (&lpHw->RTLoopCritical);
		
			if (lResult != 0) { 
				
				/* if capture ended we're either stopped or we're breaking
				 * up the files for PictureTel.
				 */
				if (lResult == IDS_CAP_END) {
					
					lpRT->ErrorCode = 0;

					if (!lpRT->bStop && lpRT->bSplitFile) {
						
						lResult = RT_SplitFile (lpHw);
						if (lResult) {
							lpRT->ErrorCode = lResult;
							break;
						}
					} else {
						break;
					} 
				} else {
					lpRT->ErrorCode = lResult;
					break;
				}
			}
		}

		/* yield a little to let the timer proc run */
		if (lpRT->VideoDoneCount == 0 && lpRT->AudioDoneCount == 0) {
			Sleep(1);
		} else {
			Sleep(0);
		}
	}

	/* did we really stop because the file was too big to write to? */
	if (lpRT->bDiskFull) {
		lpRT->ErrorCode = IDS_CAP_RECORDING_ERROR;
	}

	/* 
	 * all done
	 */
	return RT_StopCapture (lpHw);
}


int 
RT_LoadAviCapString(LPHWINFO lpHw, UINT uID, LPTSTR lpBuffer, int nBufferMax)
{
	LPRTINFO lpRT = &lpHw->RTInfo;

	if (uID >= IDS_RT_FIRST_STRING) {
		   
		if (uID == IDS_RT_VIDEO_DROP_ERROR && lpHw->dwCaptureStopped != 0) {
			return LoadString(ghDLLInst, IDS_RT_VIDEO_LOST_ERROR, lpBuffer, nBufferMax);
		} else {
			return LoadString(ghDLLInst, uID, lpBuffer, nBufferMax);
		}
	} else {
											  
		if (lpHw->hVfw == NULL) {
			return 0;
 		} else if (uID == IDS_CAP_AUDIO_DROP_ERROR) {
			return LoadString(ghDLLInst, IDS_RT_AUDIO_DROP_ERROR, lpBuffer, nBufferMax);
		} else if (lpRT->bDiskFull && uID == IDS_CAP_RECORDING_ERROR) {
			uID = lpRT->dwMaxFileSize > 0x80000000 ?IDS_RT_CAPTURE_LIMIT_4GB:IDS_RT_CAPTURE_LIMIT_2GB;
			return LoadString(ghDLLInst, uID, lpBuffer, nBufferMax);
		} else {
			return LoadString(lpHw->hVfw, uID, lpBuffer, nBufferMax);
		}
	}
}


BOOL
RT_HaveFullGop (LPHWINFO lpHw) 
{
	LPRTINFO lpRT = &lpHw->RTInfo;
	DWORD dwIndex, dwCount, i;

	/* are we currently processing a gop??? */
	if (lpRT->bHaveFullGop) {
		
		/* have we processed the last frame of the gop??? */
		if (lpRT->dwGopFrameCount != 0) {
			return TRUE;
		} else {
			lpRT->bHaveFullGop = FALSE;
			lpRT->dwGopFrameCount = 0;
			lpRT->dwGopStartIndex = lpRT->dwNextGopStartIndex;
		}
	}

	/* do we need to flush out the remaining frames??? */
	if (lpRT->bFlushToLastGOP && lpRT->dwNewStartCountDown == 0) {
			
		if (lpRT->dwDropCount > lpRT->VideoDoneCount) {
			lpRT->dwDropCount = lpRT->VideoDoneCount;
		}

		for (i=0;i<lpRT->dwDropCount;i++) {

			LPVIDEOHDR lpvh = &lpRT->videoHdr[lpRT->VideoDoneIndex];

			lpvh->dwBytesUsed = 0;
			lpvh->dwFlags &= ~WHDR_DONE;
			lpvh->dwFlags &= ~VHDR_INQUEUE;

			VC_AddReadyTail(lpHw->vh, lpvh);

			lpRT->VideoDoneIndex = (lpRT->VideoDoneIndex + 1) % lpHw->wNumVideoAllocated;
			lpRT->VideoDoneCount--;
		}

		lpRT->dwDropCount = 0;
		lpRT->bFlushToLastGOP = FALSE;
		lpRT->bFirstGop = TRUE;
	}

	if (lpRT->bFirstGop) {
		lpRT->dwGopStartIndex = lpRT->VideoDoneIndex;
		lpRT->dwGopFrameCount = 0;
	}

	dwIndex = lpRT->dwGopStartIndex;
	dwCount = lpRT->VideoDoneCount;

	/* find the end of the gop */
	lpRT->bGotStopFrame   = FALSE;
	lpRT->dwGopFrameCount = 1;

	for (i=0;i<dwCount;i++) {  

		LPVIDEOHDR lpvh = &lpRT->videoHdr[dwIndex];
		if (RT_FindLastFrame (lpHw, lpvh)) {

			lpRT->dwNextGopStartIndex = (dwIndex + 1) % lpHw->wNumVideoAllocated;
			lpRT->bHaveFullGop = TRUE;
			return TRUE;
		}

		dwIndex = (dwIndex + 1) % lpHw->wNumVideoAllocated;
		lpRT->dwGopFrameCount++;
	}

	return FALSE;
}


BOOL 
RT_FlushToLastGop (LPHWINFO lpHw)
{	
	LPRTINFO lpRT = &lpHw->RTInfo;
	DWORD dwIndex, dwCount, i;
	LPVIDEOHDR lpvh;
	
	/* we're called from the timer routine - go critical!!! */
	EnterCriticalSection (&lpHw->RTLoopCritical);

	/* make sure we have all the done buffers */
	while (lpRT->videoHdr[lpRT->VideoReadyIndex].dwFlags & VHDR_DONE) {
		
		lpvh = &lpRT->videoHdr[lpRT->VideoReadyIndex];	

		lpRT->VideoReadyIndex = (lpRT->VideoReadyIndex + 1) % lpHw->wNumVideoAllocated;
		lpRT->VideoDoneCount++;
	}

	if (lpRT->bHaveFullGop) {
		dwIndex = lpRT->dwNextGopStartIndex;
		dwCount = lpRT->VideoDoneCount - lpRT->dwGopFrameCount;
	} else {
		dwIndex = lpRT->VideoDoneIndex;
		dwCount = lpRT->VideoDoneCount;
	}

	lpRT->bGotStopFrame = FALSE;
	lpRT->dwDropCount   = 0;

 	for (i=0;i<dwCount;i++) {  

		lpvh = &lpRT->videoHdr[dwIndex];
		if (RT_FindLastFrame (lpHw, lpvh)) {
			lpRT->dwDropCount = 0;
		} else {
			lpRT->dwDropCount++;
		}

		dwIndex = (dwIndex + 1) % lpHw->wNumVideoAllocated;
	}

	lpRT->bGotStopFrame = FALSE;

	/* we want the video time to be that of the last good frame */
	lpHw->dwInterruptCount -= lpRT->dwDropCount;
	lpHw->dwVideoTime = MulDiv (lpHw->dwInterruptCount, lpHw->dwMicroPerInterrupt, 1000L);                         
	lpHw->dwTimePrevious = lpHw->dwVideoTime;
	
	lpRT->bFlushToLastGOP = TRUE;
	lpRT->dwNewStartCountDown = lpRT->VideoDoneCount - lpRT->dwDropCount;

	LeaveCriticalSection (&lpHw->RTLoopCritical);

	return TRUE;
}


BOOL
RT_MakeLastFrame (LPHWINFO lpHw, LPVIDEOHDR lpvh)
{								
	LPRTINFO lpRT = &lpHw->RTInfo;
		
	DWORD dwSize = lpvh->dwBytesUsed;
	DWORD* lpSearch = (DWORD*)lpvh->lpData;

	DWORD dwX, FrameType, dwTemporalReference;
	DWORD dwIndex = 0;

	while (dwSize != 0) {       
            
		if (lpSearch[dwIndex] == 0x00010000) {
			break;
		}
        
		dwIndex++;
		dwSize--;
	}

	if (dwSize == 0) {
		return FALSE;
	}
									   
	dwIndex++;
	dwX = ByteSwap (lpSearch[dwIndex]);
	FrameType = (dwX & 0x00380000) >> 19;
	dwTemporalReference  = (dwX & 0xFFC00000) >> 22;
	
	if (lpRT->bGotGoodStop) {
		
		if (dwTemporalReference == lpRT->dwLastReference) {
			
			if (lpRT->bStop || lpRT->bSplitFile) {
				CopyMemory (lpvh->lpData+lpvh->dwBytesUsed, lpRT->StopCodeFrame, sizeof(DWORD));
				lpvh->dwBytesUsed += sizeof(DWORD);
			}

			lpRT->bGotGoodStop = FALSE;
			lpRT->bGotIFrame   = FALSE;
			lpRT->PFrameCount  = 0L;
			
			return TRUE;
		}

	} else if (FrameType == I_FRAME || FrameType == P_FRAME) {

		if (lpRT->bSplitFile) {
			
			/* 
			 * for the purpose of PictureTel file splitting, we want to
			 * force the next frame after the stop code to be an I-Frame
			 * so the next file will start at a valid point without too
			 * much hastle. A new I-Frame comes in display order after
			 * the second P-Frame
			 */
			if (!lpRT->bGotIFrame) {
				if (FrameType == I_FRAME) {
					lpRT->bGotIFrame = TRUE;
				}
			} else if (FrameType == P_FRAME) {
				if (++lpRT->PFrameCount == 2) {
					lpRT->dwLastReference = dwTemporalReference - 1;
					lpRT->bGotGoodStop = TRUE;
				}
			}

		} else {
			lpRT->dwLastReference = dwTemporalReference - 1;
			lpRT->bGotGoodStop = TRUE;
		}
	}
	
	return FALSE;
}


BOOL
RT_FindLastFrame (LPHWINFO lpHw, LPVIDEOHDR lpvh)
{								
	LPRTINFO lpRT = &lpHw->RTInfo;
		
	DWORD dwSize = lpvh->dwBytesUsed/sizeof(DWORD);
	DWORD* lpSearch = (DWORD*)lpvh->lpData;

	MPEG_TEMPORAL_REFERENCE mtr;
	DWORD dwIndex = 0;

	while (dwSize != 0) {       
            
		if (lpSearch[dwIndex] == 0x00010000) {
			break;
		}
        
		dwIndex++;
		dwSize--;
	}

	if (dwSize == 0) {
		return FALSE;
	}
									   
	dwIndex++;
	mtr.dword = ByteSwap (lpSearch[dwIndex]);

	if (lpRT->bGotStopFrame) {
		
		if (mtr.fld.tr == lpRT->dwLastTemporalReference) {
			
			lpRT->bGotStopFrame = FALSE;
			return TRUE;
		}

	} else if (mtr.fld.pt == I_FRAME || mtr.fld.pt == P_FRAME) {

		lpRT->dwLastTemporalReference = mtr.fld.tr - 1;
		lpRT->bGotStopFrame = TRUE;
	}
	
	return FALSE;
}


BOOL 
RT_CalcVBVDelay (LPHWINFO lpHw, LPVIDEOHDR lpvh)
{
	LPRTINFO lpRT = &lpHw->RTInfo;

	double FrameRate = (lpHw->VideoMode == PAL) ? 25.00 : 29.97;
	
	unsigned __int64 Bitcount = (unsigned __int64)lpRT->dwCurrentVideoBytes * 8;
	unsigned __int64 bitcnt_EOP = Bitcount;
	unsigned __int64 VBVBufferSize = 327680;
	unsigned __int64 PictureDelay;
	unsigned __int64 PadBits = 0;

	DWORD PadBytes = 0;
	int vbv_delay;

	DWORD dwSize = lpvh->dwBytesUsed/sizeof(DWORD);
	DWORD* lpSearch = (DWORD*)lpvh->lpData;
	DWORD dwIndex = 0;

	MPEG_TEMPORAL_REFERENCE mtr;

	/* assume no padding */
	lpvh->dwReserved[PAD_SIZE_COUNT] = 0;
	
	/* 
	 * reference point for vbv_delay is directly after writing 
	 * the picture start code
     */
	while (dwSize != 0) {       
            
		if (lpSearch[dwIndex] == 0x00010000) {
			break;
		}
        
		dwIndex++;
		dwSize--;
	}

	if (dwSize == 0) {
		return FALSE;
	} 

	dwIndex += 1;
	Bitcount += (dwIndex * 32);

	mtr.dword = ByteSwap (lpSearch[dwIndex]);

	/* 
	 * first call of calc_nVBVDelay - we start with 
	 * a 7/8 filled VBV buffer (12.5% back-off) 
	 */
	if (lpRT->DecodingTime == 0) {
		PictureDelay = (((VBVBufferSize * 7)/8) * 90000)/lpRT->VideoDataRate;
	} else {
		PictureDelay = (unsigned __int64)(90000.0/(FrameRate));
	}

	/* when to decode current frame */
	lpRT->DecodingTime += PictureDelay;

	/* 
	 * if we're going to overflow the vbv buffer - pad the frame out with zero's 
	 * to maintain rate control.
	 */
	if ( (lpRT->DecodingTime - ((bitcnt_EOP*90000)/lpRT->VideoDataRate))
      > ((VBVBufferSize*90000)/lpRT->VideoDataRate)) {
		
		/*
		 * strive to maintain a 9/10 filled VBV buffer (10% back-off) 
		 */
		PadBits = (((lpRT->DecodingTime*lpRT->VideoDataRate)/90000) - ((VBVBufferSize*9)/10) ) - bitcnt_EOP;
		PadBits = (PadBits + 31) & ~31; /* dword align */
		PadBytes = (DWORD)PadBits/8;
	}

	/* calculate vbv_delay */
	vbv_delay = (int)(lpRT->DecodingTime - (((Bitcount+PadBits)*90000)/lpRT->VideoDataRate));

	/* fix up blatent under/over flows */
	if (vbv_delay < 0) {
		vbv_delay = 0;
	}
	if (vbv_delay > 65535) {
	    vbv_delay = mtr.fld.dly;
	}

	/* store vbv_delay in the picture */
	mtr.fld.dly = vbv_delay;
	lpSearch[dwIndex] = ByteSwap (mtr.dword);

	/* increment the video output size */
	lpRT->dwCurrentVideoBytes += lpvh->dwBytesUsed + PadBytes;
	lpvh->dwReserved[PAD_SIZE_COUNT] = PadBytes;

	return TRUE;
}


#define MAX_FRAMES  (lpHw->VideoMode == NTSC ?30:25)
#define MAX_SECONDS (60)
#define MAX_MINUTES (60)
#define MAX_HOURS   (24)

BOOL 
RT_SetTimeCode (LPHWINFO lpHw, BYTE *lpData, DWORD dwBytes)
{
	LPRTINFO lpRT = &lpHw->RTInfo;

    DWORD *lpSearch = (DWORD*)lpData;
	
    DWORD dwSize = dwBytes/sizeof(DWORD);
    DWORD Index = 0, i;

    DWORD dwIncrement = 0x01;
	MPEG_TIME_CODE tc;
	MPEG_TIME_CODE *ptc = &lpRT->tc;

    /* 
     * search for the gop header; 
     */
    while (dwSize >= dwIncrement) {       
            
        if (lpSearch[Index] == 0xb8010000) {
			break;
		}
        
        dwSize -= min (dwSize,dwIncrement);
	    Index += min (dwSize,dwIncrement);
    }

    /* a size of zero indicates no data is ready */
    if (dwSize <= dwIncrement) {
        return FALSE;
    }

	tc.dword = ByteSwap (lpSearch[Index + 1]);

	tc.fld.h = ptc->fld.h;
	tc.fld.m = ptc->fld.m;
	tc.fld.s = ptc->fld.s;
	tc.fld.f = ptc->fld.f;

	dwIncrement = (lpRT->FrameCount - lpRT->LastGop);
	lpRT->LastGop = lpRT->FrameCount;

	for (i=0;i<dwIncrement;i++) {

		tc.fld.f = (ptc->fld.f + 1) % MAX_FRAMES;
		if (tc.fld.f == 0) {
			tc.fld.s = (ptc->fld.s + 1) % MAX_SECONDS;
			if (tc.fld.s == 0) {
				tc.fld.m = (ptc->fld.m + 1) % MAX_MINUTES;
				if (tc.fld.m == 0) {
					tc.fld.h = (ptc->fld.h + 1) % MAX_HOURS;
				}
			}

			if (tc.fld.dff && tc.fld.m %10 != 0 && tc.fld.s == 0) {
				tc.fld.f += 2;
			}
		}

		ptc->dword = tc.dword;
	}

	/* if skipping first gop - set the broken link flag */
	if (RT_VIDEOSTARTCOUNT != 0 && lpRT->FrameCount == 0) 	{ 
		tc.fld.bl = 1;
	}
	
	lpSearch[Index + 1] = ByteSwap (tc.dword);
	return TRUE;
}			   


LRESULT
RT_CheckFileSize (LPHWINFO lpHw) 
{
	LPRTINFO lpRT = &lpHw->RTInfo;
	DWORD dwCurrentBytes = lpRT->dwSizeGOP;
	BOOL bFileTooBig = FALSE;
	DWORD dwMaxFileSize;

	/* current Broadway app checks for 2GB file size limtit */
	if (lpHw->IsBroadway && lpHw->dwMaxFileSize == 0) {
		return 0;
	}

	/* do we need to check file size */
	if (lpRT->bCreateNoFile || lpRT->bStop || lpRT->bSplitFile) { 
		return 0;
	}

	/* get video file size */
	if (lpRT->OutFile != NULL) {
		dwCurrentBytes += GetFileSize(lpRT->OutFile, NULL);
	} 	

	/* get audio file size */
	if (lpRT->AudioFile != NULL) {
		dwCurrentBytes += GetFileSize(lpRT->AudioFile, NULL);
	} 

	/* are we checking for a file spit  or a file system limit */
	if (lpHw->dwMaxFileSize != 0) { 
		dwMaxFileSize = lpHw->dwMaxFileSize;
	} else {
		dwMaxFileSize = lpRT->dwMaxFileSize;
	}		  

	if (dwCurrentBytes < dwMaxFileSize) {
		return 0;
	}

	/* PictureTel wants the mpeg files on 600mB boundries - is it time to 
     * start a new file???
	 */	
	if (lpHw->dwMaxFileSize != 0) { 

		/* 
		 * 
		 * if capturing video we want to make sure we break the file up on
		 * valid boundries (MPEG spec shtuff).
		 */
		if (lpRT->fCaptureVideo) {
			lpRT->bSplitFile = TRUE;			
		} else {
			/* if no video just split the file now!!! */
			return RT_SplitFile (lpHw);
		}
	} else {
		lpRT->bDiskFull = TRUE;	

		if (RT_StopLoop (lpHw)) {
			return IDS_CAP_RECORDING_ERROR;
		} 
	}
	
	return 0;
}


BOOL RT_StopLoop (LPHWINFO lpHw)
{
	LPRTINFO lpRT = &lpHw->RTInfo;

	if (lpRT->fCaptureVideo) {
		
		lpRT->bStop = TRUE;
		lpRT->dwStopTime = timeGetTime();
		
		return FALSE;

	} else {

		return TRUE;
	}
}
	

VOID
RT_StreamRemoveReadyHead(VCUSER_HANDLE vh, LPVIDEOHDR lpvh)
{
  	LPHWINFO lpHw = (LPHWINFO)VC_GetHWInfo(vh);
	LPRTINFO lpRT = &lpHw->RTInfo;

    LPVIDEO_BUFFER lpvb;
	DWORD dwSize;

	if (lpHw->CaptureMode != CAPMODE_RTMPEG) {
		return;
	}

    lpvb = &lpHw->CaptBufs.Bufs[lpHw->CaptBufs.dwCurrentBuf];
	dwSize = lpvb->dwFilled*sizeof(DWORD);

	if (dwSize > (lpRT->dwVideoLength - (lpRT->dwVideoFilled + lpRT->dwVideoUnused))) {
		lpvh->dwBufferLength = 0; 
	} else {
		lpvh->dwBufferLength = dwSize; 
	}

	if (dwSize >= (lpRT->dwVideoLength - lpRT->dwVideoTail)) {
		
		lpRT->dwVideoUnused = lpRT->dwVideoLength - lpRT->dwVideoTail;
		lpvh->lpData = lpRT->lpVideoBuf;
		lpRT->dwVideoTail = 0;
	} 

	lpvh->lpData = lpRT->lpVideoBuf+lpRT->dwVideoTail;
	lpvh->dwReserved[HEADER_DATA] = (DWORD)lpvh->lpData;

	lpRT->dwVideoTail += dwSize;
	lpRT->dwVideoFilled += dwSize;
}


VOID
RT_StreamAddReadyTail(VCUSER_HANDLE vh, LPVIDEOHDR lpvh)
{
  	LPHWINFO lpHw = (LPHWINFO)VC_GetHWInfo(vh);
	LPRTINFO lpRT = &lpHw->RTInfo;

	if (lpHw->CaptureMode != CAPMODE_RTMPEG) {
		return;
	}

	lpRT->dwVideoFilled -= lpvh->dwBufferLength; 
	lpvh->dwBufferLength = 0;

	lpvh->lpData = NULL;
	lpvh->dwReserved[HEADER_DATA] = 0;
}


VOID
RT_StreamReplaceReadyHead(VCUSER_HANDLE vh, LPVIDEOHDR lpvh)
{
  	LPHWINFO lpHw = (LPHWINFO)VC_GetHWInfo(vh);
	LPRTINFO lpRT = &lpHw->RTInfo;

	if (lpHw->CaptureMode != CAPMODE_RTMPEG) {
		return;
	}

	if (lpRT->dwVideoLength - lpRT->dwVideoTail == lpRT->dwVideoUnused) {
		lpRT->dwVideoTail -= lpvh->dwBufferLength;
		lpRT->dwVideoUnused = 0;
	} else if (lpRT->dwVideoTail - lpvh->dwBufferLength == 0) {
		lpRT->dwVideoTail = lpRT->dwVideoLength - lpRT->dwVideoUnused;
		lpRT->dwVideoUnused = 0;
	} else {
		lpRT->dwVideoTail -= lpvh->dwBufferLength;
	}

	lpRT->dwVideoFilled -= lpvh->dwBufferLength; 
	lpvh->dwBufferLength = 0;

	lpvh->lpData = NULL;
	lpvh->dwReserved[HEADER_DATA] = 0;
}


LRESULT 
RT_InitializeAudio (LPHWINFO lpHw)
{
	DWORD wNumAudioRequested = REQUIRED_AUDIOBUFFERS;
	LPRTINFO lpRT = &lpHw->RTInfo;
	DWORD Index;

	/* extract the waveformat */
	lpRT->WaveFormat.wFormatTag      = WAVE_FORMAT_PCM; 
	lpRT->WaveFormat.nChannels	     = lpRT->AudioMode == AUDIO_MONO ?1:2;
	lpRT->WaveFormat.nSamplesPerSec  = lpHw->AudioRate; 
	lpRT->WaveFormat.wBitsPerSample  = 16; 
    lpRT->WaveFormat.nBlockAlign     = (lpRT->WaveFormat.nChannels * lpRT->WaveFormat.wBitsPerSample) / 8;      
	lpRT->WaveFormat.nAvgBytesPerSec = lpRT->WaveFormat.nSamplesPerSec * lpRT->WaveFormat.nBlockAlign;  
	lpRT->WaveFormat.cbSize          = 0;            

	/* pcm audio only */
	if (lpRT->WaveFormat.wFormatTag != WAVE_FORMAT_PCM) {
		return IDS_RT_WAVE_FORMAT_NOT_PCM_ERROR;
	}

	/* 16 bit audio only */
	if (lpRT->WaveFormat.wBitsPerSample != 16) {
		return IDS_RT_WAVE_FORMAT_NOT_16BIT_ERROR;
	}

	/* open the wave input device */
	if (waveInOpen(&lpRT->hWave, WAVE_MAPPER, &lpRT->WaveFormat, 0, 0, 0) != 0) {
		return IDS_CAP_WAVE_OPEN_ERROR;
	}

	/* allocate the wavehdr array */
	lpRT->waveHdr = malloc (sizeof(WAVEHDR) * wNumAudioRequested);
	if (lpRT->waveHdr == NULL) {
		return IDS_CAP_WAVE_ALLOC_ERROR;
	}

	/* clear it */
	memset (lpRT->waveHdr, 0, sizeof(WAVEHDR) * wNumAudioRequested);

	/* if we're capturing no file, we need an intermediate buffer */
	if (lpRT->bCreateNoFile) {

		memset (&lpRT->waveHdrNoFile, 0, sizeof(WAVEHDR));

		lpRT->waveHdrNoFile.dwFlags = WHDR_PREPARED |  WHDR_INQUEUE;
		lpRT->waveHdrNoFile.dwBufferLength  = ((INPUT_SIZE*1) * lpRT->WaveFormat.nBlockAlign) * lpHw->AudioRate/11025;

		lpRT->waveHdrNoFile.lpData = malloc (lpRT->waveHdrNoFile.dwBufferLength);
		if (lpRT->waveHdrNoFile.lpData == NULL) {
			return IDS_CAP_WAVE_ALLOC_ERROR;
		}
	}

	/* 
	 * if capturing, we skip the first gop to compensate for a blank 
	 * line/quality problem with the ccube microcode, so we skip a 
	 * coresponding amount of audio.
	 */
	if (lpRT->fCaptureVideo && RT_VIDEOSTARTCOUNT != 0) {

		DWORD dwMicroPerFrame = (lpHw->VideoMode == PAL) ? PAL_MICROSPERFRAME : NTSC_MICROSPERFRAME;
		double fDelay = (double)MulDiv(RT_VIDEOSTARTCOUNT, dwMicroPerFrame, 1000L);
		double fBytesPerSec = (double)lpRT->WaveFormat.nAvgBytesPerSec;

		/* 
		 * to synch up with the first gop, we're going to make the first buffer
		 * the size (in time) of the first gop, and then thow it out in the
		 * audio loop.
		 */
		lpRT->waveHdr[0].dwBufferLength = (DWORD)(fBytesPerSec*(fDelay/1000.0));
		
		/* make sure it's block aligned */
		lpRT->waveHdr[0].dwBufferLength = (lpRT->waveHdr[0].dwBufferLength/lpRT->WaveFormat.nBlockAlign) * lpRT->WaveFormat.nBlockAlign;

		/* 
		 * allocate the buffer to be the bigger of this size and the size required
		 * by the audio encoder.
		 */
		lpRT->waveHdr[0].lpData = malloc (max (lpRT->waveHdr[0].dwBufferLength,(((INPUT_SIZE*1) * lpRT->WaveFormat.nBlockAlign) * lpHw->AudioRate/11025)));
		if (lpRT->waveHdr[0].lpData == NULL) {
			return IDS_CAP_WAVE_ALLOC_ERROR;
		}

		lpHw->wNumAudioAllocated++;
		Index = 1;

	} else {
		Index = 0;
	}

	/* init the wave headers */
	for (;Index<wNumAudioRequested;Index++) {
		lpRT->waveHdr[Index].dwBufferLength  = (((INPUT_SIZE*1) * lpRT->WaveFormat.nBlockAlign) * lpHw->AudioRate/11025);

		lpRT->waveHdr[Index].lpData = malloc (lpRT->waveHdr[Index].dwBufferLength);
		if (lpRT->waveHdr[Index].lpData == NULL) {
			if ((Index + 1) < MIN_AUDIOBUFFERS) {
				return IDS_CAP_WAVE_ALLOC_ERROR;
			} else {
				break;
			}
		} 
		
		lpHw->wNumAudioAllocated++;
	}

	/* prepare the headers */
	for (Index=0;Index<lpHw->wNumAudioAllocated;Index++) {
		if (waveInPrepareHeader (lpRT->hWave, &lpRT->waveHdr[Index], sizeof(WAVEHDR)) != 0) {
			return IDS_CAP_WAVE_PREPARE_ERROR;
		}
	}

	/* add the buffers */
	for (Index=0;Index<lpHw->wNumAudioAllocated;Index++) {
		if (waveInAddBuffer (lpRT->hWave, &lpRT->waveHdr[Index], sizeof(WAVEHDR)) != 0) {
			return IDS_CAP_WAVE_ADD_ERROR;
		}
	}

	/* get the audio encoder stereo type */
	switch (lpRT->AudioMode) {

		case AUDIO_MONO:
			lpRT->StereoType = MONO;
			break;
		case AUDIO_JOINTSTEREO:
			lpRT->StereoType = JOINT_STEREO;
			break;
		case AUDIO_DUALCHANNEL:
			lpRT->StereoType = DUAL_CHANNEL;
			break;
		case AUDIO_STEREO:
			lpRT->StereoType = STEREO;
			break;
	}

	/* get the audio encoder sampling frequency */
	switch (lpRT->WaveFormat.nSamplesPerSec) {

		case AUDIO_RATE_11KHZ:
			lpRT->SamplingFreq = SF_11025_HZ;
			break;
		case AUDIO_RATE_22KHZ:
			lpRT->SamplingFreq = SF_22050_HZ;
			break;
		case AUDIO_RATE_44KHZ:
			lpRT->SamplingFreq = SF_44100_HZ;
			break;
	}

	/* allocate the audio encoder */
	lpRT->hAudioEncode = CreateAudioEncoder(lpRT->SamplingFreq, lpRT->StereoType, lpRT->AudioDataRate);
	if (lpRT->hAudioEncode == NULL) {
		return IDS_CAP_COMPRESSOR_ERROR;
	}
			
	/* set up input buffer size */
	lpRT->AudioBytesToRead = INPUT_SIZE * lpRT->WaveFormat.nBlockAlign;
			
	if (lpRT->WaveFormat.nSamplesPerSec == AUDIO_RATE_22KHZ) {
		lpRT->AudioBytesToRead /= 2;
	} else if (lpRT->WaveFormat.nSamplesPerSec == AUDIO_RATE_11KHZ) {
		lpRT->AudioBytesToRead /= 4;
	}

	if ((INPUT_SIZE * lpRT->WaveFormat.nBlockAlign) > lpRT->AudioBytesToRead) {
		lpRT->InBuf = malloc (INPUT_SIZE * lpRT->WaveFormat.nBlockAlign);
		if (lpRT->InBuf == NULL) {
			return IDS_CAP_WAVE_ALLOC_ERROR;
		}

		memset (lpRT->InBuf, 0, INPUT_SIZE * lpRT->WaveFormat.nBlockAlign);
	}

	return 0;
}


LRESULT 
RT_InitializeVideo (LPHWINFO lpHw)
{
    LPVC_CALLBACK lpCallback = VC_GetCallbackTable(lpHw->vh);
	DWORD wNumVideoRequested = REQUIRED_VIDEOBUFFERS;
	LPRTINFO lpRT = &lpHw->RTInfo;

	HWND hWnd = lpHw->hWndCap;
	DWORD Index;

	/* do we use preview??? */
	lpRT->bPreview = lpHw->bPreview | lpHw->bOverlay;

	/* allocate the rate control padding buffer */
	lpRT->lpPadBuffer = malloc (PAD_BUFFERSIZE);
	if (lpRT->lpPadBuffer == NULL) {
		return IDS_CAP_VIDEO_ALLOC_ERROR;
	}

	FillMemory (lpRT->lpPadBuffer, PAD_BUFFERSIZE, 0);

	/* allocate the videohdr array */
	lpRT->videoHdr = malloc (sizeof(VIDEOHDR) * wNumVideoRequested);
	if (lpRT->videoHdr == NULL) {
		return IDS_CAP_VIDEO_ALLOC_ERROR;
	}

	/* try to allocate a womping big circular buffer memory for video */
	lpRT->dwVideoLength = REQUIRED_VIDEOBUFFER_SIZE;
	lpRT->lpVideoBuf = NULL;

	/* go for the biggest we can get!!! */
	while (lpRT->lpVideoBuf == NULL) {
		
		lpRT->lpVideoBuf = malloc (lpRT->dwVideoLength);
		if (lpRT->lpVideoBuf == NULL) {
			
			lpRT->dwVideoLength -= min (VIDEOBUFFER_SIZE_PER_SEC, lpRT->dwVideoLength);
			wNumVideoRequested -= VIDEO_BUFFERS_PER_SECOND;

			if (lpRT->dwVideoLength < MIN_VIDEOBUFFER_SIZE) 
				break;
		}
	}

	/* did we make it? */
	if (lpRT->lpVideoBuf == NULL) {
		return IDS_CAP_VIDEO_ALLOC_ERROR;
	}

	/* initialize for use with real-time buffer functions */
	lpCallback->StreamRemoveReadyHeadFunc = RT_StreamRemoveReadyHead;
	lpCallback->StreamAddReadyTailFunc = RT_StreamAddReadyTail;
	lpCallback->StreamReplaceReadyHeadFunc = RT_StreamReplaceReadyHead;

	lpRT->dwVideoFilled = 0;
	lpRT->dwVideoUnused = 0;
	lpRT->dwVideoTail   = 0;

	/* init the video headers */
	for (Index=0;Index<wNumVideoRequested;Index++) {
		lpRT->videoHdr[Index].lpData	      = NULL;
		lpRT->videoHdr[Index].dwBufferLength  = lpHw->dwMaxCalcFrameSize*sizeof(DWORD);
		lpRT->videoHdr[Index].dwBytesUsed     = 0;
		lpRT->videoHdr[Index].dwTimeCaptured  = 0;
		lpRT->videoHdr[Index].dwUser          = 0;
		lpRT->videoHdr[Index].dwFlags         = VHDR_PREPARED;
		lpRT->videoHdr[Index].dwReserved[0]   = 0;
		lpRT->videoHdr[Index].dwReserved[1]   = 0;
		lpRT->videoHdr[Index].dwReserved[2]   = 0;
		lpRT->videoHdr[Index].dwReserved[3]   = 0;

		lpHw->wNumVideoAllocated++;

		VC_AddReadyTail(lpHw->vh, &lpRT->videoHdr[Index]);
	}

	/* setup for preview if needed */
	if (lpRT->bPreview) {

		lpRT->hDC = GetDC(hWnd);
		lpRT->hdd = DrawDibOpen();

		lpRT->DstWidth  = lpHw->CaptureInfo.FrameWidth;
		lpRT->DstHeight = lpHw->CaptureInfo.FrameHeight;

		lpRT->SrcWidth  = lpHw->CaptureInfo.FrameWidth;
		lpRT->SrcHeight = lpHw->CaptureInfo.FrameHeight;

		
	  	lpRT->biOut.biSize          = sizeof (BITMAPINFOHEADER);  
		lpRT->biOut.biPlanes		= 1; 
		lpRT->biOut.biBitCount 		= 24;
		lpRT->biOut.biCompression   = FOURCC_BW10;
		lpRT->biOut.biSizeImage     = lpHw->nSizeImage; 
		lpRT->biOut.biXPelsPerMeter = 0; 
		lpRT->biOut.biYPelsPerMeter = 0; 
		lpRT->biOut.biClrUsed       = 0;       
		lpRT->biOut.biClrImportant  = 0; 
		lpRT->biOut.biWidth			= lpRT->DstWidth; 
		lpRT->biOut.biHeight		= lpRT->DstHeight;
  
	    DrawDibBegin(lpRT->hdd
                     ,lpRT->hDC
                     ,lpRT->DstWidth
                     ,lpRT->DstHeight
                     ,&lpRT->biOut
                     ,lpRT->SrcWidth
                     ,lpRT->SrcHeight
                     ,DDF_SAME_HDC
                     );
	} 


	/* The next gop is the first */
	lpRT->bFirstGop = TRUE;

	return 0;
}


long 
RT_GetFreeDiskSpaceInKB(LPRTINFO lpRT, LPCTSTR pFile)
{
	DWORD dwFreeClusters, dwBytesPerSector, dwSectorsPerCluster, dwClusters;
    char RootName[MAX_PATH];
    LPSTR ptmp; 

	/* 
	 * need to find path for root directory on drive containing
     * this file.
	 */
    GetFullPathName(pFile, sizeof(RootName)/sizeof(RootName[0]), RootName, &ptmp);

    /* truncate this to the name of the root directory (god how tedious) */
    if ((RootName[0] == TEXT('\\')) && (RootName[1] == TEXT('\\'))) {

        /*
		 * path begins with  \\server\share\path so skip the first
         * three backslashes
		 */
        ptmp = &RootName[2];
        while (*ptmp && (*ptmp != TEXT('\\'))) {
            ptmp++;
        }
        
		if (*ptmp) {
            /* advance past the third backslash */
            ptmp++;
        }

    } else {
        /* path must be drv:\path */
        ptmp = RootName;
    }

    /* find next backslash and put a null after it */
    while (*ptmp && (*ptmp != TEXT('\\'))) {
        ptmp++;
    }

    /* found a backslash ? */
    if (*ptmp) {

        /* skip it and insert null */
        ptmp++;
        *ptmp = TEXT('\0');
    }

    if (!GetDiskFreeSpace(RootName,
			&dwSectorsPerCluster,
			&dwBytesPerSector,
			&dwFreeClusters,
			&dwClusters)) {

		if (lstrlen (lpRT->sDriveLetter) == 0)  { /* there is no drive selected need to map */
		 
			int drive, curdrive;
			DWORD dwRet;
			
			char szDrive[300];
			NETRESOURCE netres;
		
			/* Save current drive. */
			curdrive = _getdrive();

			/* If we can switch to the drive, it exists. */
			for( drive = 4; drive <= 26; drive++ ) {
				  
				if( _chdrive( drive ) != 0) {
					sprintf( szDrive, "%c:", drive + 'A');
					sprintf (lpRT->sDriveLetter, "%s", szDrive);
					break;
				 }
			}

			if (lstrlen (lpRT->sDriveLetter) == 0)
				return( -1 );   /* unable to find a free drive to map */

			netres.dwType = RESOURCETYPE_DISK;
			netres.lpLocalName = (LPSTR)(LPCSTR)lpRT->sDriveLetter;
			netres.lpRemoteName	= RootName;
			netres.lpProvider = NULL;
				
			dwRet = WNetAddConnection2(
			    &netres,	// points to structure that specifies connection details
			    NULL,	    // points to password string
			    NULL,	    // points to user name string
			    0 	        // set of bit flags that specify connection options
			);
			
			if (dwRet != NO_ERROR)	{

				/* unable to find a free drive to map */
				sprintf (lpRT->sDriveLetter, "");
				return( -1 );   
			}

		    if (!GetDiskFreeSpace(lpRT->sDriveLetter,
				&dwSectorsPerCluster,
				&dwBytesPerSector,
				&dwFreeClusters,
				&dwClusters)) 
		 	{
				WNetCancelConnection2(lpRT->sDriveLetter, 0, TRUE); // Disconnect.
				sprintf (lpRT->sDriveLetter, "");
				sprintf (lpRT->sRootName, "%s", RootName);
			    return (-1);
			}

			WNetCancelConnection2(lpRT->sDriveLetter, 0, TRUE); // Disconnect.

		} 
		
		/* we hopefully have info */
		sprintf (lpRT->sRootName, "%s", RootName);

	}
    
	return(MulDiv (dwSectorsPerCluster * dwBytesPerSector,
		   dwFreeClusters,
		   1024));
}


LRESULT 
RT_InitializeOutput (LPHWINFO lpHw)
{
	LPRTINFO lpRT = &lpHw->RTInfo;
	HWND hWnd = lpHw->hWndCap;
	long size;

	/* do we need to create a file??? */
	if (lpRT->bCreateNoFile) {
		return 0;
	}
	/* get the file name to capture to */
	capFileGetCaptureFile(hWnd, lpRT->szCaptureFile, sizeof(lpRT->szCaptureFile));

	/* 
	 * verify we can check disk space. this should always work on local drives. 
	 * the only time it might ever hose is on network drives.  If this is the case,
	 * we don't want to waste time trying to reconnect to the drive.
	 *
	 * NOTE: If the calling app is Broadway, we don't ever check for disk space.
	 * Broadway does this already, wy waste time duplicating efforts!
	 */
	if (lpHw->IsBroadway || lpRT->bCreateNoFile) {
		lpRT->CheckDiskSpace = FALSE;
	} else if (!lpHw->fCapturingNow) {
		size = RT_GetFreeDiskSpaceInKB(lpRT, lpRT->szCaptureFile);
		if (size != -1) {
			lpRT->CheckDiskSpace = TRUE;
		} else {
			lpRT->CheckDiskSpace = FALSE;
		}
	}

	/* create the video/stream output file */
	if (lpRT->fCaptureVideo) {
		char *ptr = lpRT->szCaptureFile;
		int len;

		/* make sure we have enough space for ext */
		lpRT->szCaptureFile[sizeof(lpRT->szCaptureFile)-5] = '\0';
		
		/* find extention */		
		len = lstrlen (ptr);
		ptr += len;
		while (*ptr != '.' && len != 0)	{
			ptr--;
			len--;
		}

		if (len==0) {
			if (lpHw->dwMaxFileSize != 0) {

				wsprintf (buffer, "%ld.mpg", lpRT->FileCount);
				lstrcat (lpRT->szCaptureFile, buffer);  			
			} else {
				lstrcat (lpRT->szCaptureFile, ".mpg");  			
			}
		} else {
			if (lpHw->dwMaxFileSize != 0) {
				wsprintf (ptr, "%ld.mpg", lpRT->FileCount);  			
			} else {
				wsprintf (ptr, ".mpg");  			
			}
		}
								  
		lpRT->OutFile = CreateFile(lpRT->szCaptureFile, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
                                FILE_ATTRIBUTE_NORMAL, NULL);

		if (lpRT->OutFile == INVALID_HANDLE_VALUE) {
			return IDS_CAP_FILE_OPEN_ERROR;
		}
	}

	/* stream mixing or seperate files??? */
	if (lpRT->bStreamMix) {
	
		/* create the stream mixer object */
		if (lpRT->StreamMixType == StandardMix) {
			lpRT->hStream = CreateStreamMixer();
		} else if (lpRT->StreamMixType == VideoCDMix) {
			lpRT->hStream = CreateVideoCDMixer();
		}

		if (lpRT->hStream == NULL) {
			return IDS_CAP_COMPRESSOR_ERROR;
		}

		/* initialize the stream object */
		if (InitializeStream(lpRT->hStream
							 ,lpRT->OutFile
							 ,lpRT->PackSize
							 ,lpRT->PacketsPerPack
							 ,lpRT->StreamDataRate
							 ,lpRT->AudioDelay, 3072, 0.0
							 )) {

			return IDS_CAP_COMPRESSOR_ERROR;
		}

		/* get the audio and video pins */
		lpRT->hVideoPin = GetStreamVideoPin(lpRT->hStream); 
		if (lpRT->hVideoPin  == NULL) {
			return IDS_CAP_COMPRESSOR_ERROR;
		}

		lpRT->hAudioPin = GetStreamAudioPin(lpRT->hStream); 
		if (lpRT->hAudioPin  == NULL) {
			return IDS_CAP_COMPRESSOR_ERROR;
		}
	} else if (lpRT->fCaptureAudio) {
	
		char *ptr = lpRT->szAudioCaptureFile;
		int len;

		/* set audio file name = base_name.mpa */
		lstrcpyn(lpRT->szAudioCaptureFile, lpRT->szCaptureFile, sizeof(lpRT->szAudioCaptureFile)-1);

		/* make sure we have enough space for ext */
		lpRT->szAudioCaptureFile[sizeof(lpRT->szAudioCaptureFile)-5] = '\0';
		
		/* find extention */		
		len = lstrlen (ptr);
		ptr += len;
		while (*ptr != '.' && len != 0)	{
			ptr--;
			len--;
		}

		if (len==0) {
			if (lpHw->dwMaxFileSize != 0) {
				wsprintf (buffer, "%ld.mpa", lpRT->FileCount);
				lstrcat (lpRT->szAudioCaptureFile, buffer);  			
			} else {
				lstrcat (lpRT->szAudioCaptureFile, ".mpa");  			
			}
		} else {
			if (lpHw->dwMaxFileSize != 0) {
				wsprintf (ptr, "%ld.mpa", lpRT->FileCount);  			
			} else {
				wsprintf (ptr, ".mpa");  			
			}
		}

		/* open the audio file */
		lpRT->AudioFile = CreateFile(lpRT->szAudioCaptureFile, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
		                            FILE_ATTRIBUTE_NORMAL, NULL);

		if (lpRT->AudioFile == INVALID_HANDLE_VALUE) {
			return IDS_CAP_FILE_OPEN_ERROR;
		}

		/* allocate an output buffer */
		lpRT->OutData = malloc (GetMaxAudioFrameSize(lpRT->hAudioEncode));
		if (lpRT->OutData == NULL) {
			return IDS_CAP_VIDEO_ALLOC_ERROR;
		}
	}

	if (lpHw->dwMaxFileSize != 0) {
		lpRT->FileCount++;  			
	} 

	return 0;
}


LRESULT 
RT_SplitFile (LPHWINFO lpHw) 
{
	LPRTINFO lpRT = &lpHw->RTInfo;
	DWORD dwWaveElapsedMS = (DWORD)((double)lpHw->dwCurrentWaveSamples/(double)lpRT->WaveFormat.nSamplesPerSec * 1000.0);

	/* flush and free the stream mixer */
	if (lpRT->hStream != NULL) {
		
		FlushStreamData(lpRT->hStream);
		DeleteStreamMixer(lpRT->hStream);
		
		lpRT->hStream   = NULL;
		lpRT->hAudioPin = NULL;
		lpRT->hVideoPin = NULL;
		lpRT->AudioFile = NULL;
	}

	/* close the output files */
	if (lpRT->OutFile != NULL && lpRT->OutFile != INVALID_HANDLE_VALUE) {
		CloseHandle (lpRT->OutFile);
		lpRT->OutFile = NULL;
	}
	if (lpRT->AudioFile != NULL && lpRT->AudioFile != INVALID_HANDLE_VALUE) {
		CloseHandle (lpRT->AudioFile);
		lpRT->AudioFile = NULL;
	}

	/* free output buffer */
	if (lpRT->OutData != NULL) {
		free (lpRT->OutData);
		lpRT->OutData = NULL;
	}

	/* reset some stuff */
	lpRT->dwCurrentBytes  = 0L;
	lpRT->tc.dword		  = 0L;
	lpRT->PFrameCount	  = 0L;
	lpRT->bLastFrame	  = FALSE;
	lpRT->StopCodeWritten = FALSE;
	lpRT->bGotGoodStop	  = FALSE;
	lpRT->bGotIFrame	  = FALSE;
	lpRT->bSplitFile      = FALSE;

	/* resynch audio to video!!! */
	if (dwWaveElapsedMS != lpHw->dwCurrentTimeElapsedMS) {
		int nCorrection = dwWaveElapsedMS - lpHw->dwCurrentTimeElapsedMS;
		
		lpRT->AudioDelay = ((double)(double)nCorrection/1000.0);
	}

	return RT_InitializeOutput (lpHw);
}


LRESULT 
RT_StartCapture (LPHWINFO lpHw) 
{
	LPRTINFO lpRT = &lpHw->RTInfo;
	HWND hWnd = lpHw->hWndCap;
	VCCALLBACK Callback;

	/* reset the cursor */
	if (lpRT->OldCursor != NULL) {
		SetCursor(lpRT->OldCursor);
		lpRT->OldCursor = NULL;
	}

	/* was hit ok to capture specified??? */
	if (lpRT->CapParams.fMakeUserHitOKToCapture) {

		RT_LoadAviCapString(lpHw, IDS_CAP_SEQ_MSGSTART, buffer, sizeof(buffer));
		
		if (lpRT->fCaptureVideo) {
			wsprintf (message, buffer, lpRT->szCaptureFile);
		} else {
			wsprintf (message, buffer, lpRT->szAudioCaptureFile);
		}

		if (MessageBox (hWnd, message, "", MB_OKCANCEL|MB_ICONEXCLAMATION) != IDOK) {
			lpRT->bCaptureOK = FALSE;
			return IDS_CAP_END;
		}
	}

	/* does the user really want to start??? */
	if (lpHw->lpfnCapControlCallback) {
		LRESULT lResult;
		
		lResult = lpHw->lpfnCapControlCallback (hWnd, CONTROLCALLBACK_PREROLL);
		if (!lResult) {
			lpRT->bCaptureOK = FALSE;
			return IDS_CAP_END;
		}
	}

	/* if not yielding, display a wait cursor */
	if (!lpRT->CapParams.fYield) {
		lpRT->OldCursor = SetCursor(LoadCursor(NULL,MAKEINTRESOURCE(IDC_WAIT)));
	}

	/* turn preview off - we'll handle it!!! */
	if (lpRT->fCaptureVideo) {

		if (!lpHw->bOverlay) {
			capPreview (hWnd, FALSE);
		}
	} else if (lpHw->bPreview) {
		
		/* if capturing audio only, slow down preview */
		capPreviewRate(hWnd, 333);
	} else if (lpHw->bOverlay) {

		/* if capturing audio only, slow down overlay */
	    VO_NotifyStreamStart (lpHw->voh, TRUE);
	}

	/* initialize video capture */
	if (lpRT->fCaptureVideo) {
	
		/* no callback required */
		Callback.dwCallback = 0;
		Callback.dwFlags = 0;
		Callback.hDevice = 0;
		Callback.dwUser =  0;

		/* init video capture */
		if (!HW_StreamInit(lpHw->vh, (lpHw->VideoMode == PAL) ? PAL_MICROSPERFRAME : NTSC_MICROSPERFRAME)) {
			return IDS_CAP_AVI_INIT_ERROR;
		}
	}

	/* compensate for any audio delay startup... */
	if (lpRT->fCaptureAudio && lpRT->fCaptureVideo) {

		DWORD dwEndTime;

		/* compensate for the c-cube pipeline delay and audio delay */
		int nDelay = lpHw->nAudioDelay;

		/* compensate for the c-cube pipeline delay and audio delay */
		if (nDelay >= 0) {

			/* start the audio */
			if (waveInStart (lpRT->hWave) != 0) {
				return IDS_CAP_AVI_INIT_ERROR;
			}

			dwEndTime = timeGetTime() + (DWORD)nDelay;
			while (timeGetTime() < dwEndTime);

			/* start the video */
			if (!HW_StreamStart(lpHw->vh)) {
				return IDS_CAP_AVI_INIT_ERROR;
			}
		} else {

			/* start the video */
			if (!HW_StreamStart(lpHw->vh)) {
				return IDS_CAP_AVI_INIT_ERROR;
			}

			dwEndTime = timeGetTime() + (DWORD)abs(nDelay);
			while (timeGetTime() < dwEndTime);

			/* start the audio */
			if (waveInStart (lpRT->hWave) != 0) {
				return IDS_CAP_AVI_INIT_ERROR;
			}
		}

		/* used for infinite synch */
		lpRT->dwPrevious = pcGetTime();

	} else {

		/* start the audio if required */
		if (lpRT->fCaptureAudio) {
			if (waveInStart (lpRT->hWave) != 0) {
				return IDS_CAP_AVI_INIT_ERROR;
			}
		}

		/* start the video if required */
		if (lpRT->fCaptureVideo) {
			if (!HW_StreamStart(lpHw->vh)) {
				return IDS_CAP_AVI_INIT_ERROR;
			}
		}
	}

	/* inform app we've started */
	if (lpHw->lpfnStatusCallback) {
		
		RT_LoadAviCapString(lpHw, IDS_CAP_BEGIN, message, sizeof(message));
		lpHw->lpfnStatusCallback(hWnd, IDS_CAP_BEGIN, message);

		RT_LoadAviCapString(lpHw, IDS_CAP_SEQ_MSGSTOP, message, sizeof(message));
		lpHw->lpfnStatusCallback(hWnd, IDS_CAP_SEQ_MSGSTOP, message);
					
		if (lpRT->fCaptureVideo) {
			RT_LoadAviCapString(lpHw, IDS_CAP_STAT_VIDEOCURRENT, buffer, sizeof(buffer));
			wsprintf (message, buffer, 0, 0, 0, 0);

			lpHw->lpfnStatusCallback(hWnd, IDS_CAP_STAT_VIDEOCURRENT, message);
		} else {
			RT_LoadAviCapString(lpHw, IDS_RT_STAT_AUDIOCURRENT, buffer, sizeof(buffer));
			wsprintf (message, buffer, 0, 0, 0);

			lpHw->lpfnStatusCallback(hWnd, IDS_RT_STAT_AUDIOCURRENT, message);
		}
	}

	/* give us a priority boost for capture efficiency */
	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST);

	/* we have begun!!! */
	lpHw->fCapturingNow = TRUE; 
	
	return 0;
}


LRESULT 
RT_StopCapture (LPHWINFO lpHw) 
{
	LPRTINFO lpRT = &lpHw->RTInfo;
    LPVC_CALLBACK lpCallback = VC_GetCallbackTable(lpHw->vh);
	HWND hWnd = lpHw->hWndCap;
	DWORD Index, ErrorCode;

	/* 
	 * we've started stopping!!!
	 */
	lpHw->bIsStopping = TRUE;

	/* all done - lower thread priority */
	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL);
	ResetEvent (lpHw->hCaptureComplete);

	/* unhook our hooks */
	if (lpRT->hHookM != NULL) {	
		UnhookWindowsHookEx(lpRT->hHookM);
	}
	if (lpRT->hHookK != NULL) {	
		UnhookWindowsHookEx(lpRT->hHookK);
	}
	if (lpRT->hCaptureStop) {
		CloseHandle (lpRT->hCaptureStop);	
	}

	/* stop audio capture */
	if (lpRT->fCaptureAudio && lpRT->hWave != NULL) {

		/* reset wave in */
		waveInReset (lpRT->hWave);
	}

	/* stop video capture */
	if (lpRT->fCaptureVideo) {
		
		HW_StreamStop(lpHw->vh);

		VC_ResetDoneQueue(lpHw->vh);
		VC_ResetReadyQueue(lpHw->vh);

	}

	/* don't use realtime buffer routines any more */
	lpCallback->StreamRemoveReadyHeadFunc = NULL;
	lpCallback->StreamAddReadyTailFunc = NULL;
	lpCallback->StreamReplaceReadyHeadFunc = NULL;

	/* flush and free the stream mixer */
	if (lpRT->hStream != NULL) {
		FlushStreamData(lpRT->hStream);
		DeleteStreamMixer(lpRT->hStream);
	}

	/* close the output files */
	if (lpRT->OutFile != NULL && lpRT->OutFile != INVALID_HANDLE_VALUE) {
		CloseHandle (lpRT->OutFile);
	}
	if (lpRT->AudioFile != NULL && lpRT->AudioFile != INVALID_HANDLE_VALUE) {
		CloseHandle (lpRT->AudioFile);
	}

	/* free the audio encoder */
	if (lpRT->hAudioEncode != NULL) {
		DeleteAudioEncoder(lpRT->hAudioEncode);
	}

	/* free input and output buffers */
	if (lpRT->InBuf != NULL) {
		free (lpRT->InBuf);
	}
	if (lpRT->OutData != NULL) {
		free (lpRT->OutData);
	}

	/* free the wave headers */
	if (lpRT->waveHdr != NULL) {

		/* unprepare and free the wave buffers */
		for (Index=0;Index<lpHw->wNumAudioAllocated;Index++) {
			if (lpRT->waveHdr[Index].dwFlags & WHDR_PREPARED)	{
				waveInUnprepareHeader (lpRT->hWave, &lpRT->waveHdr[Index], sizeof(WAVEHDR));                      
			}
		}

		free (lpRT->waveHdr);
	}

	if (lpRT->lpVideoBuf != NULL) {
		free (lpRT->lpVideoBuf);
	}

	if (lpRT->lpPadBuffer != NULL) {
		free (lpRT->lpPadBuffer);
	}
	
	if (lpRT->waveHdrNoFile.lpData != NULL) {
		free (lpRT->waveHdrNoFile.lpData); 
	}

	/* close up the audio device */
    waveInClose (lpRT->hWave);

	/* free the video headers */
	if (lpRT->videoHdr != NULL) {
		free (lpRT->videoHdr);
	}

	/* close up the video stream */
	if (lpRT->fCaptureVideo) {
		HW_StreamFini (lpHw->vh);
	}

	/* reset the cursor */
	if (lpRT->OldCursor != NULL) {
		SetCursor(lpRT->OldCursor);
	}

	/* clean up the drawdib stuff */
	if (lpRT->hdd != NULL) {
		DrawDibEnd(lpRT->hdd);
		DrawDibClose(lpRT->hdd );
	}
	if (lpRT->hDC != NULL) {
		ReleaseDC (hWnd, lpRT->hDC);
	}

	/* restore preview */
	if (lpRT->fCaptureVideo) {
		if (lpRT->bPreview && !lpHw->bOverlay) {
			capPreview (hWnd, TRUE);
			lpHw->bPreview = TRUE;
		}
	} else if (lpHw->bPreview) {
		
		/* if capturing audio only, restore preview */
		capPreviewRate(hWnd, 15);
	} else if (lpHw->bOverlay) {

		/* if capturing audio only, restore overlay */
	    VO_NotifyStreamStart (lpHw->voh, FALSE);
	}

	/* 
	 * at this point we've actually stopped stopping, we just have to let the user
	 * display any status/errors messages...
	 */
	lpHw->bIsStopping = FALSE;

	/* report final capture status */
	if (lpRT->bCaptureOK && lpHw->lpfnStatusCallback) {
		
		if (lpRT->fCaptureAudio && lpRT->fCaptureVideo) {

			DWORD sec, ms, cf, df, vrh, vrl, wb, wrh, wrl;

			RT_LoadAviCapString(lpHw, IDS_CAP_STAT_VIDEOAUDIO, buffer, sizeof(buffer));
						
			sec = lpHw->dwCurrentTimeElapsedMS/1000;
			ms  = lpHw->dwCurrentTimeElapsedMS - (sec*1000);
			cf  = lpHw->dwCurrentVideoFrame;
			df  = lpHw->dwCurrentVideoFramesDropped;
			
			if (lpHw->VideoMode == PAL)	 {
				vrh = 25; vrl = 000;
			} else {
				vrh = 29; vrl = 970;
			}

			wb  = lpHw->dwCurrentWaveSamples * lpRT->WaveFormat.nBlockAlign;
			wrh = lpRT->WaveFormat.nSamplesPerSec/1000;
			wrl = lpRT->WaveFormat.nSamplesPerSec- (wrh * 1000);

			wsprintf (message, buffer, sec, ms, cf, df, vrh, vrl, wb, wrh, wrl);
			
			lpHw->lpfnStatusCallback(hWnd, IDS_CAP_STAT_VIDEOAUDIO, message);

		} else if (!lpRT->fCaptureAudio && lpRT->fCaptureVideo) {

			DWORD sec, ms, cf, df, vrh, vrl;

			RT_LoadAviCapString(lpHw, IDS_CAP_STAT_VIDEOONLY, buffer, sizeof(buffer));
						
			sec = lpHw->dwCurrentTimeElapsedMS/1000;
			ms  = lpHw->dwCurrentTimeElapsedMS - (sec*1000);
			cf  = lpHw->dwCurrentVideoFrame;
			df  = lpHw->dwCurrentVideoFramesDropped;
			
			if (lpHw->VideoMode == PAL)	 {
				vrh = 25; vrl = 000;
			} else {
				vrh = 29; vrl = 970;
			}

			wsprintf (message, buffer, sec, ms, cf, df, vrh, vrl);

			lpHw->lpfnStatusCallback(hWnd, IDS_CAP_STAT_VIDEOONLY, message);
		} else {

			DWORD sec, ms, wb, wrh, wrl;

			RT_LoadAviCapString(lpHw, IDS_RT_STAT_AUDIOONLY, buffer, sizeof(buffer));
						
			sec = lpHw->dwCurrentTimeElapsedMS/1000;
			ms  = lpHw->dwCurrentTimeElapsedMS - (sec*1000);
			wb  = lpHw->dwCurrentWaveSamples * lpRT->WaveFormat.nBlockAlign;
			wrh = lpRT->WaveFormat.nSamplesPerSec/1000;
			wrl = lpRT->WaveFormat.nSamplesPerSec- (wrh * 1000);

			wsprintf (message, buffer, sec, ms, wb, wrh, wrl);
			
			lpHw->lpfnStatusCallback(hWnd, IDS_RT_STAT_AUDIOONLY, message);
		}
	}

	if (lpRT->ErrorCode) {
	
		if (lpHw->lpfnErrorCallback) {
			RT_LoadAviCapString(lpHw, lpRT->ErrorCode, message, sizeof(message));
			
			lpHw->lpfnErrorCallback(hWnd, lpRT->ErrorCode, message);
		}
	}

	/* inform app we've stopped */
	if (lpRT->bCaptureOK && lpHw->lpfnStatusCallback) {
		RT_LoadAviCapString(lpHw, IDS_CAP_END, message, sizeof(message));
		lpHw->lpfnStatusCallback(hWnd, IDS_CAP_END, message);
	}

	ErrorCode = lpRT->ErrorCode;

	/* all done capturing!!! */	
	lpHw->hCaptureThread = NULL;
	lpHw->fCapturingNow = FALSE;

	return ErrorCode ?FALSE:TRUE;
}

	  
LRESULT 
RT_ProcessAudio (LPHWINFO lpHw) 
{
	LPRTINFO lpRT = &lpHw->RTInfo;
	HWND hWnd = lpHw->hWndCap;
	int OutSize;
	short *InData;
	SBuff *buff;

	/* is the current audio buffer done??? */
	while (lpRT->waveHdr[lpRT->AudioReadyIndex].dwFlags & WHDR_DONE) {

		/* are we about to drop a buffer due to lack of real time??? */
		if (lpRT->AudioDoneCount >= (lpHw->wNumAudioAllocated -1)) {
			lpHw->NoBuffers++;
			return IDS_CAP_AUDIO_DROP_ERROR;
		}

	   /* 
	    * if capturing, we skip the first gop to compensate for a blank 
		* line/quality problem with the ccube microcode, so we skip a 
		* corespondin amount of audio.
		*/
		if (lpRT->fCaptureVideo && lpRT->AudioStartCount < RT_AUDIOSTARTCOUNT)	{		
			LPWAVEHDR  lpwh = &lpRT->waveHdr[lpRT->AudioReadyIndex];

			/* all done with this one - set the size back to the standard size */
			lpwh->dwBufferLength = lpRT->waveHdr[1].dwBufferLength;
			lpwh->dwBytesRecorded = 0;
			lpwh->dwFlags &= ~WHDR_DONE;

			/* put the buffer back to the driver */
			if (waveInAddBuffer (lpRT->hWave, lpwh, sizeof(WAVEHDR)) != 0) {
				return IDS_CAP_WAVE_ADD_ERROR;
			}			  

			lpRT->AudioStartCount++;
			lpRT->AudioDoneIndex = (lpRT->AudioDoneIndex + 1) % lpHw->wNumAudioAllocated;
		
		} else {
			lpRT->AudioDoneCount++;					 
		}


		lpRT->AudioReadyIndex = (lpRT->AudioReadyIndex + 1) % lpHw->wNumAudioAllocated;
	}
			
	/* 
	 * infinite calibration, baby!!! well almost - we'll modify the 
	 * calibration factor every 5 minutes and the longest we'll modify
	 * the calibration factor for is 5 hours - to avoid rollover.
	 */
	if (lpRT->fCaptureVideo) {
		
		MMTIME mmTime;
		DWORD dwTime = pcGetTime();
		DWORD dwSpan = dwTime - lpRT->dwPrevious;

		mmTime.wType = TIME_BYTES;
		if (lpRT->cbStart == 0) {
		
			if (dwTime > 100) {
				if (waveInGetPosition (lpRT->hWave, &mmTime, sizeof(MMTIME)) == 0) {
					lpRT->cbStart = mmTime.u.cb;
					lpRT->dwStartTime = lpRT->dwPrevious = dwTime;
				}
	
				lpRT->dwPrevious = dwTime;
			}
		
		} else if (dwSpan >= 300000 && dwTime < 18000000) {
			
			if (waveInGetPosition (lpRT->hWave, &mmTime, sizeof(MMTIME)) == 0) {
				
				DWORD cbLast = mmTime.u.cb;

				double dfCalc = ((double)(((double)(cbLast-lpRT->cbStart)/(double)lpRT->WaveFormat.nBlockAlign)/(double) (dwTime-lpRT->dwStartTime))*(double)1000.0);
				double dfDiff = dfCalc/(double)(lpRT->WaveFormat.nSamplesPerSec);

				VC_TimerDisable (lpHw->vh);
				lpHw->dfAudioFudgeFactor = ((lpHw->dfAudioFudgeFactor * 10.0) + dfDiff) / 11.0;
				VC_TimerEnable (lpHw->vh);
			}

			lpRT->dwPrevious = dwTime;
		}
	}	
	
	/* try to get an audio buffer to process */
	if (lpRT->lpProcessWaveHdr == NULL && lpRT->AudioDoneCount > 0) { 
		lpRT->lpProcessWaveHdr = &lpRT->waveHdr[lpRT->AudioDoneIndex];
		lpRT->AudioDoneIndex = (lpRT->AudioDoneIndex + 1) % lpHw->wNumAudioAllocated;

		/* 
		 * see if there's enough disk space avalable for this buffer plus 
		 * some headway (twice the buffer size just to be safe!!!) 
		 */
		if (lpRT->CheckDiskSpace) {
		
			int size = RT_GetFreeDiskSpaceInKB(lpRT, lpRT->szCaptureFile);
			
			if (size == -1) {
				lpRT->CheckDiskSpace = FALSE;
			} else if (size <= (long)((lpRT->lpProcessWaveHdr->dwBufferLength * 5)/1024)) {
				return IDS_CAP_FILE_WRITE_ERROR;
			}
		}
	}

	/* do we have an audio buffer to process??? */
	if (lpRT->lpProcessWaveHdr != NULL && (lpRT->hAudioPin == NULL || PinBufferReady(lpRT->hAudioPin))) {

		OutSize = GetNextAudioFrameSize(lpRT->hAudioEncode);
		InData = (short *)(lpRT->lpProcessWaveHdr->lpData + lpRT->AudioEncodeIndex);
				
		if (lpRT->InBuf != NULL) {
			CopyMemory (lpRT->InBuf, InData, lpRT->AudioBytesToRead);
			InData = (short*)lpRT->InBuf;
		}
		
		/* are we in no file mode??? */
		if (lpRT->bCreateNoFile) {

			/* encode the audio data */
			lpRT->MaxSB = 32;
			ProcessFrame(lpRT->hAudioEncode, TRUE, &lpRT->MaxSB, InData, lpRT->waveHdrNoFile.lpData + lpRT->AudioByteCount);

		} else if (lpRT->bStreamMix) {

			buff = GetPinBuffer(lpRT->hAudioPin);

			/* encode the audio data */
			lpRT->MaxSB = 32;
			ProcessFrame(lpRT->hAudioEncode, TRUE, &lpRT->MaxSB, InData, buff->data);
					
			/* send it to the stream mixer */
			buff->ValidBytes = OutSize;
			PutPinBuffer(lpRT->hAudioPin, buff);
		} else {
		    unsigned long Size;

			/* encode the audio data */
			lpRT->MaxSB = 32;
			ProcessFrame(lpRT->hAudioEncode, TRUE, &lpRT->MaxSB, InData, lpRT->OutData);

			/* write it to the file */
			WriteFile(lpRT->AudioFile, lpRT->OutData, OutSize, &Size, NULL);
			if ((int)Size != OutSize) {
				return IDS_CAP_FILE_WRITE_ERROR;
			}
		}

		/* update the file size */
		lpHw->dwCurrentWaveSamples += (lpRT->AudioBytesToRead/lpRT->WaveFormat.nBlockAlign); 
		lpRT->dwCurrentBytes += OutSize;
		lpRT->AudioByteCount += OutSize;

		/* are we done processing this audio buffer??? */
		lpRT->AudioEncodeIndex += lpRT->AudioBytesToRead;
		if (lpRT->AudioEncodeIndex == lpRT->lpProcessWaveHdr->dwBufferLength) {
				
			/* update the status */						
			if (!lpRT->fCaptureVideo) {

				lpHw->dwCurrentTimeElapsedMS = (DWORD)((double)lpHw->dwCurrentWaveSamples/(double)lpRT->WaveFormat.nSamplesPerSec * 1000.0);

				/* update user status */
				if (lpHw->lpfnStatusCallback && lpRT->AudioDoneCount <= MAX_AUDIO_DONE) {

					static DWORD cb, sec, ms;
						
					RT_LoadAviCapString(lpHw, IDS_RT_STAT_AUDIOCURRENT, buffer, sizeof(buffer));
							
					cb   = lpHw->dwCurrentWaveSamples * lpRT->WaveFormat.nBlockAlign;
					sec  = lpHw->dwCurrentTimeElapsedMS/1000;
					ms   = lpHw->dwCurrentTimeElapsedMS - (sec*1000);

					wsprintf (message, buffer, cb, sec, ms);
				
					lpHw->lpfnStatusCallback(hWnd, IDS_RT_STAT_AUDIOCURRENT, message);
				}
			}
					
			/* does user wish to continue capture??? */
			if (lpHw->lpfnCapControlCallback) {	

				LRESULT lResult;

				lResult = lpHw->lpfnCapControlCallback (hWnd, CONTROLCALLBACK_CAPTURING);
				if (!lResult) {
					return IDS_CAP_END;
				}
			}

			/* yield callback is called at least once per frame */
			if (lpHw->lpfnYieldCallback) {
				lpHw->lpfnYieldCallback (hWnd);
			}

			/* 
			 * we're mpeg stream based, not frame based, so if the user 
			 * doesn't want the data written to disk, and we haven't 
			 * been set to no file mode we just quit!!!
			 */
			if (lpHw->lpfnWaveStreamCallback) {
			
				WAVEHDR *lpWaveHdr;
				LRESULT lResult;


				if (lpRT->bCreateNoFile) {
					lpWaveHdr = &lpRT->waveHdrNoFile;	

					lpWaveHdr->dwFlags |= WHDR_DONE;
					lpWaveHdr->dwFlags &= ~WHDR_INQUEUE;

				} else {
					lpWaveHdr = lpRT->lpProcessWaveHdr;
				}

				lpWaveHdr->dwBytesRecorded = lpRT->AudioByteCount;

				lResult = lpHw->lpfnWaveStreamCallback (hWnd, lpWaveHdr);
				
				if (!lResult && !lpRT->bCreateNoFile) {
					if (RT_StopLoop (lpHw)) {
						return IDS_CAP_END;
					}
				}

				if (lpRT->bCreateNoFile) {
					lpWaveHdr->dwFlags &= ~WHDR_DONE;
					lpWaveHdr->dwFlags |= WHDR_INQUEUE;
				} 
			}

			/* all done with this one */
			lpRT->lpProcessWaveHdr->dwBytesRecorded = 0;
			lpRT->lpProcessWaveHdr->dwFlags &= ~WHDR_DONE;

			/* put the buffer back to the driver */
			if (waveInAddBuffer (lpRT->hWave, lpRT->lpProcessWaveHdr, sizeof(WAVEHDR)) != 0) {
				return IDS_CAP_WAVE_ADD_ERROR;
			}			  

			lpRT->lpProcessWaveHdr = NULL;
			lpRT->AudioEncodeIndex = 0;
			lpRT->AudioByteCount = 0;
			lpRT->AudioDoneCount--;

			/* 
			 * check if max file size is exceeded...
			 */
			return RT_CheckFileSize (lpHw);
		}
	}

	return 0;
}


LRESULT
RT_ProcessVideo (LPHWINFO lpHw) 
{
	LPRTINFO lpRT = &lpHw->RTInfo;
	HWND hWnd = lpHw->hWndCap;
	int OutSize;
	short *InData;
	SBuff *buff;
	
	/* 
	 * did we drop any frames in hardware that we can't compensate for??? 
	 */
	if (!lpRT->bStop && (lpHw->TimerOverRun+lpHw->FrameTooBig+lpHw->CopyOverRun+lpHw->NoBuffers+lpHw->dwCaptureStopped) != 0) {

		if (lpHw->dwCurrentVideoFrame != 0) {
			lpHw->dwCurrentVideoFramesDropped++;
		}
		
		return IDS_RT_VIDEO_DROP_ERROR;
	}

	/* is the video current buffer done??? */
	while (lpRT->videoHdr[lpRT->VideoReadyIndex].dwFlags & VHDR_DONE) {
		
		LPVIDEOHDR lpvh = &lpRT->videoHdr[lpRT->VideoReadyIndex];	

		/* are we about to drop a frame due to lack of real time??? */
		if (lpRT->VideoDoneCount >= (lpHw->wNumVideoAllocated - 1)) {
			
			if (lpHw->dwCurrentVideoFrame != 0) {
				lpHw->dwCurrentVideoFramesDropped++;
			}

			lpHw->NoBuffers++;
			return IDS_RT_VIDEO_DROP_ERROR;
		}

		/* 
		 * we skip the first gop to compensate for a blank line/quality problem
		 * with the ccube microcode
		 */
		if (lpRT->VideoStartCount < RT_VIDEOSTARTCOUNT)	{		
				
			/* all done with this one */
			lpvh->dwBytesUsed = 0;
			lpvh->dwFlags &= ~WHDR_DONE;
			lpvh->dwFlags &= ~VHDR_INQUEUE;

			/* put the buffer back to the capture driver */
			VC_AddReadyTail(lpHw->vh, lpvh);

			lpRT->VideoDoneIndex = (lpRT->VideoDoneIndex + 1) % lpHw->wNumVideoAllocated;
			lpRT->VideoStartCount++;
		
		} else {
			lpRT->VideoDoneCount++;
		}
		
		lpRT->VideoReadyIndex = (lpRT->VideoReadyIndex + 1) % lpHw->wNumVideoAllocated;
	}

	/* try to get a buffer to process */
	if (lpRT->lpProcessVideoHdr == NULL && lpRT->VideoDoneCount > 0) { 

		if (!RT_HaveFullGop (lpHw)) {
			return 0;
		}
		
		lpRT->lpProcessVideoHdr = &lpRT->videoHdr[lpRT->VideoDoneIndex];
		lpRT->VideoDoneIndex = (lpRT->VideoDoneIndex + 1) % lpHw->wNumVideoAllocated;

		/* recalculate the vbv_delay */
		RT_CalcVBVDelay (lpHw, lpRT->lpProcessVideoHdr);

		/* insert correct time code */
		if (lpRT->lpProcessVideoHdr->dwFlags & VHDR_KEYFRAME) {
			RT_SetTimeCode (lpHw, lpRT->lpProcessVideoHdr->lpData, lpRT->lpProcessVideoHdr->dwBytesUsed);
		}

		/* if flushing to gop end decrement new start countdown */
		if (lpRT->bFlushToLastGOP && lpRT->dwNewStartCountDown != 0) {
			lpRT->dwNewStartCountDown--;
		}

		/* decrement gop frame count, increment total frame count */
		lpRT->dwGopFrameCount--;
		lpRT->FrameCount++;

		/* 
		 * see if there's enough disk space avalable for this buffer plus 
		 * some headway just to be safe!!!) 
		 */
		if (lpRT->CheckDiskSpace) {
			
			long size = RT_GetFreeDiskSpaceInKB(lpRT, lpRT->szCaptureFile);
			
			if (size == -1) {
				lpRT->CheckDiskSpace = FALSE;
			} else if (size <= (long)((lpRT->lpProcessVideoHdr->dwBytesUsed * 5)/1024)) {
				return IDS_CAP_FILE_WRITE_ERROR;
			}
		}
		
		/* only draw a frame if we have time */
		if (lpRT->VideoDoneCount <= MAX_VIDEO_DONE && lpRT->AudioDoneCount <= MAX_AUDIO_DONE) {
			
			/* preview??? */
			if (lpRT->bPreview && (lpRT->lpProcessVideoHdr->dwFlags & VHDR_KEYFRAME)) {
				DWORD Width, Height;
				 
				/* are we scaleing preview??? */
				if (lpRT->CapStatus.fScale) {
					RECT rect;
	
					GetWindowRect (hWnd, &rect);
							
					Width = rect.right-rect.left; 
					Height = rect.bottom-rect.top;
			
				} else {
	
					Width = lpRT->SrcWidth; 
					Height = lpRT->SrcHeight;
				}

				lpRT->biOut.biSizeImage = lpRT->lpProcessVideoHdr->dwBytesUsed;

				/* display it!!! */
				DrawDibDraw(lpRT->hdd
   					        ,lpRT->hDC     
   						    ,0
    						,0
		    		        ,Width
			    		    ,Height
							,&lpRT->biOut
   							,lpRT->lpProcessVideoHdr->lpData
   							,0
							,0
							,lpRT->SrcWidth
							,lpRT->SrcHeight
							,DDF_SAME_HDC
							);

			} else if (lpHw->bOverlay) {
				VO_DrawFrameNow (lpHw->voh);						
			}
			
		} 

		/* we need to stop with the last frame */
		if ((lpRT->bSplitFile || lpRT->bStop) && !lpRT->bLastFrame) {
			lpRT->bLastFrame = RT_MakeLastFrame (lpHw, lpRT->lpProcessVideoHdr);
		}
	}

	/* do we have a video buffer to process */			   
	if (lpRT->lpProcessVideoHdr != NULL && (lpRT->hVideoPin == NULL || PinBufferReady(lpRT->hVideoPin))) {

		DWORD PadBytes = lpRT->lpProcessVideoHdr->dwReserved[PAD_SIZE_COUNT];

		if (PadBytes != 0) {
			OutSize = min (PAD_BUFFERSIZE, PadBytes);
			InData = (short *)lpRT->lpPadBuffer;
		} else {
			OutSize = lpRT->lpProcessVideoHdr->dwBytesUsed - lpRT->VideoEncodeIndex;
			InData = (short *)(lpRT->lpProcessVideoHdr->lpData + lpRT->VideoEncodeIndex);
		}

		/* are we in no file mode??? */
		if (lpRT->bCreateNoFile) {
			if (PadBytes == 0) {
				OutSize = lpRT->lpProcessVideoHdr->dwBytesUsed;
			}
		} else if (lpRT->bStreamMix) {
					
			buff = GetPinBuffer(lpRT->hVideoPin);
			OutSize = min (OutSize, buff->DataSize);

			/* send it to the stream mixer */
			CopyMemory (buff->data, InData, OutSize);
			buff->ValidBytes = (int)OutSize;
			PutPinBuffer(lpRT->hVideoPin, buff);
		}  else {
			unsigned long Size;

			/* write it to the file */
			WriteFile(lpRT->OutFile, InData, OutSize, &Size, NULL);
			if ((int)Size != OutSize) {
				return IDS_CAP_FILE_WRITE_ERROR;
			}
		}

		/* update the file size... */
		lpRT->dwCurrentBytes += OutSize; 

		/* are we done processing this video buffer??? */
		if (PadBytes == 0) { 
			lpRT->VideoEncodeIndex += OutSize;
		} else {
			lpRT->lpProcessVideoHdr->dwReserved[PAD_SIZE_COUNT] -= OutSize;
		}

		if (lpRT->VideoEncodeIndex == lpRT->lpProcessVideoHdr->dwBytesUsed) {

			/* update the status */
			lpHw->dwCurrentTimeElapsedMS = MulDiv (++lpHw->dwCurrentVideoFrame, lpHw->dwMicroPerInterrupt, 1000L);                         

			/* update user status - if there is time */
			if (lpHw->lpfnStatusCallback && lpRT->VideoDoneCount <= MAX_VIDEO_DONE && lpRT->AudioDoneCount <= MAX_AUDIO_DONE) {

				static DWORD cf, sec, ms;
					
				RT_LoadAviCapString(lpHw, IDS_CAP_STAT_VIDEOCURRENT, buffer, sizeof(buffer));
							
				cf   = lpHw->dwCurrentVideoFrame;
				sec  = lpHw->dwCurrentTimeElapsedMS/1000;
				ms   = lpHw->dwCurrentTimeElapsedMS - (sec*1000);

				wsprintf (message, buffer, cf, 0, sec, ms);
				
				lpHw->lpfnStatusCallback(hWnd, IDS_CAP_STAT_VIDEOCURRENT, message);
			}

			/* does user wish to continue capture??? */
			if (lpHw->lpfnCapControlCallback) {	
				LRESULT lResult;

				lResult = lpHw->lpfnCapControlCallback (hWnd, CONTROLCALLBACK_CAPTURING);
				if (!lResult) {
					if (RT_StopLoop (lpHw)) {
						return IDS_CAP_END;
					}
				}
			}	

			/* yield callback is called at least once per frame */
			if (lpHw->lpfnYieldCallback) {
				lpHw->lpfnYieldCallback (hWnd);
			}

			/* 
			 * we're mpeg stream based, not frame based, so if the user 
			 * doesn't want the data written to disk, we just quit!!!
			 */
			if (lpHw->lpfnVideoStreamCallback) {

				LRESULT lResult;

				lResult = lpHw->lpfnVideoStreamCallback (hWnd, lpRT->lpProcessVideoHdr);
				if (!lResult && !lpRT->bCreateNoFile) {
					if (RT_StopLoop (lpHw)) {
						return IDS_CAP_END;
					}
				}
			}
				
			/* all done with this one */
			lpRT->lpProcessVideoHdr->dwBytesUsed = 0;
			lpRT->lpProcessVideoHdr->dwFlags &= ~WHDR_DONE;
			lpRT->lpProcessVideoHdr->dwFlags &= ~VHDR_INQUEUE;

			/* put the buffer back to the capture driver */
			VC_AddReadyTail(lpHw->vh, lpRT->lpProcessVideoHdr);

			lpRT->lpProcessVideoHdr = NULL;
   			lpRT->VideoEncodeIndex = 0;
			
			lpRT->VideoDoneCount--;

			/* stop on last frame */
			if (lpRT->bLastFrame) {
				lpRT->StopCodeWritten = TRUE;
				return IDS_CAP_END;															  
			}

			/* 
			 * check if max file size is exceeded...
			 */
			return RT_CheckFileSize (lpHw);
		}
	}

	return 0;
}


