/*
 * hardware.h
 *
 * 32-bit Video Capture driver
 * hardware-specific data structures and definitions for Video Tool
 *
 * Mark Vrabel, October 1995
 */

#ifndef _HARDWARE_H_
#define _HARDWARE_H_

#ifdef __cplusplus
extern "C" {				    
#endif

/* register layout details of video tool board */
#include <vtdioc.h>
#include <vt_acc.h>
#include <vt_regs.h>
#include <vidtool.h>
#include <bwuser.h>
#include <bwdlgs.h>
#include <rcuser.h>
#include <videobuf.h>
#include "realtime.h"

#define PAL_PASS_SIZE        (((360L*288L)*2L)/sizeof(DWORD))
#define NTSC_PASS_SIZE       (((360L*242L)*2L)/sizeof(DWORD))
#define MAX_PASS_SIZE        (PAL_PASS_SIZE)
#define NUM_PASS_BUFS        (1L)

#define MAX_CAPT_SIZE        ((0x10000L)/sizeof(DWORD))
#define NUM_CAPT_BUFS        (1L)

#define REQUIRED_BUFFER_SIZE ((NUM_PASS_BUFS*MAX_PASS_SIZE) + ((NUM_CAPT_BUFS)*MAX_CAPT_SIZE)) 

#define Timeout(From,ms) ((timeGetTime()-From) > ms)
#define PERIOD_MIN (5)
#define MAX_NOFRAME (1000/PERIOD_MIN)


#define FORBIDDEN  0
#define I_FRAME    1
#define P_FRAME    2
#define B_FRAME    3


#define STR_BROADWAY_HARDWARE_SUBKEY "SOFTWARE\\DataTranslation\\Broadway\\HardwareOptions"
#define STR_AUDIO_CAL_FACTOR         "AudioCalFactor"
#define STR_DRIVER_INSTALL			 "DriverInstallComplete"


#define VOVERLAY_HANDLE  DWORD

__inline DWORD ByteSwap(DWORD dwX)
{

#ifdef _X86_

    _asm    mov     eax, dwX
    _asm    bswap   eax
    _asm    mov     dwX, eax

    return dwX;

#else

    return _lrotl(((dwX & 0xFF00FF00) >> 8) | ((dwX & 0x00FF00FF) << 8), 16);

#endif

}



#pragma pack(1) 

/* hardware-specific data structure - stored in device extension and
 * accessed by VC_GetHWInfo(vh)
 */
typedef struct _HWINFO {
    DWORD SizeStruct;					// size of this structure

    /* Device identification */
    DWORD DeviceIndex;					// device index (device node for Win95)
    DEVID DevID;						// board hardware device ID
	UCHAR RevisionID;					// board hardware revision
	VCUSER_HANDLE vh;	 				// owns this device
    
    volatile VT_REGS FAR *lpRegs;		// device register information

    /* video buffers */
	DWORD dwPhysicalAddr;				// base physical address
	DWORD FAR *lpLinearAddr;			// base linear address
	DWORD dwMemorySize;					// size of video memory
	
	DWORD dwCaptHead,dwCaptTail;		// head and tail of circular capture buffer
	DWORD dwAvgCalcFrameSize;			// average calculated frame size
	DWORD dwMinCalcFrameSize;			// minimum calculated frame size
	DWORD dwMaxCalcFrameSize;			// maximum calculated frame size

    VIDEO_BUFFER_LIST PassBufs;			// passthru video buffers
    VIDEO_BUFFER_LIST CaptBufs;			// capture video buffers

    /* stream position variables */
    DWORD  dwInterruptCount;			// true count of ints since stream began
    DWORD  dwVideoTime;					// time (based on frame syncs) since start
    DWORD  dwNextFrameNr;				// next expected frame
    DWORD  dwMicroPerInterrupt;			//  - interrupt rate based on NTSC/PAL
    DWORD  dwMicroPerFrame;				// microsecs per frame desired
    DWORD  dwNextFrameTime;				// next capture time in msecs from stream start
    BOOL   bVideoIn;					// are we actually doing video in ?
		
    DWORD  dwTimePrevious;				// time of previous frame
    DWORD  dwLastInterrupt;				// time of previous interrupt
	DWORD  dwStartTime;					// actual system start time
    DWORD  dwEndTime;					// actual system end time
    int    nAudioDelay;					// calculated audio delay
	double dfAudioFudgeFactor;

    /* format information */
	BW_CAPTUREOPTIONS CaptureOptions;			// capture options
	BW_RECODEOPTIONS RecodeOptions;				// Realtime IBP capture options
	BW_CAPTURE_PREFERENCES CapturePreferences;  // capture preferences
	BW_RECODE_PREFERENCES RecodePreferences;	// Realtime IBP capture preferences

    CAPTUREINFO CaptureInfo;			// hardware capture info
    ULONG VideoMode;					// NTSC or PAL
    ULONG VideoConnector;				// Composite of SVideo
	DWORD nSizeImage;					// size of frame buffer

    /* state flags */
    BOOL bCapture;						// capture flag
    BOOL bOverlay;						// overlay flag
	BOOL bPreview;						// preview flag
	BOOL bYield;						// yeild flag
	BOOL bStoppingNow;					// stopping flag
    
	/* single frame capture information */
    LPUCHAR lpLastFrame;				// storage for last good frame
	DWORD dwLastFrameMax;				// size of the buffer
    DWORD dwLastFrameSize;				// size of last frame (in bytes)
    BOOL bCaptureFrame;					// capture frame flag
    HANDLE hFrameEvent;					// set when single frame is ready

    /* overlay shtuff */
    VOVERLAY_HANDLE  voh;				// handle to overlay engine
    
    /* debug log file info */
    BOOL bLogFileEnable;				// use log file flag
    DWORD TimerOverRun;					// missed timer ticks
    DWORD FifoOverFlow;					// fifo over flow
    DWORD NoStartCode;					// missed start code
    DWORD FrameAllZeros;				// bogus frame of all zeros
    DWORD FrameTooBig;					// Frame bigger than buffer
    DWORD ExtractFrame;					// VC_ExtractFrame error
	DWORD CopyOverRun;					// buffer copy overrun error
	DWORD NoBuffers;					// no buffers on done queue
	DWORD FifoReset;					// had to reset fifo to resynch
	DWORD Repeated;						// frames repeated
	DWORD Skipped;						// frames skipped
	DWORD MultiFrame;					// mutiframe count
	DWORD ResetFifoCount;				// fifo reset flag
	DWORD BadToggle;                    // toggle mode error
	BOOL  bIsStopping;					// realtime stopping flag
	DWORD dwRepeatFrames;				// number of frames to repeat
	DWORD dwCaptureStopped;             // realtime capture stopped (end of tape)

	/* realtime capture... */
	RTINFO RTInfo;						// realtime capture structure
	CAPTURE_MODE CaptureMode;			// Capture mode - I-frame or IBP
	FARPROC OrigCapWndProc;				// original capture window proc
	HWND hWndCap;						// capture window

	/* status for realtime capture */
	DWORD dwCurrentVideoFrame;			// Current frame count    
	DWORD dwCurrentVideoFramesDropped;	// number of frames dropped
	DWORD dwCurrentWaveSamples;			// Current audio samples
	DWORD dwCurrentTimeElapsedMS;		// elapsed capture time
	BOOL fCapturingNow;					// stream capture flag
	UINT wNumVideoAllocated;			// video buffers allocated
	UINT wNumAudioAllocated;			// audio buffers allocated
	
	/* capture window calback procedures */
	FARPROC lpfnErrorCallback;			// error callback
	FARPROC lpfnStatusCallback;			// status callback
	FARPROC lpfnYieldCallback;			// yield callback
	FARPROC lpfnFrameCallback;			// frame callback
	FARPROC lpfnVideoStreamCallback;	// video stream callback
	FARPROC lpfnWaveStreamCallback;		// audio callback
	FARPROC lpfnCapControlCallback;		// capture control callback

	/* misc info */
	HANDLE hVfw;						// video for window dll module handle
	BOOL fCaptureInit;					// used when yielding, capture is initialized
	HANDLE hCaptureThread;				// used when as background capture thread when yielding
	HANDLE hCaptureComplete;			// used to signal capture loop it's done 
	BOOL bSkipFrame;					// used to signal capture loop it's time to skip a frame 
	BOOL bRepeatFrame;					// used to signal capture loop it's time to repeat a frame 
	BOOL bGotIFrame;					// got I-Frame flag for MPEG frame skip/repeat
	BOOL bMultiFrame;					// more than one frame in circular buffer
	BOOL bPreviousOverlay;				// overlay mode prior to stream capture
	BOOL bNoDialog;						// do configuration with no dialog
	BOOL bNoRestart;					// don't restart microcode
	BOOL dwTimeOffset;					// stream capture timer offset
	BOOL dwLast;						// used to check for buffer overruns
	DWORD dwMaxFileSize;				// PictureTel's file boundry size
	BOOL IsBroadway;					// Broadway app flag
	BOOL StreamInit;					// initialized for mpeg capture
	CRITICAL_SECTION RTLoopCritical;	// Critical section for Real-time capture loop
	BOOL bCreateNoFile;					// set when capturing w/no file
	DWORD dwLostFrames;					// frames lostin a row
	DWORD dwSkipFrame;					// count of frames to skip on a bad vsynch
	BOOL bCanDrawNow;					// can draw overlay now flag
	BOOL bCriticalInit;					// critical section init flag
	BOOL bRestart;						// real-time restart flag
	DWORD dwTimeRestart;				// c-cube restart time;

	/* global stuff for broadway app */
    UINT AudioEnable;					// audio enable
    UINT AudioMode;						// audio channels
    UINT AudioRate;						// audio sample rate
    UINT Quality;						// audio capture quality

} HWINFO, *PHWINFO, FAR *LPHWINFO;

#pragma pack()



/*
 * Inline handle verification
 */
#define IsBadHWINFO(p) ((p) == NULL || IsBadWritePtr((p), sizeof(HWINFO)) || (p)->SizeStruct != sizeof(HWINFO))


/* frame timing. we get one interrupt per frame (we only enable
 * even field interrupts) so the ms per interrupt is the frame not field rate
 */
#define PAL_MICROSPERFRAME         (40000)
#define NTSC_MICROSPERFRAME        (33367)

/* functions called from dispatch code via callback table */
BOOL HW_ConfigFormat(VCUSER_HANDLE, LPCONFIG_INFO);
BOOL HW_ConfigDisplay(VCUSER_HANDLE, LPCONFIG_INFO);
BOOL HW_ConfigSource(VCUSER_HANDLE, LPCONFIG_INFO);
BOOL HW_ConfigInstall(VCUSER_HANDLE vh, LPCONFIG_INFO lpGeneric);
BOOL HW_ConfigProperties(VCUSER_HANDLE vh, LPCONFIG_INFO lpGeneric);

DWORD HW_GetOverlayMode(VCUSER_HANDLE);
BOOL HW_SetKeyRGB(VCUSER_HANDLE, LPRGBQUAD);
BOOL HW_SetKeyPalIdx(VCUSER_HANDLE, ULONG);
ULONG HW_GetKeyColour(VCUSER_HANDLE);
BOOL HW_SetOverlayWindow (VCUSER_HANDLE vh, HWND hWnd);
BOOL HW_SetOverlayRects(VCUSER_HANDLE, LPOVERLAY_RECTS);
BOOL HW_SetOverlayOffset(VCUSER_HANDLE, LPRECT);
BOOL HW_Overlay(VCUSER_HANDLE, BOOL);
BOOL HW_Capture(VCUSER_HANDLE, BOOL);

BOOL HW_DrawFrame(VCUSER_HANDLE, LPDRAWBUFFER);
BOOL HW_Frame(VCUSER_HANDLE, LPVIDEOHDR);

BOOL HW_StreamInit(VCUSER_HANDLE vh, ULONG microsecperframe);
BOOL HW_StreamFini(VCUSER_HANDLE vh);
BOOL HW_StreamStart(VCUSER_HANDLE vh);
BOOL HW_StreamStop(VCUSER_HANDLE vh);
ULONG HW_StreamGetPosition(VCUSER_HANDLE);
VOID HW_StreamReportSkip(VCUSER_HANDLE vh, DWORD dwCount);

ULONG HW_RepeatFrame(VCUSER_HANDLE vh, LPUCHAR lpInput, ULONG InputLength, LPUCHAR lpOutput, ULONG OutputLength, LPULONG lpTimeStamp);					 
BOOL HW_InterruptAcknowledge(VCUSER_HANDLE vh, BOOL FAR *lpServiced, BOOL FAR *lpRepeat);
ULONG HW_CaptureService(VCUSER_HANDLE vh, LPUCHAR lpBuffer, LPULONG pTimeStamp, ULONG BufferLength, BOOL FAR* lpKeyFrame);

BOOL HW_Cleanup(VCUSER_HANDLE);
BOOL HW_DeviceOpen(VCUSER_HANDLE vh);

BOOL HW_Start (LPHWINFO lpHw, BOOL bLoad, BOOL bStop);
BOOL HW_Stop (LPHWINFO lpHw, BOOL bUnload, BOOL bStop);

DWORD HW_InterruptNoPassthru(VCUSER_HANDLE vh, LPHWINFO lpHw, DWORD Status);
DWORD HW_InterruptPassthru(VCUSER_HANDLE vh, LPHWINFO lpHw, DWORD Status);
ULONG HW_CaptureFrame(VCUSER_HANDLE vh, LPUCHAR lpBuffer, ULONG BufferLength);

BOOL HW_InitializeBuffers (VCUSER_HANDLE vh);
VOID HW_EventNotify (VCUSER_HANDLE vh);
BOOL HW_VerifyCapture ();
void HW_DynamicFifoReset(LPHWINFO lpHw);


/* virtual overlay functions */
BOOL WINAPI VO_InitializeOverlay (LPHWINFO lpHw, VOVERLAY_HANDLE FAR *lpvoh);
BOOL WINAPI VO_TerminateOverlay (VOVERLAY_HANDLE voh);
BOOL WINAPI VO_SetOverlaySourceFormat (VOVERLAY_HANDLE voh, LONG Width, LONG Height, LONG BlankLines);
BOOL WINAPI VO_SetOverlayBlankLines (VOVERLAY_HANDLE voh,LONG BlankLines);

BOOL WINAPI VO_SetOverlayDestFormat (VOVERLAY_HANDLE voh, LONG Width, LONG Height);
BOOL WINAPI VO_SetOverlayParentWindow(VOVERLAY_HANDLE voh, HWND hWnd);
BOOL WINAPI VO_StartOverlay (VOVERLAY_HANDLE voh);
BOOL WINAPI VO_StopOverlay(VOVERLAY_HANDLE voh);
BOOL WINAPI VO_PauseOverlay(VOVERLAY_HANDLE voh);
BOOL WINAPI VO_ContinueOverlay(VOVERLAY_HANDLE voh);
BOOL WINAPI VO_DrawFrame(VOVERLAY_HANDLE voh);
BOOL WINAPI VO_NotifyStreamStart (VOVERLAY_HANDLE voh, BOOL bStart);
BOOL WINAPI VO_DrawFrameNow (VOVERLAY_HANDLE voh);



#ifdef __cplusplus
}
#endif

 
#endif // _HARDWARE_H_



