/*
 * vcsubs.c
 *
 *
 * 32-bit Video Capture driver
 * Video capture subroutine support
 *
 *
 * Mark Vrabel, October 95
 */

#include <vcdriver.h>
#include "vcupriv.h"

#define MIN_RESOLUTION   1

/*
 * VC_Init
 *
 * Create the device object, and any necessary related setup, and
 * allocate device extension data. The device extension data is
 * a DEVICE_INFO struct plus however much data the caller wants for
 * hardware-specific data.
 *
 * parameters:
 *  pDriverObject - pointer to driver object (arg to DriverEntry)
 *  RegistryPathName - entry for this driver in registry (arg to DriverEntry)
 *  HWInfoSize - amount of data to allocate at end of DeviceExtension
 *
 * returns pointer to device extension data as DEVICE_INFO struct.
 */
VCUSER_HANDLE
VC_Init(ULONG HWInfoSize)
{
    VCUSER_HANDLE vh;    

    /* try to allocate the device info structure */
    vh = (VCUSER_HANDLE)GlobalAllocPtr(GPTR, sizeof(VCUSER_STRUCT)+HWInfoSize);
    if (vh == NULL) {
        return NULL;
    }

    /* zero out the structure */
    FillMemory (vh, sizeof(VCUSER_STRUCT)+HWInfoSize, 0);

    /* initialise the rest of the structure */
    vh->SizeStruct   = sizeof(VCUSER_STRUCT);
    vh->State        = State_Idle;               

	InitializeCriticalSection (&vh->BufferCritical);
	vh->bCriticalInit = TRUE;
    
    timeGetDevCaps (&vh->tc, sizeof(TIMECAPS));
    vh->Resolution = max(vh->tc.wPeriodMin, MIN_RESOLUTION);

    timeBeginPeriod(vh->Resolution);

    vh->hWorkEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    vh->hCompleteEvent = CreateEvent(NULL, FALSE, FALSE, NULL);

    if ((vh->hWorkEvent == NULL) || (vh->hCompleteEvent == NULL)) {
        return NULL;
    }

    vh->hTimerEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (vh->hTimerEvent == NULL) {
        return NULL;
    }

    return vh;
}


/* 
 * VC_Cleanup
 *
 * call to unload or abort load of the driver 
 */
VOID
VC_Cleanup(VCUSER_HANDLE vh)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return;
    }

    /* make sure the timer is dead */
    VC_StopEventTimer (vh);
    timeEndPeriod(vh->Resolution);

    /* cleanup the queues */
    VC_StreamReset(vh); 

    /* make sure we don't leave a worker thread hanging */
    VC_StreamFini(vh);

    /* and that all events are closed */
    if (vh->hWorkEvent != NULL) {
        CloseHandle(vh->hWorkEvent);
    }
    if (vh->hCompleteEvent != NULL) {
        CloseHandle(vh->hCompleteEvent);
    }
    if (vh->hTimerEvent == NULL) {
        CloseHandle(vh->hTimerEvent);
    }

    /* call the device-specific cleanup if there is one */
    if (vh->Callback.CleanupFunc != NULL) {
        vh->Callback.CleanupFunc(vh);
    }

	if (vh->bCriticalInit) {
		DeleteCriticalSection (&vh->BufferCritical);
	}
		
	/* zero out the structure */
    FillMemory (vh, sizeof(VCUSER_STRUCT), 0);

    /* Free the device info structure */
    GlobalFreePtr(vh);
}


/*
 * get the hardware specific portion of the device extension
 */
LPVOID
VC_GetHWInfo(VCUSER_HANDLE vh)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return NULL;
    }


	return( (LPVOID) (((LPUCHAR)vh) + sizeof(VCUSER_STRUCT)));
}


/*
 * get the callback table into which we can insert functions we support 
 */
LPVC_CALLBACK
VC_GetCallbackTable(VCUSER_HANDLE vh)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return NULL;
    }

    return ( &vh->Callback );
}


/*
 * set the minimum size for queued buffers
 */
VOID
VC_SetImageSize(VCUSER_HANDLE vh, DWORD ImageSize)
{
    /* verify user handle */
    if (IsBadVCUSER_HANDLE(vh)) {
        return;
    }

    vh->ImageSize = ImageSize;
}


/*
 * delay for a number of milliseconds. This is accurate only to
 * +- 15msecs at best.
 */
VOID
VC_Delay(UINT nMillisecs)
{   
   Sleep (nMillisecs);
}


/*
 * Block by polling for a number of microseconds
 */
#pragma optimize("", off)
VOID
VC_Stall(int nMicrosecs)
{
    /* fake it out - close enough??? */

   DWORD i;
   DWORD nLoops = (DWORD)((DWORD)nMicrosecs*20L);

   for (i=0;i<nLoops;i++);
}
#pragma optimize("", on)



